/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_MOSAL_DLL_H
#define H_MOSAL_DLL_H

#ifndef MT_KERNEL

typedef void* MOSAL_dll_hndl_t;

typedef enum {
  MOSAL_DLL_LOAD_GLOBAL = 1, /* External symbols defined in the library will be made available to
                                subsequently loaded libraries */
  MOSAL_DLL_LOAD_LAZY   = (1<<1) /* Resolve undefined symbols as code from the dynamic library is executed */
} MOSAL_dll_load_flags_t;


/******************************************************************************
 *  Function (user-mode only):
 *    MOSAL_dll_open
 *
 *  Description:
 *    Load a dynamically loadable library
 *    If the library is already loaded via another call to this function, 
 *    only a reference count is updated and the same handle is returned.
 *
 *  Parameters: 
 *    dllpath (IN) - Path/filename of DLL to load. 
 *                   If not full path, search is done as defined in the OS 
 *                   for library search path.
 *    load_flags (IN)  - Flags for loading operation.
 *    dll_hndl_p (OUT) - Returned handle to loaded DLL.
 *
 *  Returns:
 *    MT_OK
 *    MT_ENOENT - Failed to open library file (not found or not enough permissions)
 *
 ******************************************************************************/
call_result_t MOSAL_dll_open(const char* dllpath, 
                             const MOSAL_dll_load_flags_t load_flags,
                             MOSAL_dll_hndl_t *dll_hndl_p);

/******************************************************************************
 *  Function (user-mode only):
 *    MOSAL_dll_close
 *
 *  Description:
 *    Unload a dynamically loadable library.
 *    If MOSAL_dll_open invoked more than once for given library, this function
 *    will only decrement its reference count.
 *
 *  Parameters: 
 *    dll_hndl (IN) - Handle of DLL as returned from MOSAL_dll_open().
 *
 *  Returns:
 *    MT_OK
 *    MT_EINVAL - Invalid dll handle.
 *
 ******************************************************************************/
call_result_t MOSAL_dll_close(MOSAL_dll_hndl_t dll_hndl);

/******************************************************************************
 *  Function (user-mode only):
 *    MOSAL_dll_get_sym
 *
 *  Description:
 *    Get loading (relocation) address of given symbol within given library.
 *
 *  Parameters: 
 *    dll_hndl (IN) - Handle of DLL as returned from MOSAL_dll_open().
 *
 *  Returns:
 *    MT_OK
 *    MT_EINVAL - Invalid dll handle or unknown symbol.
 *
 ******************************************************************************/
call_result_t MOSAL_dll_get_sym(MOSAL_dll_hndl_t dll_hndl, const char* sym, void **sym_pp);


#endif /* !MT_KERNEL */

#endif /* H_MOSAL_DLL_H */

