/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_MOSAL_TIMER_H
#define H_MOSAL_TIMER_H

#include <mosal_timer_imp.h>

#ifdef MT_KERNEL

/*  typedef for MOSAL INTERRUPT object */
struct MOSAL_ISR;
typedef struct MOSAL_ISR MOSAL_ISR_t;

/******************************************************************************
 *  Function (kernel-mode only):
 *    MOSAL_ISR_set
 *
 *  Description:
 *    connect interrupt handler
 *
 *  Parameters:
 *	  isr_p(IN) MOSAL_ISR_t *
 *		  interrupt object
 *    handler(IN) intr_handler_t
 *        Pointer to the intr. handler.
 *    irq(IN)   MOSAL_IRQ_ID_t
 *        The IRQ number of the intr which invokes this handler.
 *    name(IN) char *
 *        Just informationa name. In Linux environment
 *        that name will be presented in /proc/interrupts
 *    dev_id(IN)   MT_ulong_ptr_t
 *        Unique device ID. Use on intr sharing.
 *        If NULL, the device may NOT share the IRQ
 *        with other devices (handlers).
 *        Otherwise, this pointer may be used as the "This"
 *        pointer of device data object when calling handler.
 *
 *  Returns:
 *    call_result_t
 *        0-OK -1-Error
 *
 ******************************************************************************/
call_result_t MOSAL_ISR_set(MOSAL_ISR_t * isr_p,
			    MOSAL_ISR_func_t handler,
			    MOSAL_IRQ_ID_t irq,
			    char *name, MT_ulong_ptr_t dev_id);

/******************************************************************************
 *  Function (kernel-mode only):
 *    MOSAL_ISR_unset
 *
 *  Description:
 *    disconnect interrupt handler
 *
 *  Parameters:
 *	  isr_p(IN) MOSAL_ISR_t *
 *		  interrupt object
 *
 *  Returns:
 *    call_result_t
 *        0-OK -1-Error
 *
 ******************************************************************************/
call_result_t MOSAL_ISR_unset(MOSAL_ISR_t * isr_p);

/******************************************************************************
 *  Function (kernel-mode only):
 *    MOSAL_set_intr_handler
 *
 *  Description:
 *    Sets the given interrupt handler to be called back on the given IRQ.
 *    If dev_id!=NULL, the handler would be linked to IRQ only if previously
 *    set handler were sharing handlers (had dev_id!=NULL), too.
 *    This means that a non-sharing handler may be the only handler which
 *    is called back on a given IRQ.
 *
 *  Parameters:
 *    handler(IN) intr_handler_t
 *        Pointer to the intr. handler.
 *    irq(IN)   MOSAL_IRQ_ID_t
 *        The IRQ number of the intr which invokes this handler.
 *    name(IN) char *
 *        Just informationa name. In Linux environment
 *        that name will be presented in /proc/interrupts
 *    dev_id(IN)   void*
 *        Unique device ID. Use on intr sharing.
 *        If NULL, the device may NOT share the IRQ
 *        with other devices (handlers).
 *        Otherwise, this pointer may be used as the "This"
 *        pointer of device data object when calling handler.
 *
 *  Returns:
 *    call_result_t
 *        0-OK -1-Error
 *
 ******************************************************************************/
call_result_t MOSAL_set_intr_handler(intr_handler_t handler,
				     MOSAL_IRQ_ID_t irq,
				     char *name, void *dev_id);

/******************************************************************************
 *  Function (kernel-mode only):
 *    MOSAL_unset_intr_handler
 *
 *  Description:
 *    Removes the given interrupt handler of the callback chain on the given IRQ.
 *
 *  Parameters:
 *    handler(IN) intr_handler_t
 *        Pointer to the intr. handler.
 *    irq(IN) MOSAL_IRQ_ID_t
 *        The IRQ number of the intr which invokes this handler.
 *    dev_id(IN) void*
 *        Unique device ID. Use on intr sharing.
 *
 *  Returns:
 *    call_result_t
 *        MT_OK if success, MT_ERROR if failed.
 *
 ******************************************************************************/
call_result_t MOSAL_unset_intr_handler(intr_handler_t handler,
				       MOSAL_IRQ_ID_t irq, void *dev_id);

#endif				/* MT_KERNEL */

/* ////////////////////////////////////////////////////////////////////////////// */

/*  Time functions  */

/* ////////////////////////////////////////////////////////////////////////////// */

#ifdef MT_KERNEL
 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_time_get_clock
 *
 *  Description:
 *		get current system clock (in microseconds)
 *
 *  Parameters:
 *		ts(OUT) - pointer to a structure, describing the time
 *
 *  Returns:
 *
 *  Notes:
 *		Callers of this routine must be running at IRQL <= DISPATCH_LEVEL
 *
 ******************************************************************************/
void MOSAL_time_get_clock(MOSAL_timespec_t * ts);


 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_time_compare
 *
 *  Description:
 *		compare 2 absolute times
 *
 *  Parameters:
 *		ts1(IN) - pointer to a structure, describing the time
 *		ts2(IN) - pointer to a structure, describing the time
 *
 *  Returns:
 *		positive - when ts1 > ts2
 *		negative - when ts1 < ts2
 *		zero	 - when ts1 = ts2
 *
 *
 ******************************************************************************/
static __INLINE__ int MOSAL_time_compare(MOSAL_timespec_t * ts1,
					 MOSAL_timespec_t * ts2)
{
	if (ts1->tv_sec > ts2->tv_sec)
		return 1;
	if (ts1->tv_sec < ts2->tv_sec)
		return -1;
	if (ts1->tv_nsec < ts2->tv_nsec)
		return -1;
	return ts1->tv_nsec > ts2->tv_nsec;
}

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_time_add_usec
 *
 *  Description:
 *		enlarge an absolute time 'ts' by 'usec' of microseconds
 *
 *  Parameters:
 *		ts(IN) 		- pointer to a structure, describing the time
 *		usecs(IN) 	- a POSITIVE number of microseconds to add
 *
 *  Returns:
 *		updated 'ts'
 *
 ******************************************************************************/
static __INLINE__ MOSAL_timespec_t *MOSAL_time_add_usec(MOSAL_timespec_t * ts,
							long usecs)
{
	ts->tv_sec += usecs / 1000000L;
	ts->tv_nsec += (usecs % 1000000L) * 1000;
	if (ts->tv_nsec > 1000000000L) {
		ts->tv_sec++;
		ts->tv_nsec -= 1000000000L;
	}
	return ts;
}

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_time_init
 *
 *  Description:
 *		Zero 'ts' structure
 *
 *  Parameters:
 *		ts(IN) 		- pointer to a structure, describing the time
 *		usecs(IN) 	- a POSITIVE number of microseconds to add
 *
 *  Returns:
 *		updated 'ts'
 *
 ******************************************************************************/
static __INLINE__ void MOSAL_time_init(MOSAL_timespec_t * ts)
{
	ts->tv_sec = ts->tv_nsec = 0;
}
#endif				/* MT_KERNEL */

 /******************************************************************************
 *  Function (inline): 
 *		MOSAL_get_time_counter
 *
 *  Description:
 *		get te current value of a counter that progresses monotonically with time
 *
 *  Parameters:
 *
 *  Returns:
 *		value of the counter or 0 if not supported by the architecture
 *
 ******************************************************************************/
static __INLINE__ u_int64_t MOSAL_get_time_counter(void)
{
	MT_MOSAL_GET_TIME_COUNTER;
}

/******************************************************************************
 *  Function: MOSAL_get_counts_per_sec
 *
 *  Description: get number of counts in 1 sec (refer to MOSAL_get_time_counter)
 *
 *  Parameters:
 *
 *  Returns:
 *    Number of counts in 1 sec or 0 if not supported
 *
 ******************************************************************************/
u_int64_t MOSAL_get_counts_per_sec(void);

#endif				/* H_MOSAL_TIMER_H */
