/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_MTL_COMMON_H
#define H_MTL_COMMON_H

#ifdef __cplusplus
extern "C" {
#endif

#if defined(__mppc__) && ! defined(MT_BIG_ENDIAN)
#define MT_BIG_ENDIAN 1
#endif

#include <mtl_types.h>
#include <bit_ops.h>

#define MAX_MTL_LOG_TYPES      3
#define MAX_MTL_LOG_SEVERITIES 8
#define MAX_MTL_LOG_LEN        512
#define MAX_MTL_LOG_LAYER        32
#define MAX_MTL_LOG_INFO        512
	typedef enum {
		mtl_log_trace = 0,
		mtl_log_debug = 1,
		mtl_log_error = 2
	} mtl_log_types;

/*  Prepend File-name + Line-number to formatted message.
 *  Example:  MTL_DEBUG4(MT_FLFMT("(1+1/n)^n -> %g (%s)"), M_E, "Euler")
 */
#define MT_FLFMT(fmt)  "%s[%d]: " fmt "\n", mtl_basename(__FILE__), __LINE__
#define MT_STDMSG(fmt) "%s:%s:%s:%d:"MT_PID_FMT": " fmt "\n", mtl_process_name(), __func__, mtl_basename(__FILE__), __LINE__, MOSAL_getpid()
#define MT_STDFMT(fmt) "%s["MT_PID_FMT"]: %s:%u: " fmt "\n", MOSAL_get_proc_name(), MOSAL_getpid(), mtl_basename(__FILE__), __LINE__
/* OS-dependent stuff */
#include <mtl_sys_defs.h>

/* Convenient macros doing  cast & sizeof */
#define TNMALLOC(t, n)      (t*)MALLOC((n) * sizeof(t))
#define TMALLOC(t)          TNMALLOC(t, 1)
#define TNVMALLOC(t, n)     (t*)VMALLOC((n) * sizeof(t))
#define TVMALLOC(t)         TNVMALLOC(t, 1)
#define TNINTR_MALLOC(t, n) (t*)INTR_MALLOC((n) * sizeof(t))
#define TINTR_MALLOC(t)     TNINTR_MALLOC(t, 1)

#ifndef POWER2
#define POWER2(power) (1 << (power))
#else
#error use different name for power of 2
#endif

#ifdef USE_RELAY_MOD_NAME
	extern void mtl_log_set_name(char *mod_name);
#endif

	extern const char *mtl_strerror(call_result_t errnum);
	extern const char *mtl_strerror_sym(call_result_t errnum);
	extern const char *mtl_basename(const char *filename);	/* trim dir-path */

/******************************************************************************
 *  Function: mtl_log_set
 *
 *  Description: Setup log print in kernel module.
 *
 *  Parameters:
 *    layer(IN)  (LEN s) char*
 *    info(IN)	 (LEN s) char*
 *
 *  Returns:
 ******************************************************************************/
	void mtl_log_set(char *layer, char *info);

	extern void mtl_log(const char *layer, mtl_log_types log_type, char sev,
			    const char *fmt, ...)
	    __attribute__ ((format(printf, 4, 5)));

	extern void mtl_common_cleanup(void);

/* slear lower bits in "val" using '1 << mask'
 * e.g. MT_MASKX(0xFFFFF,8) -> 0xFFF00
 */
#define MT_ROUND_DOWN(fname, ftype) \
static inline ftype MT_DOWN_ALIGNX_##fname(ftype val, unsigned long mask) \
{ \
	return (val & ((~(ftype)0UL) << mask)); \
}

/*
 * Round up 'val' to its lower 'mask' bits.
 * e.g. MT_UP_ALIGNX(0x1002, 8) -> 0x1100
 */
#define MT_ROUND_UP(fname, ftype) \
static inline ftype MT_UP_ALIGNX_##fname(ftype val, unsigned long mask) \
{ \
	return ((val + ( ((ftype)1 << mask) - 1) ) & ((~(ftype)0UL) << mask) ); \
}

MT_ROUND_DOWN(VIRT, MT_virt_addr_t)
MT_ROUND_UP  (VIRT, MT_virt_addr_t)

MT_ROUND_DOWN(PHYS, MT_phys_addr_t)
MT_ROUND_UP  (PHYS, MT_phys_addr_t)

MT_ROUND_DOWN(U32, u_int32_t)
MT_ROUND_UP  (U32, u_int32_t)

MT_ROUND_DOWN(U64, u_int64_t)
MT_ROUND_UP  (U64, u_int64_t)

MT_ROUND_DOWN(ULONG_PTR, MT_ulong_ptr_t)
MT_ROUND_UP  (ULONG_PTR, MT_ulong_ptr_t)

MT_ROUND_DOWN(SIZE, MT_size_t)
MT_ROUND_UP  (SIZE, MT_size_t)

MT_ROUND_DOWN(ULONG, unsigned long)
MT_ROUND_UP  (ULONG, unsigned long)

#undef MT_ROUND_DOWN
#undef MT_ROUND_UP


/* move to pci_ids.h */
#define MT_MELLANOX_IEEE_VENDOR_ID      0x02c9
#define MT_MELLANOX_PCI_VENDOR_ID       0x15B3
#define MT_TOPSPIN_PCI_VENDOR_ID        0x1867

/* some standard macros for tracing */
#define FUNC_IN MTL_DEBUG2("==> %s\n", __func__)
#define FUNC_OUT MTL_DEBUG2("<== %s\n", __func__)
#define MT_RETURN(rc) { FUNC_OUT; return (rc); }
#define MT_RETV { FUNC_OUT ; return; }

#ifdef __cplusplus
}
#endif
#endif				/* H_MTL_COMMON_H */
