/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_MOSAL_MEM_H
#define H_MOSAL_MEM_H

#include <mtl_types.h>
#include <mtl_errno.h>
#include <mtl_common.h>
#include <mosal_arch.h>
#include "mosal_mlock.h"

/*============ macro definitions =============================================*/
#define VA_NULL ((MT_virt_addr_t)0)
#define PA_NULL ((MT_phys_addr_t)0)
#define INVAL_PHYS_ADDR ((MT_phys_addr_t)_UI64_MAX)

typedef u_int32_t mem_attr_t;

typedef char *MOSAL_shmem_key_t;

/*MOSAL shared memory flags */
#define MOSAL_SHM_CREATE    0x1	/*Create a new shared memory region */
#define MOSAL_SHM_EXCL      0x2	/*Ensure that the new region has been created */
#define MOSAL_SHM_READONLY  0x4	/*Create read-only region */
#define MOSAL_SHM_HUGEPAGE  0x8	/*Create huge page */
/************************************/

#define MIN_PAGE_SZ						MOSAL_SYS_PAGE_SIZE
#define MIN_PAGE_SZ_MASK				(~((MT_virt_addr_t)MIN_PAGE_SZ - 1))
#define MIN_PAGE_SZ_ALIGN(x)		    (((x) + ~MIN_PAGE_SZ_MASK) & MIN_PAGE_SZ_MASK)

#define MOSAL_PAGE_ALIGN(va, size)		(((va) + ~MOSAL_PAGE_MASK((va)+(size))) \
												& MOSAL_PAGE_MASK(va + size))

typedef enum {
	MOSAL_MEM_FLAGS_NO_CACHE = 1,	/* non-chached mapping (to be used for address not in main memory) */
	MOSAL_MEM_FLAGS_PERM_READ = (1 << 1),	/* Read permission */
	MOSAL_MEM_FLAGS_PERM_WRITE = (1 << 2)
	    /* Write permission */
	    /* Note: currently write permission implies read permissions too */
} MOSAL_mem_flags_enum_t;

typedef u_int32_t MOSAL_mem_flags_t;	/* To be used with flags from MOSAL_mem_flags_enum_t */

static __INLINE__ const char *MOSAL_prot_ctx_str(const MOSAL_prot_ctx_t
						 prot_ctx)
{
	if (prot_ctx == MOSAL_get_kernel_prot_ctx()) {
		return "KERNEL";
	} else if (prot_ctx == MOSAL_get_current_prot_ctx()) {
		return "USER";
	} else {
		return "INVALID";
	}
}

static inline const char *MOSAL_perm_str(MOSAL_mem_perm_t req_perm)
{
	switch ( req_perm & 3 ) {
		case 0:
			return "NONE";
		case MOSAL_PERM_READ:
			return "READ";

		case MOSAL_PERM_WRITE:
			return "WRITE";

		case MOSAL_PERM_READ | MOSAL_PERM_WRITE:
			return "READ | WRITE";
	}
}

/******************************************************************************
 *
 *  Function(User Space only): MOSAL_io_remap
 *
 *  Description: Map a physical contigous buffer to virtual address.
 *
 *  Parameters: 
 *      pa   (IN) MT_phys_addr_t
 *           Physical address.
 *      size (IN) u_int32_t
 *           Size of memory buffer in bytes
 *               
 *
 *  Returns: On success returns pointer to new virtual memory buffer else
 *           returns zero.
 *
 *  Notes: The returned address will be page alligned. In case 'size' is not 
 *         page alligned the amount of allocated memory can be bigger than the 
 *         requested.
 *
 ******************************************************************************/
MT_virt_addr_t
#ifndef MTL_TRACK_ALLOC
    MOSAL_io_remap
#else
    MOSAL_io_remap_memtrack
#endif
    (MT_phys_addr_t pa, MT_size_t size);

#ifdef MTL_TRACK_ALLOC
#define MOSAL_io_remap(pa, size)                                                                                  \
                             ({                                                                                   \
                                MT_virt_addr_t __memtrack_rc;                                                                \
                                __memtrack_rc = MOSAL_io_remap_memtrack((pa), (size));                                           \
                                if ( __memtrack_rc != VA_NULL ) {                                                            \
                                  memtrack_alloc(MEMTRACK_IOREMAP, (unsigned long)(__memtrack_rc), (size), __FILE__, __LINE__, GFP_KERNEL);\
                                }                                                                                 \
                                __memtrack_rc;                                                                               \
                             })

#endif

void
#ifndef MTL_TRACK_ALLOC
 MOSAL_io_unmap
#else
 MOSAL_io_unmap_memtrack
#endif
 (MT_virt_addr_t va);

#ifdef MTL_TRACK_ALLOC
#define MOSAL_io_unmap(va)  do {                                                                        \
                              memtrack_free(MEMTRACK_IOREMAP, (unsigned long)(va), __FILE__, __LINE__); \
                              MOSAL_io_unmap_memtrack(va);                                                       \
                            }                                                                           \
                            while (0)
#endif

/******************************************************************************
 *
 *  Function: MOSAL_virt_to_phys
 *
 *  Description: Translate virtual address to physical.
 *
 *  Parameters: 
 *      prot_ctx(IN) source of the address (kernel or user)
 *      va (IN) const MT_virt_addr_t 
 *          Virtual address.
 *      pa_p(OUT) returned physical address
 *              
 *  Returns: MT_OK On success
 *           MT_ENOMEM when not in address space
 *           MT_ENORSC when physical page is not available
 *           MT_EINVAL invalid value of prot_ctx
 *
 ******************************************************************************/
call_result_t MOSAL_virt_to_phys(MOSAL_prot_ctx_t prot_ctx,
				 const MT_virt_addr_t va,
				 MT_phys_addr_t * pa_p);


/******************************************************************************
 *  Function: MOSAL_is_iomem
 *
 *  Description:
 *    check if the region supplied represent io memory
 *  Parameters: 
 *    prot_ctx(in): protection context
 *    va(in): start of buffer
 *    size(in): size of buffer
 *  Returns:
 *    MT_OK in case of success
 ******************************************************************************/
call_result_t MOSAL_is_iomem(MOSAL_protection_ctx_t prot_ctx,
							 MT_virt_addr_t va,
							 MT_size_t size,
							 int *iomem_p);

/******************************************************************************
 *
 *  Function (Kernel space only): MOSAL_virt_to_phys_ex
 *
 *  Description: Translate virtual address to physical.
 *
 *  Parameters: 
 *      prot_ctx(IN) source of the address (kernel or user)
 *      va (IN) Virtual address
 *      page_pp(OUT) pointer to return struct page * mapped to va
 *      pa_p(OUT) returned physical address
 *      wr_enable(IN) if set pte is made write enabled
 *              
 *  Returns: MT_OK On success
 *           MT_ENOMEM when not in address space
 *           MT_ENORSC when physical page is not available
 *           MT_EINVAL invalid value of prot_ctx
 *
 ******************************************************************************/
#if defined(MT_KERNEL)
call_result_t MOSAL_virt_to_phys_ex(MOSAL_prot_ctx_t prot_ctx,
				    const MT_virt_addr_t va,
				    struct page **page_pp,
				    MT_phys_addr_t * pa_p);
#endif

/******************************************************************************
 *
 *  Function (Kernel space only): MOSAL_phys_to_virt
 *
 *  Description: Translate physical address to virtual.
 *
 *  Parameters: 
 *      pa (IN) const MT_phys_addr_t 
 *          Physical address.
 *              
 *  Returns: On success returns a virtual address in current address space, corresponding to pa. 
 *		Else returns zero.
 *
 ******************************************************************************/
MT_virt_addr_t MOSAL_phys_to_virt(const MT_phys_addr_t pa);

/******************************************************************************
 *
 *  Function (user_only): MOSAL_virt_to_bus
 *
 *  Description: Translate a virtual address to a bus address.
 *
 *  Parameters: 
 *      va (IN) MT_virt_addr_t
 *           Virtual address.
 *              
 *  Returns: On success returns bus address pointed by va. Else returns zero.
 *
 ******************************************************************************/
MT_phys_addr_t MOSAL_virt_to_bus(MT_virt_addr_t va);

/******************************************************************************
 *
 *  Function (only include): MOSAL_test_set_bit32
 *
 *  Description: atomically set bit and return its previous value. 
 *
 *  Parameters: 
 *      off (IN) bit offset 0-31 
 *       va (IN) void pointer to dword containing bit.
 *     
 *  Returns: previous bit value.
 *
 ******************************************************************************/
#define MOSAL_test_set_bit32(off, va)   MOSAL_arch_test_set_bit32((off), (va))

/******************************************************************************
 *
 *  Function (only include): MOSAL_test_clear_bit32
 *
 *  Description: atomically clear bit and return its previous value. 
 *
 *  Parameters: 
 *      off (IN) bit offset 0-32  
 *       va (IN) void pointer to dword containing bit.
 *     
 *  Returns: previous bit value.
 *
 ******************************************************************************/
#define MOSAL_test_clear_bit32(off, va)   MOSAL_arch_test_clear_bit32((off), (va))

#define PAGE_SIZE_4M  0x400000
#define PAGE_SIZE_2M  0x200000
#define PAGE_SIZE_4K  0x1000
#define PAGE_SIZE_8K  0x2000
#define PAGE_SIZE_16K 0x4000
#define PAGE_SIZE_64K 0x10000
#define PAGE_SHIFT_4M 22
#define PAGE_SHIFT_2M 21
#define PAGE_SHIFT_4K 12
#define PAGE_SHIFT_8K 13
#define PAGE_SHIFT_16K 14
#define PAGE_SHIFT_64K 16

/******************************************************************************
 *  Function: MOSAL_get_page_shift
 *
 *  Description: 
 *    get page shift for the va in the protection context specified by prot_ctx.
 *
 *  Parameters: 
 *    prot_ctx (IN) protection context
 *    va (IN) virtual address 
 *    page_shift_p (OUT) returned page shift
 *
 *  Returns: MT_OK
 *           MT_ENOMEM when address not valid
 *
 *  Note: If the address va does not belong to address space in the specified
 *        context the function fails.
 *
 ******************************************************************************/
call_result_t MOSAL_get_page_shift(MOSAL_prot_ctx_t prot_ctx, MT_virt_addr_t va,
				   unsigned *page_shift_p);

/******************************************************************************
 *  Function: (inline function) MOSAL_get_page_size
 *
 *  Description:
 *    get page size for the va in the protection context specified by prot_ctx.
 *
 *  Parameters: 
 *    prot_ctx(IN) protection context
 *    va(IN) virtual address 
 *    page_size_p(OUT) returned page size
 *
 *  Returns: MT_OK
 *           MT_ENOMEM when address not valid
 *
 *  Note: If the address va does not belong to address space in the specified
 *        context the function fails.
 *
 ******************************************************************************/
static __INLINE__ call_result_t MOSAL_get_page_size(MOSAL_prot_ctx_t prot_ctx,
						    MT_virt_addr_t va,
						    unsigned *page_size_p)
{
	unsigned int page_shift;
	call_result_t rc;

	rc = MOSAL_get_page_shift(prot_ctx, va, &page_shift);
	if (rc != MT_OK) {
		return rc;
	}
	*page_size_p = 1 << page_shift;
	MTL_TRACE1(MT_FLFMT("%s: va=" VIRT_ADDR_FMT ", prot=%s, page_size=%d"),
		   __func__, va, MOSAL_prot_ctx_str(prot_ctx), *page_size_p);
	return rc;
}

/******************************************************************************
 *  Function 
 *    MOSAL_shmget:
 *
 *  Description:
 *    Retrieve a shared memory region for the given key
 *  Parameters: 
 *    key(IN) MOSAL_shmem_key_t - A unique key for the memory region 
 *    size(IN) MT_size_t - size of needed memory region
 *    flags(IN) - permitions flags that may be:
 *      MOSAL_SHM_CREATE    - create 
 *      MOSAL_SHM_EXCL      - fall if the memory region with the given key 
 *      MOSAL_SHM_READONLY  - readonly region (by default region will be created 
 * with read-write permitions
 * already exists
 *    
 *    id_p(OUT) MOSAL_shmid_t - id of created region (in case of success)
 *
 *  Returns:
 *    MT_OK - in case of success
 *    MT_EACCES - if user has no permitions to access this memory region
 *    MT_EAGAIN - no resources
 *    MT_EBUSY  - if MOSAL_SHM_CREATE | MOSAL_SHM_EXCL was set in flags and the region with 
 * such a key already exists
 *    MT_EINVAL  - otherwise
 *
 ******************************************************************************/
call_result_t MOSAL_shmget(MOSAL_shmem_key_t key,
			   MT_size_t size,
			   u_int32_t flags, MOSAL_shmid_t * id_p);

/******************************************************************************
 *  Function 
 *    MOSAL_shmat:
 *
 *  Description:
 *    Attaches previously allocated shared memory region to virtual address 
 *  space of the calling process.
 *  Parameters: 
 *    id(IN) MOSAL_shmem_key_t - ID of shared memory region 
 *    flags(IN) - permitions flags that may be:
 *      MOSAL_SHM_READONLY  - create read-only memory region
 *      otherwise the region will be mapped with read-write permitions
 *    addr_p(OUT) - Start virtual address of the mapped shared region in case of success
 *
 *  Returns:
 *    MT_OK - incase of seccess
 *    MT_EACCES - If calling process has no permissions to access the region with
 *  the given ID
 *    MT_EINVAL - Bad id
 *    MT_EAGAIN - Couldn't allocate page tables for a new address range or for a descriptor
 *    MT_ERROR  - otherwise
 *
 ******************************************************************************/
call_result_t MOSAL_shmat(MOSAL_shmid_t id, int flags, void **addr_p);

/******************************************************************************
 *  Function 
 *    MOSAL_shmdt:
 *
 *  Description:
 *    Detaches virtual address mapping of previously attached shared memory region .
 *  Parameters: 
 *    addr(IN) void * - Virtual address do detach
 *  Returns:
 *      MT_OK - success
 *      MT_EINVAL - Bad address
 ******************************************************************************/
call_result_t MOSAL_shmdt(void *addr);

/******************************************************************************
 *  Function 
 *    MOSAL_shmrm:
 *
 *  Description:
 *    Schadules for removing the shared memory region with the given ID
 *  Parameters: 
 *    id(IN) MOSAL_shmid_t - ID of the memory region
 *  Returns:
 *    MT_OK in case of success
 *    MT_EACCES if called by the user which is not a creator of the region
 *    MT_EINVAL -otherwise
 ******************************************************************************/
call_result_t MOSAL_shmrm(MOSAL_shmid_t id);

/* this macro allows to call mapping functions from user space */
/* #define DEBUG_MAP_FUNCTIONS */

#if defined(MT_KERNEL)

typedef struct MOSAL_phys_ctg_st *MOSAL_phys_ctg_t;
typedef struct MOSAL_iomem_st *MOSAL_iomem_t;

typedef enum {
	MOSAL_MEM_ALLOC_FLAGS_NONE = 0,
	MOSAL_MEM_ALLOC_FLAGS_LAST
} MOSAL_mem_alloc_flags_t;

typedef struct {
	MT_phys_addr_t pa;	/* physical address of the buffer */
	MT_u_int_t order;	/* log2 of size in bytes */

	unsigned int k_map_count;	/* number of times the object is mapped */
	MOSAL_mem_flags_t k_map_flags;	/* mapping flags */
	MT_virt_addr_t k_va;	/* virtual address */

	unsigned int u_map_count;	/* number of times the object is mapped */
	MOSAL_mem_flags_t u_map_flags;	/* mapping flags */
	MT_virt_addr_t u_va;	/* virtual address */
} MOSAL_phys_ctg_props_t;

typedef struct {
	MT_phys_addr_t pa;	/* physical address of the buffer */
	MT_virt_addr_t va;	/* virtual address of the buffer */
	MT_size_t size;		/* size in bytes */
	MOSAL_prot_ctx_t prot_ctx;	/* protection context to which object is mapped */
} MOSAL_iomem_props_t;

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_phys_ctg_alloc
 *
 *  Description: allocate physically contigous host memory (RAM)
 *
 *  Parameters:
 *    log2_size(in) log2 of size of requested memory
 *    log2_align(i) min number of zeros of the physical address (alignment)
 *    flags(in) physical memory required atributes
 *    pobj_p(out) returned object
 *               
 *
 *  Returns: MT_OK success
 *           MT_EINVAL invalid argument
 *           MT_ENOMEM no memory
 *
 ******************************************************************************/
call_result_t MOSAL_mem_phys_ctg_alloc(IN MT_u_int_t log2_size,
				       IN MT_u_int_t log2_align,
				       IN MOSAL_mem_alloc_flags_t flags,
				       OUT MOSAL_phys_ctg_t * pobj_p);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_phys_ctg_free
 *
 *  Description: de-allocate memory and destroy the object.
 *
 *  Parameters:
 *    pobj(in) object to be destroyed
 *               
 *
 *  Returns: MT_OK success
 *
 *  Notes: 1. The function unmaps the physical memory first from both protection
 *            contexts and then destroys the object
 *
 ******************************************************************************/
call_result_t MOSAL_mem_phys_ctg_free(IN MOSAL_phys_ctg_t pobj);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_phys_ctg_map
 *
 *  Description: map the physical memory to prot_ctx's address space
 *
 *  Parameters:
 *    pobj(in) object to map
 *    prot_ctx(in) protection context
 *    flags(in) mapping attributes
 *    va_p(out) virtual address
 *               
 *
 *  Returns: MT_OK success
 *           MT_EINVAL nvalid argument
 *           MT_ENOMEM no memory
 *           MT_EINTR
 *
 *  Note: 1. va_p may be passed NULL. In this case mapping will take place and
 *           the virtual address can be later retrieved by a call to
 *           MOSAL_mem_phys_ctg_get_props
 *        2. the function may be called multiple times for kernel context and once
 *           for user context (any one). Calling more than once for the same
 *           user context will fail. For kernel mapping, a ref count is maintained
 *           and the same number of calls to MOSAL_mem_phys_ctg_unmap must be made
 *           to actually unmap the memory.
 *
 ******************************************************************************/
call_result_t MOSAL_mem_phys_ctg_map(IN MOSAL_phys_ctg_t pobj,
				     IN MOSAL_prot_ctx_t prot_ctx,
				     IN MOSAL_mem_flags_t flags,
				     OUT MT_virt_addr_t * va_p);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_phys_ctg_unmap
 *
 *  Description: unmap the physical memory from prot_ctx's address space
 *
 *  Parameters:
 *    pobj(in) object to map
 *    prot_ctx(in) protection context
 *
 *  Returns:
 *    MT_ENOMEM  when it is not mapped
 *    MT_OK success
 *
 ******************************************************************************/
call_result_t MOSAL_mem_phys_ctg_unmap(IN MOSAL_phys_ctg_t pobj,
				       IN MOSAL_prot_ctx_t prot_ctx);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_phys_ctg_get_props
 *
 *  Description: get the properties of the object
 *
 *  Parameters:
 *    pobj(in) object to map
 *    props_p(out) returned properties
 *
 *  Returns: MT_OK success
 *
 ******************************************************************************/
call_result_t MOSAL_mem_phys_ctg_get_props(IN MOSAL_phys_ctg_t pobj,
					   OUT MOSAL_phys_ctg_props_t *
					   props_p);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_iomem_map
 *
 *  Description: map the io memory to prot_ctx's address space
 *
 *  Parameters:
 *    prot_ctx(in) protection context
 *    flags(in) mapping attributes
 *    pa(in) physical address which belongs to an io device
 *    size(in) size of mapping
 *    iomemobj_p(out) object representing the mapping. allocated by the caller.
 *    va_p(out) virtual address
 *               
 *
 *  Returns: MT_OK success
 *           MT_ENOMEM no memory
 *
 *  Note: 1. va_p may be passed NULL. In this case mapping will take place and
 *           the virtual address can be later retrieved by a call to MOSAL_iomem_get_props
 *        2. The function may be called multiple times for kernel context and once
 *           for for user context (any one). Calling more than once for the same
 *           user context will fail.
 *
 ******************************************************************************/
call_result_t MOSAL_iomem_map(IN MOSAL_prot_ctx_t prot_ctx,
			      IN MOSAL_mem_flags_t flags,
			      IN MT_phys_addr_t pa,
			      IN MT_size_t size,
			      OUT MOSAL_iomem_t * iomemobj_p,
			      OUT MT_virt_addr_t * va_p);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_iomem_unmap
 *
 *  Description: unmap the io memory from prot_ctx's address space
 *
 *  Parameters:
 *    iomemobj(in) object handle
 *    prot_ctx(in) protection context
 *
 *  Returns: MT_OK success
 *
 ******************************************************************************/
call_result_t MOSAL_iomem_unmap(IN MOSAL_iomem_t iomemobj,
				IN MOSAL_prot_ctx_t prot_ctx);

/******************************************************************************
 *
 *  Function(Kernel Space only): MOSAL_mem_iomem_get_props
 *
 *  Description: get the properties of the object
 *
 *  Parameters:
 *    iomemobj(in) object handle
 *    props_p(out) returned properties
 *
 *  Returns: MT_OK success
 *
 ******************************************************************************/
call_result_t MOSAL_iomem_get_props(IN MOSAL_iomem_t iomemobj,
				    OUT MOSAL_iomem_props_t * props_p);

#endif

#endif				/* H_MOSAL_MEM_H */
