/*
  This software is available to you under a choice of one of two
  licenses.  You may choose to be licensed under the terms of the GNU
  General Public License (GPL) Version 2, available at
  <http://www.fsf.org/copyleft/gpl.html>, or the OpenIB.org BSD
  license, available in the LICENSE.TXT file accompanying this
  software.  These details are also available at
  <http://openib.org/license.html>.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.

  Copyright (c) 2004 Mellanox Technologies Ltd.  All rights reserved.
*/

#ifndef H_MOSAL_TIMER_IMP_H
#define H_MOSAL_TIMER_IMP_H

#ifdef MT_KERNEL

#if LINUX_KERNEL_2_4 || LINUX_KERNEL_2_6
/* This code is only for kernel 2.4 or 2.6 */

#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/time.h>
#if defined(__ia64__)
#include <asm/delay.h>
#endif


/*  ISRs */

typedef int MOSAL_IRQ_ID_t;

/* IRQ back compat. */
#ifndef IRQ_HANDLED
typedef void irqreturn_t;
#define IRQ_HANDLED
#endif

typedef struct pt_regs MOSAL_intr_regs_t;
typedef irqreturn_t(*intr_handler_t) (int irq, void *dev_id,
				      MOSAL_intr_regs_t * regs_p);

/*  user ISR function prototype */
typedef irqreturn_t(*MOSAL_ISR_func_t) (MT_ulong_ptr_t func_ctx, void *isr_ctx1,
					void *isr_ctx2);

struct MOSAL_ISR {
	MOSAL_ISR_func_t func;
	MOSAL_IRQ_ID_t irq;
	char *name;
	MT_ulong_ptr_t ctx;
	void *isr_ctx1;
	void *isr_ctx2;
};

/*  DPCs */

/*  types of DPC (if doesn't need to relay info from ISR to DPC - use MOSAL_NO_CTX value) */
typedef enum { MOSAL_NO_CTX, MOSAL_SINGLE_CTX,
	    MOSAL_MULTIPLE_CTX } MOSAL_DPC_type_t;

/*  the structure contains the context, relayed to user DPC  */

struct DPC_CONTEXT {
	MT_ulong_ptr_t func_ctx;	/*  DPC ("static") context */
	void *isr_ctx1;		/*  "dynamic" context, relayed by ISR */
	void *isr_ctx2;		/*  "dynamic" context, relayed by ISR */
};

/*  user DPC function prototype */
typedef void (*MOSAL_DPC_func_t) (struct DPC_CONTEXT * func_ctx);

/*  MOSAL DPC object */
struct MOSAL_DPC {
	struct tasklet_struct tasklet;
	MOSAL_DPC_type_t type;	/*  type of DPC */
	MOSAL_DPC_func_t func;	/*  user DPC to be called */
	struct DPC_CONTEXT dpc_ctx;	/*  user DPC context */
	MOSAL_spinlock_t lock;	/*  spinlock */
};

/*  Macros  */
#define MOSAL_DPC_enable(MOSAL_DPC_p)  	tasklet_enable(&MOSAL_DPC_p->tasklet)
#define MOSAL_DPC_disable(MOSAL_DPC_p)  tasklet_disable(&MOSAL_DPC_p->tasklet)

/*  TIMERs */

/*  MOSAL timer object */
struct MOSAL_timer {
	struct timer_list timer;	/*  OS DPC object */
};

#endif				/* LINUX_KERNEL_2_4 */

#endif				/* MT_KERNEL */

#if defined(powerpc) || defined(__powerpc__)
static __inline__ unsigned int mftbu(void)
{
	unsigned int re;
	asm volatile ("mftbu %0":"=r" (re));
	return re;
}
static __inline__ unsigned int mftb(void)
{
	unsigned int re;
	asm volatile ("mftb %0":"=r" (re));
	return re;
}

static __inline__ u_int64_t MOSAL_ppc_cputime_get(void)
{

	u_int64_t upper, lower, temp;
	do {
		upper = mftbu();
		lower = mftb();
		temp = mftbu();	/*  (?? bypass Possible bug from ~0x0 to 0x0) */
	} while (temp != upper);
	return (upper << 32) | lower;

}				/*  gilboa_cputime_get */

#endif				/* __powerpc__ */

#ifdef MT_KERNEL

#if (LINUX_KERNEL_2_4 || LINUX_KERNEL_2_6)
/* This code is only for kernel 2.4 or kernel 2.6 */

/* ////////////////////////////////////////////////////////////////////////////// */

/*  DPC (=tasklet) functions  */

/* ////////////////////////////////////////////////////////////////////////////// */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Component: 
 *		MOSAL tasklet/DPC implementation [tasklets and DPC are synonyms]
 *
 * Data object: 
 * 		MOSAL DPC object	- struct MOSAL_DPC,	wrapping tasklet in Linux and DPC in Windows
 *
 * Functions:
 *		void MOSAL_DPC_init(MOSAL_DPC_t  *d ...)			- initialize a MOSAL DPC object;
 *		void MOSAL_DPC_schedule(MOSAL_DPC_t *d)				- schedule a MOSAL DPC object;
 *		void MOSAL_DPC_schedule_ctx(MOSAL_DPC_t *d)			- schedule a MOSAL DPC object with parameters;
 *		call_result_t MOSAL_DPC_add_ctx(MOSAL_DPC_t *d,...) -  add DPC request context to a MOSAL DPC object;
 *		
 *
 * Macros:
 *		MOSAL_DPC_enable(MOSAL_DPC_t *d)					- enable DPC (only Linux)
 *		MOSAL_DPC_disable(MOSAL_DPC_t *d)					- disable DPC (only Linux)
 *
 * Example of usage (taken from CM\msgdspch.c and ported to MOSAL API):
 *
 *		**  declare a MOSAL DPC object 'cm_machine_tasklet', calling function 'cm_machine' 
 *		MOSAL_DPC_t cm_machine_tasklet;
 *		MOSAL_DPC_init (&cm_machine_tasklet, cm_machine, 0, 0);
 *
 *		**  schedule it **
 *		MOSAL_DPC_schedule( &cm_machine_tasklet );
 * 
 * Notes:
 *		1. There are no static initialization like Linux's DECLARE_TASKLET() !
 *		2. DPC function has Linux's prototype: void a_DPC_function(unsigned long ), 
 *		   but the meaning of the parameters depends on the type of DPC.
 *		3. There are 2 types of DPC: MOSAL_SINGLE_CTX and MOSAL_MULTIPLE_CTX.
 *      4. DPC of MOSAL_SINGLE_CTX type:
 *			- allows only one interrupt per one DPC invocation;
 *			- doesn't relay dynamic parameters from ISR to DPC;
 *			- The parameter of a_DPC_function() is in fact a user callback context;
 *		5. DPC of MULITPLE_CTX type:
 *			- allows several interrupts before and during a DPC invocation;
 *			- on every interrupt ISR fills a DPC context with static and dynamic parameters and
 *			  enqueues it to the DPC context chain. The DPC handles all the chain while it works.
 *			- The parameter of a_DPC_function() is in fact a pointer to DPC_CONTEXT_t (see below);
 *		   
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

struct DPC_CONTEXT;
typedef struct DPC_CONTEXT DPC_CONTEXT_t;

/*  typedef for MOSAL DPC object */
struct MOSAL_DPC;
typedef struct MOSAL_DPC MOSAL_DPC_t;

/*  Functions */

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_DPC_init
 *
 *  Description:
 *		init a MOSAL DPC object
 *
 *  Parameters:
 *		d			- MOSAL DPC object
 *		func		- user DPC;
 *		data		- its data;
 *		type 		- type of DPC
 *
 *  Returns:
 *
 *  Notes:
 *		(Windows) Callers of this routine must be running at IRQL PASSIVE_LEVEL
 *
 ******************************************************************************/
void MOSAL_DPC_init(MOSAL_DPC_t * d, MOSAL_DPC_func_t func,
		    MT_ulong_ptr_t func_ctx, MOSAL_DPC_type_t type);

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_DPC_schedule
 *
 *  Description:
 *		schedule user DPC 
 *
 *  Parameters:
 *		d			- MOSAL DPC object
 *
 *  Returns:
 *
 *  Notes:
 *		(Windows) Callers of this routine must be running at IRQL PASSIVE_LEVEL
 *
 ******************************************************************************/
MT_bool MOSAL_DPC_schedule(MOSAL_DPC_t * d);

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_DPC_schedule_ctx
 *
 *  Description:
 *		schedule user DPC with relaying a context
 *
 *  Parameters:
 *		d				- MOSAL DPC object
 *		isr_ctx1		- context, relayed by ISR, inserting DPC;
 *		isr_ctx2		- context, relayed by ISR, inserting DPC; 
 *
 *  Returns:
 *
 *  Notes:
 *		(Windows) Callers of this routine must be running at IRQL PASSIVE_LEVEL
 *
 ******************************************************************************/
MT_bool MOSAL_DPC_schedule_ctx(MOSAL_DPC_t * d, void *isr_ctx1, void *isr_ctx2);

#ifdef SUPPORT_MULTIPLE_CTX
 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_DPC_add_ctx
 *
 *  Description:
 *		add DPC request context to a MOSAL DPC object
 *
 *  Parameters:
 *		d			- MOSAL DPC object
 *		ctx1		- context, relayed by ISR, inserting DPC;
 *		ctx2		- context, relayed by ISR, inserting DPC; 
 *
 *  Returns:
 *		MT_ENORSC	- if no ctx structures
 *		MT_OK		- otherwise
 *
 *  Notes:
 *		A helper routine for ISR, inserting DPC 
 *
 ******************************************************************************/
call_result_t MOSAL_DPC_add_ctx(MOSAL_DPC_t * d, PVOID ctx1, PVOID ctx2);
#  endif			/* SUPPORT_MULTIPLE_CTX */

/* ////////////////////////////////////////////////////////////////////////////// */

/*  Timer functions  */

/* ////////////////////////////////////////////////////////////////////////////// */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Component: 
 *		MOSAL timer functions implementation 
 *
 * Data object: 
 * 		MOSAL timer object	- struct MOSAL_timer, wrapping OS-dependent structures
 *
 * Functions:
 *		void MOSAL_timer_init(MOSAL_timer_t *t)		- initialize a MOSAL timer object;
 *		void MOSAL_timer_add(MOSAL_timer_t *t ...)	- start timer;
 *		void MOSAL_timer_del(MOSAL_timer_t *t ...)	- cancel timer;
 *		void MOSAL_timer_mod(MOSAL_timer_t *t ...)	- restart timer;
 *
 * Example of usage (taken from CMkernel\cm_sm.c and ported to MOSAL API):
 *
 *		**  declare a MOSAL timer object 'cm_machine_tasklet', calling function 'cm_machine' **
 *		MOSAL_timer_t try_timer;
 *		MOSAL_timer_init (&try_timer);
 *
 *		**  start timer **
 *		MOSAL_timer_add( &try_timer, cmk_try_timeout, connection->local_com_id, usec );
 * 
 * Portability notes:
 *		1. Functions MOSAL_timer_del() and MOSAL_timer_mod() return 'void' (and not 'int' as in Linux);
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*  timer  object */
struct MOSAL_timer;
typedef struct MOSAL_timer MOSAL_timer_t;

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_timer_init
 *
 *  Description:
 *		init a MOSAL DPC object
 *
 *  Parameters:
 *		d			- MOSAL timer object
 *		func		- user DPC;
 *		data		- its data;
 *
 *  Returns:
 *
 *  Notes:
 *		1. Callers of this routine must be running at IRQL PASSIVE_LEVEL;
 *		2. Every timer must use its own MOSAL timer object ! 
 *		3. Different MOSAL timer objects may use the same DPC function;
 *
 ******************************************************************************/
__INLINE__ void MOSAL_timer_init(MOSAL_timer_t * t);

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_timer_add
 *
 *  Description:
 *		start timer
 *
 *  Parameters:
 *		d			- MOSAL timer object
 *		func		- user DPC;
 *		data		- its data;
 *		usecs		- interval; 'func' will be called in 'usecs' microseconds
 *
 *  Returns:
 *
 *  Notes:
 *		Callers of this routine must be running at IRQL <= DISPATCH_LEVEL
 *
 ******************************************************************************/
__INLINE__ void MOSAL_timer_add(MOSAL_timer_t * t, MOSAL_DPC_func_t func,
				MT_ulong_ptr_t data, long usecs);

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_timer_del
 *
 *  Description:
 *		delete timer
 *
 *  Parameters:
 *		d			- MOSAL timer object
 *
 *  Returns:
 *
 *  Notes:
 *		Callers of this routine must be running at IRQL <= DISPATCH_LEVEL 
 *    NOTE for Linux - *** do not use the callback function to call add_timer to
 *                     add this timer to the list as it may cause a race condition ***
 *
 ******************************************************************************/
__INLINE__ void MOSAL_timer_del(MOSAL_timer_t * t);

 /******************************************************************************
 *  Function (kernel-mode only):
 *		MOSAL_timer_mod
 *
 *  Description:
 *		stop the running timer and restart it in 'usecs' microseconds
 *
 *  Parameters:
 *		d			- MOSAL timer object
 *		usecs		- interval; 'func' will be called in 'usecs' microseconds
 *
 *  Returns:
 *
 *  Notes:
 *		Callers of this routine must be running at IRQL <= DISPATCH_LEVEL
 *
 ******************************************************************************/
__INLINE__ void MOSAL_timer_mod(MOSAL_timer_t * t, long usecs);

#endif				/* (LINUX_KERNEL_2_4 || LINUX_KERNEL_2_6) */
#endif				/* MT_KERNEL */

#if defined(__i386__) || defined(__x86_64__)
#define 	MT_MOSAL_GET_TIME_COUNTER		\
  u_int32_t low, high;	\
  __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high)); \
  return (((u_int64_t)high)<<32) | (u_int64_t)low

#elif defined(__ia64__) && defined(MT_KERNEL)
  #define MT_MOSAL_GET_TIME_COUNTER return (u_int64_t)ia64_get_itc()
#else
#define 	MT_MOSAL_GET_TIME_COUNTER		\
  return 0
#endif

typedef struct timespec MOSAL_timespec_t;

#endif				/* H_MOSAL_TIMER_IMP_H */
