#!/bin/sh  
# uninstall.sh - uninstallation script for vapi package


if [ $UID != 0 ]; then
  echo Must be root to run this script.
  exit 1
fi

name=`basename $0`
cd `dirname $0`
package_dir=`pwd`

LOGFILE=/dev/null

usage() {
    cat >&2 <<EOF
Usage: $name [--rel-file <path>] [--force] 
      --rel-file   Remove vapi-linux installation based on info in given file 
                   (default: /etc/vapi-linux-release)
      --force      Don't prompt for each file removal 
  -h, --help
EOF
}

prompt_removal() {
  local target_path=$1
  local backup_path=$2
  local remove_approved
  if [ -e $target_path ]; then

    if [ "x$FORCE" == "xforce" ]; then 
      remove_approved="y" 
    else
      read -p "Remove $target_path [N,y] ? " remove_approved
    fi

    if [ "x$remove_approved" == "xy" ] || [ "x$remove_approved" == "xY" ]; then
      if [ "x$backup_path" == "x" ]; then
        echo "Removing $target_path" >> $LOGFILE 2>&1
        rm -rf $target_path >> $LOGFILE 2>&1
      else
        echo "Saving $target_path at $backup_path" >> $LOGFILE 2>&1
        mv -f $target_path $backup_path  >> $LOGFILE 2>&1
      fi
    else
      echo "Cannot continue without removing $target_path first."
      exit 1
    fi

  fi
}

TEMP=`getopt -o h --long rel-file:,force,help \
    -n 'uninstall.sh' -- "$@"`

if [ $? != 0 ]; then usage ; exit 1; fi

eval set -- "$TEMP"

REL_FILE=/etc/vapi-linux-release

while true; do
    case "$1" in
           --rel-file)       REL_FILE=$2; shift 2;;
           --force)          FORCE="force"; shift;;
        -h|--help)           usage; exit 0;;
        --) shift; break;;
    esac
done

if [ ! -f $REL_FILE ]; then
  echo "Release file \"$REL_FILE\" not found."
  exit 1
fi
# Get installation info from REL_FILE
source $REL_FILE

KER_REV=$KERNELRELEASE

# Report given parameters
echo "Uninstalling VAPI for Linux: BUILD_ID=`cat ./BUILD_ID`"
echo "prefix=$prefix" 
echo "MTHOME=$MTHOME"
echo "src_path=$src_path"
echo "KER_REV=$KER_REV"
#

echo "Unloading VAPI modules..."
/sbin/modprobe -r mod_ib_mgt
/sbin/modprobe -r mod_thh
/sbin/modprobe -r mod_rhh
# In case module removal failed...
rm -f /dev/ib_mgt /dev/vipkl /dev/mosal

echo "Cleaning source tree at $src_path ..."
prompt_removal ${src_path}

# Remove H files
echo "Removing H files under ${prefix}/include..."
EXPORTED_HEADERS="mtl_common.h ib_defs.h mtl_types.h bit_ops.h mtl_sys_defs.h mtl_sys_types.h mtl_x86_types.h mtl_x86_64_types.h mtl_ia64_types.h mtl_errno.h mtl_errno_priv.h mtl_log.h memtrack.h mtperf.h mosal_dll.h  mosalu_socket.h  mosalu_socket_imp.h mt_getopt.h mt_getopt_imp.h vapi.h evapi.h vapi_features.h vapi_types.h vapi_common.h"
for hfile in $EXPORTED_HEADERS ; do
  rm -f ${prefix}/include/$hfile
done

echo "Removing binaries under ${prefix}/bin ..."
# Remove binaries
exes="open_hcas close_hcas vstat vping rctp perf_main check_catastrophic minism"
for exe in $exes; do
  rm -f ${prefix}/bin/$exe
done

echo "Removing libraries under ${prefix}/lib and under ${prefix}/lib64 ..."
# Remove libs
libs="libmosal.so libmtl_common.so libmpga.so libvapi.so libthhul.so librhhul.so libib_mgt.so"
for lib in $libs; do
  rm -f ${prefix}/lib/$lib
  rm -f ${prefix}/lib64/$lib
done
echo "Updating LD cache (ldconfig)..."
/sbin/ldconfig

# Cleaning modprobe|modules.conf
if [ "`echo $KER_REV | cut -c 1-3`" == "2.4" ]; then
  echo "Cleaning modules.conf..."
  MODULES_CONF_LINE="include /etc/modules.conf.vapi"
  cp /etc/modules.conf /etc/modules.conf.bak
  grep -v "$MODULES_CONF_LINE" /etc/modules.conf.bak > /etc/modules.conf
  rm -f /etc/modules.conf.vapi

else # kernel 2.6 - use modprobe.conf
  echo "Cleaning modprobe.conf.vapi"
  MODULES_CONF_LINE="include /etc/modprobe.conf.vapi"
  cp /etc/modprobe.conf /etc/modprobe.conf.bak
  grep -v "$MODULES_CONF_LINE" /etc/modprobe.conf.bak > /etc/modprobe.conf
  rm -f /etc/modprobe.conf.vapi > /dev/null 2>&1
fi

# Remove modules
mod_root=/lib/modules/$KER_REV
echo "Removing kernel modules under $mod_root ..."
modules="mlxsys mod_vip mod_thh mod_rhh mod_ib_mgt"
for module in $modules; do
  mod_path=`find $mod_root/kernel -regex ".*$module\.k?o" 2>/dev/null` 
  if [ $? == 0 ]; then rm -f $mod_path; fi
  mod_path=`find $mod_root/extra -regex ".*$module\.k?o" 2>/dev/null` 
  if [ $? == 0 ]; then rm -f $mod_path; fi
done
# Remove directories if empty
rmdir $mod_root/kernel/drivers/infiniband > /dev/null 2>&1
rmdir $mod_root/extra > /dev/null 2>&1

echo "Running /sbin/depmod"
/sbin/depmod
if [ $? != 0 ]; then
  echo "Failed depmod."
fi

rm -f /etc/vapi-linux-release > /dev/null 2>&1

echo "VAPI-linux was uninstalled."

rm -rf ${MTHOME}
exit 0
