/*
 *  i2c.h - i2c driver header file
 *
 *
 *  Copyright (C) Mellanox Techniologies Ltd.
 *
 *  Creation date: 30-5-2000
 *
 *  Description: header file for i2c driver. Te card is base on the PCF8584
 *               I2C bus controller.
 *
 *  Changes:
 */


#ifndef H_I2C_H
#define H_I2C_H

#include <mtcr_types.h>



/*============== type definitions =========================*/

/* define all supported serial clock rates */
typedef enum {
  clk_rate_90k,
  clk_rate_45k,
  clk_rate_11k,
  clk_rate_1k5
}
clk_rate_t;


/*============= macro definitions =========================*/

#define __NR_i2c_syscall 31 /* the system call which will be substituted by our system call */

#define MT_EI2CNOTRAN  -1402 /* remove when updated in cvs !!! */


/* for each operation code there is an associated struct that defines the data that should
   be passed which is relevant to the required operation */

/* operation codes passed to the susyem call */
#define I2C_OP_INIT 0
#define I2C_OP_MASTER_TRANSMIT 1
#define I2C_OP_MASTER_RECEIVE  2
#define I2C_OP_SEND_STOP       3




/* struct definition for I2C_OP_INIT */
typedef struct
{
  call_result_t ret_val; /* return value */
  u_int16_t base; /* base address of card */
  u_int8_t  own; /* own address of the card */
  clk_rate_t sclk;   /* the required serial clock rate */
}
i2c_op_init_st;


/* struct definition for I2C_OP_MASTER_TRANSMIT */
typedef struct
{
  call_result_t ret_val; /* return value */
  u_int8_t slv_addr;
  void *buffer;
  u_int32_t count;
  u_int32_t *bytes_sent_p;
  bool sendSTOP;
  u_int32_t *key_p;
}
i2c_op_master_transmit_st;


/* struct definition for I2C_OP_MASTER_RECEIVE */
typedef struct
{
  call_result_t ret_val; /* return value */
  u_int8_t slv_addr;
  void *buffer;
  u_int32_t count;
  u_int32_t *bytes_received_p;
  bool sendSTOP;
  u_int32_t key;
}
i2c_op_master_receive_st;


/* struct definition for I2C_OP_SEND_STOP */
typedef struct
{
  call_result_t ret_val; /* return value */
  u_int32_t key;
}
i2c_op_send_stop_st;






/* functions prototypes */


/***************************************************************************************************
 * Function: i2c_setup
 *
 * Description: sets up and initializes the card
 *
 * Parameters: base(IN): the base address of the card in the io space
 *             own (IN): the own address of the card in the i2c bus
 *             sclck(IN):the required serial clock rate
 *
 * Returns: MT_OK    success
 *          MT_ERROR failure to initialize
 *
 * Notes:
 *
 **************************************************************************************************/
call_result_t i2c_setup(u_int16_t base, u_int8_t own, clk_rate_t sclk);





/***************************************************************************************************
 * Function: i2c_master_transmit
 *
 * Description: master transmit data
 *
 * Parameters: slv_addr(IN): the address of the receipent
 *             buffer (IN): pointer to data to be transmitted
 *             count(IN): number of bytes to be sent
 *             bytes_sent_p(OUT): pointer to var to receive number of bytes sent
 *             sendSTOP(IN): flag when 1 send STOP - otherwise don't send STOP
 *             key_p(IN/OUT): pointer to key. If the bus is not locked the caller must set
 *                            it to zero, and upon return should save the value for future call
 *                            to i2c_master_receive.
 *
 * Returns: MT_OK    success
 *          MT_ERROR failure to initialize
 *
 * Notes:
 *
 **************************************************************************************************/
call_result_t i2c_master_transmit(unsigned char slv_addr, void *buffer, int count, int *bytes_sent_p,
                                  bool sendSTOP, u_int32_t *const key_p);






/***************************************************************************************************
 * Function: i2c_master_receive
 *
 * Description: master receive data
 *
 * Parameters: slv_addr(IN): the address of the receipent
 *             buffer (OUT): pointer where to put data received
 *             count(IN): number of bytes to be receive
 *             bytes_received_p(OUT): pointer to var to receive number of bytes received
 *             key(IN): key to autenticate the caller.
 *
 * Returns: MT_OK    success
 *          MT_ERROR failure to initialize
 *
 * Notes:
 *
 **************************************************************************************************/
call_result_t i2c_master_receive(u_int8_t slv_addr, void *buffer, u_int16_t count,
                                 u_int32_t *bytes_received_p, bool sendSTOP,
                                 const u_int32_t key);




/***************************************************************************************************
 * Function: i2c_send_stop
 *
 * Description: send STOP to the i2c bus
 *
 * Parameters: key(IN): pointer to a key previously received from master transmit
 *             key(IN): key to autenticate the caller.
 *
 * Returns: MT_OK    success
 *          MT_ERROR failure to initialize
 *
 * Notes:
 *
 **************************************************************************************************/
call_result_t i2c_send_stop(const u_int32_t key);



/***************************************************************************************************
 * Function: i2c_present
 *
 * Description: check if the module is preperly installed
 *
 * Parameters:
 *
 * Returns: TRUE - module properly installed
 *          FALSE - module was not properly installed
 *
 * Notes:
 *
 **************************************************************************************************/
bool i2c_present(void);


#endif  /* H_I2C_H */
