/*
 *
 *  minit.c - CR Space access init
 *
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/pci.h>
#include <fcntl.h>
#include <unistd.h>

#include "mst_pci.h"
#include "mst_pciconf.h"
#include "mst_calbr.h"
#ifdef MST_CPU_ppc_8xx
  #include "mst_ppc.h"
#endif

#define CLEAR(st) memset(&(st), 0, sizeof(st))

void usage(const char *n)
{

    printf("\nFor PCI devices:\n\t%s <device> <busdevfn> <bar#>\n\n",n);
    printf("For PCICONF devices:\n\t%s <device> <busdevfn> <addr_reg> <data_reg>\n\n",n);
    printf("For CALIBRE\n\t%s <device> <base> <own> <clk>\n\n",n);
#ifdef MST_CPU_ppc_8xx
    printf("For PPC    \n\t%s <device> <cs>\n\n",n);
#endif
    exit(1);
}

char *printpci(unsigned int bus, unsigned int devfn)
{
    static char buf[10];
    sprintf(buf, "%02x:%02x.%1x", bus, devfn >> 3, devfn & 0x7);
    return buf;
}

void getpci(char *str, unsigned int *bus, unsigned int *devfn)
{
    unsigned int dev, fn;
    char         *p1 = strchr(str, ':');
    char         *p2 = strchr(str, '.');

    if (!p1 || !p2)
    {
        printf("Invalid \"busdevfn\" specification. Should be XX:XX.X\n");
        exit(1);
    }
    *p1++ = '\0';
    *p2++ = '\0';
    *bus = strtol(str, NULL, 16);
    dev = strtol(p1, NULL, 16);
    fn = strtol(p2, NULL, 16);
    *devfn = PCI_DEVFN(dev, fn);
}

int main(int ac, char *av[])
{
    char          *devname;
    int           rc = 0, fd = 0;

    if (ac < 2)
        usage(av[0]);

    devname = av[1];
    if (strstr(devname, "pciconf"))
    {
        /* PCICONF initialization */
        struct mst_pciconf_init_st pciconf_init;
        int                        use_as_i2cm = ! !strstr(devname,"i2cm");
        CLEAR(pciconf_init);

        if (ac < 5)
            usage(av[0]);
        if ((fd = open(devname, O_RDWR | O_SYNC)) < 0)
        {
            perror("open");
            return 1;
        }
        getpci(av[2], &pciconf_init.bus, &pciconf_init.devfn);
        pciconf_init.addr_reg = strtol(av[3], NULL, 0);
        pciconf_init.data_reg = strtol(av[4], NULL, 0);
        rc = ioctl(fd, PCICONF_INIT, &pciconf_init);
        if (rc >= 0)
        {
            char buf[256];
            sprintf(buf, "%-32s - PCI configuration cycles access%s.\n"
                    "%34s bus:dev.fn=%s addr.reg=%d data.reg=%d\n",
                    devname, use_as_i2cm?" (use as I2C master only) ":"", " ",
                    printpci(pciconf_init.bus, pciconf_init.devfn),
                    pciconf_init.addr_reg, pciconf_init.data_reg);
            rc = (write(fd, buf, strlen(buf)) != strlen(buf));
            if (rc)
                perror("write");
        }
        else
            perror("ioctl (PCICONF_INIT)");
    }
    else if (strstr(devname, "pci"))
    {
        /* PCI initialization */
        struct mst_pci_init_st pci_init;
        int                    use_as_i2cm = ! !strstr(devname,"i2cm");
        CLEAR(pci_init);

        if (ac < 4)
            usage(av[0]);
        if ((fd = open(devname, O_RDWR | O_SYNC)) < 0)
        {
            perror("open");
            return 1;
        }
        getpci(av[2], &pci_init.bus, &pci_init.devfn);
        pci_init.bar = strtol(av[3], NULL, 0);
        rc = ioctl(fd, PCI_INIT, &pci_init);
        if (rc >= 0)
        {
            struct mst_pci_params_st pars;
            CLEAR(pars);
            rc = ioctl(fd, PCI_PARAMS, &pars);
            if (rc >= 0)
            {
                char buf[256];
                sprintf(buf, "%-32s - PCI direct access%s.\n"
                        "%34s bus:dev.fn=%s bar=0x%08llx size=0x%llx\n",
                        devname, use_as_i2cm?" (use as I2C master only) ":"", " ",
                        printpci(pci_init.bus, pci_init.devfn),
                        pars.bar, pars.size);
                rc = (write(fd, buf, strlen(buf)) != strlen(buf));
                if (rc)
                    perror("write");
            }
            else
                perror("ioctl (PCI_PARAMS)");
        }
        else
            perror("ioctl (PCI_INIT)");
    }
    else if (strstr(devname, "calbr") || strstr(devname, "calibre"))
    {
        /* Calibre initialization */
        struct mst_calbr_init_st calbr_init;
        CLEAR(calbr_init);

        if (ac < 5)
            usage(av[0]);
        if ((fd = open(devname, O_RDWR | O_SYNC)) < 0)
        {
            perror("open");
            return 1;
        }
        calbr_init.base = strtol(av[2], NULL, 0);
        calbr_init.own  = strtol(av[3], NULL, 0);
        calbr_init.clk_rate  = (calbr_i2c_clk_rate_t)strtol(av[4], NULL, 0);
        rc = ioctl(fd, CALBR_INIT, &calbr_init);
        if (rc >= 0)
        {
            char buf[256];
            sprintf(buf, "%-32s - Calibre board as I2C master\n"
                    "%34s base:0x%x own:0x%x clk_rate_idx:%d\n",
                    devname, " ",
                    calbr_init.base, calbr_init.own, calbr_init.clk_rate);
            write(fd, buf, strlen(buf));
        }
    }
    else if (strstr(devname, "vtop"))
    {
        char buf[256];

        if ((fd = open(devname, O_RDWR | O_SYNC)) < 0)
        {
            perror("open");
            return 1;
        }
        sprintf(buf, "%-32s - Virtual to physical addresses translation driver\n",
                devname);
        write(fd, buf, strlen(buf));
    }
    else if (strstr(devname, "usb"))
    {
      /* Nothing has to be done about usb devices */
    }
#ifdef MST_CPU_ppc_8xx
    else if (strstr(devname, "ppc"))
    {
		struct mst_ppc_init_st ppc_init;
        if (ac < 3)
            usage(av[0]);
        if ((fd = open(devname, O_RDWR | O_SYNC)) < 0)
        {
            perror("open");
            return 1;
        }
        ppc_init.cs = strtol(av[2], NULL, 0);
        ppc_init.size = strstr(devname, "47396") ? 0x200000 : 0x10000;
        rc = ioctl(fd, PPC_INIT, &ppc_init);
        if (rc >= 0)
        {
            struct mst_ppc_params_st pars;
            CLEAR(pars);
            rc = ioctl(fd, PPC_PARAMS, &pars);
            if (rc >= 0)
            {
                char buf[256];
                sprintf(buf, "%-32s - PPC direct access.\n"
                        "%34s cs=%d addr=0x%08x size=0x%x\n",
                        devname, " ", pars.cs, pars.addr, pars.size);
                rc = (write(fd, buf, strlen(buf)) != strlen(buf));
                if (rc)
                    perror("write");
            }
            else
                perror("ioctl (PCI_PARAMS)");
        }
        else
            perror("ioctl (PCI_INIT)");
    }
#endif
    else
    {
        printf("Invalid device name. Use that:\n");
        usage(av[0]);
    }
    return rc;
}
