/*
 *  i2c.c - implementation of an i2c driver for the Calibre ISA card
 *
 *  Copyright (C) Mellanox Techniologies Ltd.
 *
 *  Creation date: 30-5-2000
 *
 *  Description: This file contains the implementation of a driver for the Calibre ISA card.
 *               It uses two IO addresses, one for control and one for data.
 *
 *  Changes:
 *
 */



/* #define __SMP__   */ /* put in comment if compiling in single CPU machine */

#include <linux/module.h>
/* #include <linux/modversions.h> */
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <mtcr_types.h>
#include <i2c.h>
#include <linux/init.h>
#include <linux/version.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a, b, c) ((a)*65536+(b)*256+(c))
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/jiffies.h>
#endif

#define MTL_ERROR3 printk

/* Deal with CONFIG_MODVERSIONS */
/* #if CONFIG_MODVERSIONS==1 */
/* #define MODVERSIONS */
/* #include <linux/modversions.h> */
/* #endif */

/*========= macro definitions ====================================================*/

#define I2C_LOCK_TIMEOUT_SEC 1

#define I2C_TIMEOUT_JIFFIES (I2C_LOCK_TIMEOUT_SEC * HZ)


/* bits definitions in control register */
#define I2C_PIN  0x80
#define ES0  0x40
#define ES1  0x20
#define ES2  0x10
#define ENI  0x08
#define STA  0x04
#define STO  0x02
#define ACK  0x01

/* bits definitions in status register */
#define I2C_PIN  0x80
#define STS  0x20
#define BER  0x10
#define AD0  0x08
#define LRB  0x08
#define AAS  0x04
#define LAB  0x02
#define BB_  0x01


/* allowable internal clock frequencies */
#define CLK_3M    0x00
#define CLK_4M33  0x10
#define CLK_6M    0x14
#define CLK_8M    0x18
#define CLK_12M   0x1c

#define CARD_I_CLK CLK_12M   /* the card uses 12Mhz internal clock frequency */

/* serial clock frequencies */
#define SCLK_90K    0x00
#define SCLK_45K    0x01
#define SCLK_11K    0x02
#define SCLK_1K5    0x03


#define MAX_DELAY_ITER 20000 /* max number of delay iterations in functions that poll */

MODULE_AUTHOR("Eli Cohen - Mellanox Technologies LTD");
MODULE_DESCRIPTION("I2C driver for Calibre ISA I2C card");
MODULE_SUPPORTED_DEVICE("ICA93LV");


/*========= exported symbols =====================================================*/
EXPORT_SYMBOL(i2c_setup);
EXPORT_SYMBOL(i2c_master_transmit);
EXPORT_SYMBOL(i2c_master_receive);
EXPORT_SYMBOL(i2c_send_stop);
EXPORT_SYMBOL(i2c_present);


/*========= static variables declarations ========================================*/
/* indicates whether the module has been initialized and is ready to e used */
static bool module_opened = FALSE;

static u_int16_t card_base_address; /* base address of the card in io space */
static u_int16_t card_conf_address; /* the address of the control/status register */
static u_int8_t card_i2c_address; /* the I2C own address of the card */
static clk_rate_t card_ser_clock; /* the serial clock rate selected for the device */

/* determines whether to insert a system call to interface with teh driver */
static bool use_syscall;

/* a key used to allow access to the I2C bus when in a middle of a transaction
   requiring repeated start. */
static u_int32_t access_key;

/* flag that when set indicates that there is a transaction going on */
static bool transaction_on;

/* the last time when the bus was locked */
static u_int32_t lock_time;



/* define macros that allow to pass parameters from insmod */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(card_base_address, "i");
MODULE_PARM(card_i2c_address, "i");
MODULE_PARM(card_ser_clock, "i");
MODULE_PARM(use_syscall, "i");
#endif
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


/* pointer to hold original handler for the system call number we use */
#ifdef SYSCALL_SUPPORT
static void *sys_old;
#endif

/*========= static functions prototypes ==========================================*/
static call_result_t wait_bus_free(u_int32_t count);
static call_result_t wait_PIN_low(u_int32_t count);
static u_int8_t get_status(void);
static void short_delay(void);
#ifdef SYSCALL_SUPPORT
static asmlinkage int i2c_syscall(u_int32_t command, void *param_p);
#endif
static u_int32_t new_key(void);
void check_timeout_and_release(void);




/*========= external definitions =================================================*/
extern void *sys_call_table[]; /* the system call handlers table */



#ifndef BINDKERNEL

static call_result_t __init init_ica93lv_module(void)
{
    char s[20];
    call_result_t ret = MT_OK;

    if ( (card_base_address == 0) |
            (card_i2c_address == 0) |
            ( card_ser_clock > clk_rate_90k ) ) {
        return MT_OK; /* we let the module to be loaded but all the functions will fail
                         because 'module_started' is FALSE */
    }
    else {
        switch ( card_ser_clock ) {
            case clk_rate_90k:
                strcpy(s, "90 kbps");
                break;

            case clk_rate_45k:
                strcpy(s, "45 kbps");
                break;

            case clk_rate_11k:
                strcpy(s, "11 kbps");
                break;

            case clk_rate_1k5:
                strcpy(s, "1.5 kbps");
                break;

            default:
                strcpy(s, "invalid serial clock rate");
                return MT_EINVAL;
        }
    }


    if ( use_syscall ) {
        /* only use syscalls if reqested */
        /* when the module is intended to be used by kernel code the it is
           not necessary to insert a system call */
#ifdef SYSCALL_SUPPORT
        sys_old = sys_call_table[__NR_i2c_syscall];
        sys_call_table[__NR_i2c_syscall] = i2c_syscall;
#endif
    }


    access_key = 0; /* initialize access key */
    transaction_on = FALSE; /* no transacion in progress */
    lock_time = jiffies;


/*    if ( check_region(card_base_address, 2) < 0 ) return MT_ENORSC; */
    if (request_region(card_base_address, 2, "mod_i2c") == NULL) return MT_ENORSC;

    ret = i2c_setup(card_base_address, card_i2c_address, card_ser_clock);

    if(ret != MT_OK)
    {
        release_region(card_base_address, 2);
        card_base_address = 0;
    }

    return MT_OK;
}

static void __exit cleanup_ica93lv_module(void)
{
    if(card_base_address != 0)
        release_region(card_base_address, 2);
#ifdef SYSCALL_SUPPORT
    sys_call_table[__NR_i2c_syscall] = sys_old;
#endif
}

module_init(init_ica93lv_module);
module_exit(cleanup_ica93lv_module);

#endif

/*
 *  i2c_setup()
 */
call_result_t i2c_setup(u_int16_t base, u_int8_t own, clk_rate_t sclk)
{
    u_int8_t S2, data;
    int i;


    /* check that the clock rate is valid */
    switch (sclk) {
        case clk_rate_90k:
            card_ser_clock = clk_rate_90k;
            S2 = SCLK_90K | CARD_I_CLK;
            break;

        case clk_rate_45k:
            card_ser_clock = clk_rate_45k;
            S2 = SCLK_45K | CARD_I_CLK;
            break;


        case clk_rate_11k:
            card_ser_clock = clk_rate_11k;
            S2 = SCLK_11K | CARD_I_CLK;
            break;

        case clk_rate_1k5:
            card_ser_clock = clk_rate_1k5;
            S2 = SCLK_1K5 | CARD_I_CLK;
            break;


        default:
            return MT_EINVAL;
    }


    /* check that the i2c address is valid */
    if ( own & '\x80' ) {
        /* this card supports only 7 bits address */
        return MT_EINVAL;
    } else {
        card_i2c_address = own;
    }


    /* card's addresses in the io space */
    card_base_address = base;
    card_conf_address = card_base_address + 1;


    /* set the cards own address */

    outb_p(I2C_PIN, card_conf_address);
    for (i=0; i<50; ++i) short_delay();
    outb_p(card_i2c_address, card_base_address);
    for (i=0; i<50; ++i) short_delay();

    data = inb_p(card_base_address);
    for (i=0; i<50; ++i) short_delay();
    if ( data != card_i2c_address ) {
        if(card_base_address != 0)
            printk("<1>cannot read back card i2c address\n");
        return MT_ERROR;
    }

    /* set serial clock frequency */
    outb_p(I2C_PIN | ES1, card_conf_address);
    for (i=0; i<50; ++i) short_delay();
    outb_p(S2, card_base_address);
    for (i=0; i<50; ++i) short_delay();

    /* enable the i2c lines */
    outb_p(I2C_PIN | ES0 | ACK, card_conf_address);
    for (i=0; i<50; ++i) short_delay();


    /* mark module as initialized */
    module_opened = TRUE;

    return MT_OK;
}





/*
 *  i2c_master_transmit()
 */
call_result_t i2c_master_transmit(unsigned char slv_addr, void *buffer, int count, int *bytes_sent_p,
        bool sendSTOP, u_int32_t *const key_p)
{
    unsigned char status, data;
    int sent, i;

    if ( module_opened == FALSE ) {
        return MT_ENOINIT;
    }


    if ( slv_addr & 0x80 ) {
        /* invalid slave address */
        check_timeout_and_release();
        MTL_ERROR3("%s(%d): Invalid slave address.\n", __func__, __LINE__);
        return MT_EINVAL;
    }

    if ( key_p == NULL ) {
        return MT_EINVAL;
    }

    if ( transaction_on == FALSE ) {
        /* no transaction in progress */
        if ( *key_p != 0 ) {
            /* tried to continue a transaction when there is no transaction in progress */
            MTL_ERROR3("%s(%d): Invalid outstanding transaction.\n", __func__, __LINE__);
            return MT_EI2CNOTRAN;
        }
    }
    else {
        /* a transaction is already in progress */
        check_timeout_and_release();
        MTL_ERROR3("%s(%d): Device busy.\n", __func__, __LINE__);
        return MT_EBUSY;
    }

    /* wait for the bus to become free */
    if ( wait_bus_free(MAX_DELAY_ITER) == MT_EBUSBUSY ) {
        MTL_ERROR3("%s(%d): Bus busy.\n", __func__, __LINE__);
        return MT_EBUSBUSY;
    }



    /* load slave address to S0 register */
    outb_p( slv_addr << 1, card_base_address);

    /*  Retry first transaction */
    for(i = 0; i < MAX_DELAY_ITER; i++)
    {

        /* generate START and clock out the slave address */
        outb_p(I2C_PIN | ES0 | STA | ACK, card_conf_address);

        if ( wait_PIN_low(MAX_DELAY_ITER) == MT_EI2PINHI ) {
            MTL_ERROR3("%s(%d): Pin signal doens't becomes low.\n", __func__, __LINE__);
            return MT_EI2PINHI;
        }


        status = get_status();

        if ( ! (status & LRB) )
            break;
    }


    for (i=0, sent=0; i<=count; ++i) {

        if ( wait_PIN_low(MAX_DELAY_ITER) == MT_EI2PINHI ) {
            MTL_ERROR3("%s(%d): Pin signal doens't becomes low.\n", __func__, __LINE__);
            return MT_EI2PINHI;
        }


        status = get_status();

        if ( status & LRB ) {
            /* NACK received */

            outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address); /* send STOP */
            *bytes_sent_p = sent;
            transaction_on = FALSE;
            MTL_ERROR3("%s(%d): Nack received after sending %x bytes.\n", __func__, __LINE__, sent);
            return MT_EI2CNACK;
        }

        if ( sent >= count ) {
            break;
        }

        data = *((u_int8_t *)buffer+i);
        outb_p(data, card_base_address);
        sent++;
    }

    /* only send STOP if explicitly requested */
    if ( sendSTOP == TRUE ) {
        outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address); /* send STOP */
    }
    else {
        /* we are not sending STOP => transaction becomes on */
        access_key = new_key(); /* allocate new key */
        *key_p = access_key;
        transaction_on = TRUE;
        lock_time = jiffies; /* record the time the bus was locked */
    }


    *bytes_sent_p = sent;
    return MT_OK;
}


/*
 *  i2c_master_receive()
 */
call_result_t i2c_master_receive(u_int8_t slv_addr, void *buffer, u_int16_t count,
        u_int32_t *bytes_received_p, bool sendSTOP,
        const u_int32_t key)
{
    u_int16_t received;
    u_int8_t status, data;
    bool first = TRUE;

    if ( module_opened == FALSE ){
        return MT_ENOINIT;
    }




    if ( transaction_on == FALSE ) {
        /* no transaction in progress */
        if ( key != 0 ) {
            /* tried to continue a transaction when there is no transaction in progress */
            return MT_EI2CNOTRAN;
        }
        else {

            /* wait for the I2C bus to become free */
            if ( wait_bus_free(MAX_DELAY_ITER) == MT_EBUSBUSY ) {
                /*         outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address);  */ /* send STOP */
                return MT_EBUSBUSY;
            }

            /* make sure the bus is enabled and can address S0 */
            outb_p(I2C_PIN | ES0, card_conf_address);
        }
    }
    else {
        /* a transaction is already in progress */
        if ( key != access_key ) {
            /* key mismatch, request rejected */
            check_timeout_and_release();
            return MT_EBUSY;
        }
        else {
            /* repeated start */
            outb_p(ES0 | STA | ACK, card_conf_address);
        }
    }





    /* load slave adddress + read */
    outb_p((slv_addr << 1) | 0x01, card_base_address);





    /* generate START and clock out the slave address */
    outb_p(I2C_PIN | ES0 | STA | ACK, card_conf_address);

    received = 0;

    while ( 1 ) {

        if ( wait_PIN_low(MAX_DELAY_ITER) == MT_EI2PINHI ) {
            return MT_EI2PINHI;
        }

        status = get_status();

        if ( status & LRB ) {
            /* NACK received */
            outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address); /* send STOP */
            data = inb_p(card_base_address); /* read byte */
            *((char *)buffer+received) = data;
            *bytes_received_p = received;
            return MT_EI2CNACK;
        }

        if ( received == (count-1) ) {
            break;
        }

        data = inb_p(card_base_address);
        if ( first == TRUE ) {
            first = FALSE;
        } else {
            *((char *)buffer+received-1) = data;
        }

        received++;
    }

    /* send NACK to terminate reception */
    outb_p(ES0, card_conf_address);

    data = inb_p(card_base_address);
    *((u_int8_t *)buffer+received-1) = data;

    received++;

    if ( wait_PIN_low(MAX_DELAY_ITER) == MT_EI2PINHI ) {
        return MT_EI2PINHI;
    }


    if ( sendSTOP == TRUE ) {
        /* send STOP */
        outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address);
        transaction_on = FALSE; /* finish transaction */
    }


    data = inb_p(card_base_address);
    *((char *)buffer+received-1) = data;


    *bytes_received_p = received;

    return MT_OK;
}



/*
 *  i2c_send_stop()
 */
call_result_t i2c_send_stop(const u_int32_t key)
{

    if ( module_opened == FALSE ) {
        return MT_ENOINIT;
    }


    if ( key != access_key ) {
        /* no access writes to caller */
        check_timeout_and_release();
        return MT_EBUSY;
    }

    outb_p(I2C_PIN | ES0 | STO | ACK, card_conf_address); /* send STOP */

    transaction_on = FALSE;
    return MT_OK;
}


/*
 * i2c_present()
 *    Since the module can be loaded even if it was not passed the parameters
 *    This function can be used to verify if the module is usable.
 *    It should be noted however that even if the function returns TRUE it does not
 *    imply that the hardware really exists
 */
bool i2c_present(void)
{
    return module_opened;
}





/*===================== static fucntions declarations ==================*/




/***************************************************************************************************
 * Function: wait_bus_free
 *
 * Description: wait for the bus to become free
 *
 * Parameters: count(IN): max number of iterations of short_delay() before the function returns
 *                        MT_EBUSBUSY
 *
 * Returns: MT_OK or MT_EBUSBUSY
 *
 * Notes:
 *
 **************************************************************************************************/
static call_result_t wait_bus_free(u_int32_t count)
{
    u_int8_t status;
    u_int32_t i;
    static u_int32_t busy_time_stamp;
    static bool counting = FALSE;

    status = inb_p(card_conf_address);
    /* wait until bus is free */
    for (i=0; i<count; ++i ) {
        if ( (status & BB_) == BB_ ) {
            return MT_OK;
        }
        short_delay();
        status = inb_p(card_conf_address);
    }


    if ( counting == TRUE ) {
        if ( (jiffies - busy_time_stamp) > (5 * 100) ) {
            i2c_send_stop(access_key);
            counting = FALSE;
        }
    }
    else {
        busy_time_stamp = jiffies;
        counting = TRUE;
    }

    return MT_EBUSBUSY;
}




#ifdef SYSCALL_SUPPORT
/***************************************************************************************************
 * Function: i2c_syscall
 *
 * Description: this function runs in kernel space and identifies the operation code
 *              required by the user
 *
 * Parameters: command(IN): the code of the operation required
 *             param_p(IN/OUT): this is a void pointer whose type depends on the type of operation
 *                              required
 *
 * Returns: MT_OK - success
 *          MT_ERROR - error
 *
 * Notes:
 *
 **************************************************************************************************/
static asmlinkage call_result_t i2c_syscall(u_int32_t command, void *param_p)
{
    i2c_op_init_st init_var;
    i2c_op_master_transmit_st master_transmit_var;
    i2c_op_master_receive_st master_receive_var;
    i2c_op_send_stop_st send_stop_var;
    u_int8_t *k_buf;
    u_int32_t bytes_sent, bytes_received, l_key;
    call_result_t rc;


    switch ( command ) {
        case I2C_OP_INIT:
            /* copy param_p to local variable */
            copy_from_user(&init_var, param_p, sizeof(init_var));

            rc = i2c_setup(init_var.base, init_var.own, init_var.sclk);
            copy_to_user(&((i2c_op_init_st *)param_p)->ret_val, &rc, sizeof(rc));
            return rc;

        case I2C_OP_MASTER_TRANSMIT:
            /* copy param_p to local variable */
            copy_from_user(&master_transmit_var, param_p, sizeof(master_transmit_var));

            /* allocate buffer in kernel space where to put the data to be transmitted */
            k_buf = kmalloc(master_transmit_var.count, GFP_KERNEL);
            if ( k_buf == NULL ) {
                /* failure allocating buffer */
                rc = MT_EKMALLOC;
                copy_to_user(&((i2c_op_master_transmit_st *)param_p)->ret_val, &rc, sizeof(rc));
                return rc;
            }

            /* copy contents of buffer from user to kernel space */
            copy_from_user(k_buf, master_transmit_var.buffer, master_transmit_var.count);

            copy_from_user(&l_key, master_transmit_var.key_p, sizeof(u_int32_t));

            rc = i2c_master_transmit(master_transmit_var.slv_addr,
                    k_buf,
                    master_transmit_var.count,
                    &bytes_sent,
                    master_transmit_var.sendSTOP,
                    &l_key);

            /* free allocated buffer */
            kfree(k_buf);

            /* return to user number of bytes transmitted */
            copy_to_user(((i2c_op_master_transmit_st *)param_p)->bytes_sent_p, &bytes_sent, sizeof(int));
            copy_to_user(&(((i2c_op_master_transmit_st *)param_p)->ret_val), &rc, sizeof(rc));
            copy_to_user(((i2c_op_master_transmit_st *)param_p)->key_p, &l_key, sizeof(u_int32_t));

            return rc;


        case I2C_OP_MASTER_RECEIVE:
            /* copy param_p to local variable */
            copy_from_user(&master_receive_var, param_p, sizeof(master_receive_var));

            /* allocate buffer in kernel space where to put the data read */
            k_buf = kmalloc(master_receive_var.count, GFP_KERNEL);
            if ( k_buf == NULL ) {
                /* failure allocating buffer */
                rc = MT_EKMALLOC;
                copy_to_user(&((i2c_op_master_receive_st *)param_p)->ret_val, &rc, sizeof(rc));
                return rc;
            }


            rc = i2c_master_receive(master_receive_var.slv_addr,
                    k_buf,
                    master_receive_var.count,
                    &bytes_received,
                    master_receive_var.sendSTOP,
                    master_receive_var.key);


            /* copy contents of buffer from user to kernel space */
            copy_to_user(master_receive_var.buffer, k_buf, bytes_received);

            /* free allocated buffer */
            kfree(k_buf);

            /* return to user number of bytes received */
            copy_to_user(master_receive_var.bytes_received_p, &bytes_received, sizeof(int));
            copy_to_user(&((i2c_op_master_receive_st *)param_p)->ret_val, &rc, sizeof(rc));

            return rc;

        case I2C_OP_SEND_STOP:
            copy_from_user(&send_stop_var, param_p, sizeof(send_stop_var));

            rc = i2c_send_stop(send_stop_var.key);
            copy_to_user(&((i2c_op_send_stop_st *)param_p)->ret_val, &rc, sizeof(rc));
            return rc;

        default:
            rc = MT_ERROR;
            copy_to_user(&((i2c_op_send_stop_st *)param_p)->ret_val, &rc, sizeof(rc));
            return rc;
    }
}
#endif



/***************************************************************************************************
 * Function: wait_PIN_low
 *
 * Description: wait for the I2C_PIN (Pending Interrupt Not) bit to low, signifying that the device
 *              needs service due to character transmitted, received etc.
 *
 * Parameters: count(IN): max number of iterations of short_delay() before the function returns
 *                        MT_ERROR
 *
 * Returns: MT_OK or MT_ERROR
 *
 * Notes:
 *
 **************************************************************************************************/
static call_result_t wait_PIN_low(u_int32_t count)
{
    u_int8_t status;
    u_int32_t i;

    status = inb_p(card_conf_address);
    /* wait until I2C_PIN becomes low */
    for (i=0; i<count; ++i ) {
        if ( (status & I2C_PIN) == 0 ) {
            return MT_OK;
        }
        short_delay();
        status = inb_p(card_conf_address);
    }
    return MT_EI2PINHI;
}



/***************************************************************************************************
 * Function: get_status
 *
 * Description: get the status of the I2C
 *
 * Parameters:
 *
 * Returns: the staus byte of the device
 *
 * Notes:
 *
 **************************************************************************************************/
static unsigned char get_status(void)
{
    return inb_p(card_conf_address);
}




/***************************************************************************************************
 * Function: short_delay
 *
 * Description: do a short delay. this is needed to do efficient polling of the chip's
 *              status flags
 *
 * Parameters:
 *
 * Returns:
 *
 * Notes:
 *
 **************************************************************************************************/
static void short_delay(void)
{
    schedule();
}





/***************************************************************************************************
 * Function: new_key
 *
 * Description: return a new, random non zero, 32 bit number to be used as a key
 *
 * Parameters:
 *
 * Returns:
 *
 * Notes:
 *
 **************************************************************************************************/
static u_int32_t new_key(void)
{
    return 0x12345678; /* do real implementation ??????? */
}



/***************************************************************************************************
 * Function: check_timeout_and_release
 *
 * Description: check if the bus is locked (transactio_on) for too much time and release it.
 *
 * Parameters:
 *
 * Returns:
 *
 * Notes:
 *
 **************************************************************************************************/
void check_timeout_and_release(void)
{
    if ( (jiffies - lock_time) >= I2C_TIMEOUT_JIFFIES ) {
        i2c_send_stop(access_key);

        i2c_setup(card_base_address, card_i2c_address, card_ser_clock);
    }
}
