#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <linux/version.h>

#include "mst_pci.h"

//Compatibility for old userspace (we had a bug in PCI_PARAMS
//documentation + the interface was not 32/64 bit compatible)
#define PCI_PARAMS_OLD _IOR(PCI_MAGIC,2,struct mst_pci_init_st)
struct mst_pci_params_old_st {
	unsigned int bar;
	unsigned int size;
};
#define PARAMS_OLD PCI_PARAMS_OLD

#define  INIT   PCI_INIT
#define  STOP   PCI_STOP
#define  PARAMS PCI_PARAMS

static char* name="mst_pci";

#if CONFIG_COMPAT && !HAVE_COMPAT_IOCTL
static unsigned int
ioctl_cmds[] ={
  PARAMS_OLD,
  PARAMS
};
#endif

/* Allow minor numbers 0-255 */
#define MAXMINOR 256
#define BUFFER_SIZE 256

MODULE_AUTHOR("Michael S. Tsirkin - Mellanox Technologies LTD");
MODULE_DESCRIPTION("Mellanox Software Tools driver for accessing Tavor PCI memory");
MODULE_SUPPORTED_DEVICE("MT23108");

struct dev_data {
  unsigned long bar;
  volatile ssize_t size;
  struct semaphore sem;
  char buf[BUFFER_SIZE];
  int bufused;
};

static struct dev_data devices[MAXMINOR];

/* Some buggy 2.4 kernels seem to have trash in pciconf headers */
/* #define OLD_PCI_METHOD */

#ifdef OLD_PCI_METHOD
/*Prevent the user from extending the area with
 * mremap */
static const unsigned int addresses[] = {
  PCI_BASE_ADDRESS_0,
  PCI_BASE_ADDRESS_1,
  PCI_BASE_ADDRESS_2,
  PCI_BASE_ADDRESS_3,
  PCI_BASE_ADDRESS_4,
  PCI_BASE_ADDRESS_5,
  0
};

static unsigned long mst_pci_resource_start(struct pci_dev *dev, int bar)
{
 unsigned int curr=0;
 pci_read_config_dword(dev, addresses[bar],&curr);
 if (curr & PCI_BASE_ADDRESS_SPACE_IO) {
   curr &= PCI_BASE_ADDRESS_IO_MASK;
 } else {
   /* 32-bit specific. For 64 bit, don't mask the MSB?*/
   curr &= PCI_BASE_ADDRESS_MEM_MASK;
 }
 return curr;
}

static unsigned long mst_pci_resource_end(struct pci_dev *dev, int bar)
{
  unsigned int curr=0, mask=0, size;

  pci_read_config_dword(dev, addresses[bar],&curr);
  cli();
  pci_write_config_dword(dev, addresses[bar],~0);
  pci_read_config_dword(dev, addresses[bar],&mask);
  pci_write_config_dword(dev, addresses[bar],curr);
  sti();

  if (!mask) return 0;

  /*
   * apply the I/O or memory mask to current position.
   * note that I/O is limited to 0xffff, and 64-bit is not
   * supported by this simple implementation
   */
  if (curr & PCI_BASE_ADDRESS_SPACE_IO) {
    curr &= PCI_BASE_ADDRESS_IO_MASK;
  } else {
    curr &= PCI_BASE_ADDRESS_MEM_MASK;
  }

  /* extract the type, and the programmable bits */
  if (mask & PCI_BASE_ADDRESS_SPACE_IO) {
    mask &= PCI_BASE_ADDRESS_IO_MASK;
    size = (~mask + 1) & 0xffff; /* Bleah */
  } else {
    mask &= PCI_BASE_ADDRESS_MEM_MASK;
    size = ~mask + 1;
  }

  return size?curr+size-1:curr;
}

static inline unsigned long mst_pci_resource_len(struct pci_dev *dev, int bar)
{

  unsigned long start;
  unsigned long end;

  start=mst_pci_resource_start(dev,bar);
  end=mst_pci_resource_end(dev,bar);

  return (start==0 && start==end) ? 0: end - start + 1;
}
#else /* OLD_PCI_METHOD */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#if 0
static struct page *
nopage(struct vm_area_struct *vma,
                           unsigned long address, int write_access)
{ return NOPAGE_SIGBUS; /* send a SIGBUS */}
#endif
#else
#if 0
/* Before 2.4 we had to do ugly tricks to get bar size */
static unsigned long
nopage(struct vm_area_struct *vma,
                           unsigned long address, int write_access)
{ return 0; /* send a SIGBUS */}
#endif
static const unsigned int addresses[] = {
  PCI_BASE_ADDRESS_0,
  PCI_BASE_ADDRESS_1,
  PCI_BASE_ADDRESS_2,
  PCI_BASE_ADDRESS_3,
  PCI_BASE_ADDRESS_4,
  PCI_BASE_ADDRESS_5,
  0
};

/* Emulate Linux 2.4 on 2.2 */
/* The code originally comes from "Linux Device Drivers" by Alessandro
Rubini,   published by   O'Reilly & Associates.  Ch. 15*/
static unsigned long pci_resource_start(struct pci_dev *dev, int bar)
{
 unsigned int curr=0;
 pci_read_config_dword(dev, addresses[bar],&curr);
 if (curr & PCI_BASE_ADDRESS_SPACE_IO) {
   curr &= PCI_BASE_ADDRESS_IO_MASK;
 } else {
   /* 32-bit specific. For 64 bit, don't mask the MSB?*/
   curr &= PCI_BASE_ADDRESS_MEM_MASK;
 }
 return curr;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,17)
static unsigned long pci_resource_end(struct pci_dev *dev, int bar)
{
  unsigned int curr=0, mask=0, size;

  pci_read_config_dword(dev, addresses[bar],&curr);
  cli();
  pci_write_config_dword(dev, addresses[bar],~0);
  pci_read_config_dword(dev, addresses[bar],&mask);
  pci_write_config_dword(dev, addresses[bar],curr);
  sti();

  if (!mask) return 0;

  /*
   * apply the I/O or memory mask to current position.
   * note that I/O is limited to 0xffff, and 64-bit is not
   * supported by this simple implementation
   */
  if (curr & PCI_BASE_ADDRESS_SPACE_IO) {
    curr &= PCI_BASE_ADDRESS_IO_MASK;
  } else {
    curr &= PCI_BASE_ADDRESS_MEM_MASK;
  }

  /* extract the type, and the programmable bits */
  if (mask & PCI_BASE_ADDRESS_SPACE_IO) {
    mask &= PCI_BASE_ADDRESS_IO_MASK;
    size = (~mask + 1) & 0xffff; /* Bleah */
  } else {
    mask &= PCI_BASE_ADDRESS_MEM_MASK;
    size = ~mask + 1;
  }

  return curr+size;
}

static inline unsigned long pci_resource_len(struct pci_dev *dev, int bar)
{

  unsigned long start;
  unsigned long end;

  start=pci_resource_start(dev,bar);
  end=pci_resource_end(dev,bar);

  return (start==0 && start==end) ? 0: end - start + 1;
}

#endif /*2.2.17*/

#endif

#endif /* OLD_PCI_METHOD */



static int ioctl (struct inode *inode, struct file *file, unsigned int opcode, unsigned long udata_l)
{
  void* udata=(void*)udata_l;
  int minor=MINOR(inode->i_rdev);
  struct dev_data* dev=&devices[minor];
  int ret=0;

  /* By convention, any user gets read access
   * and is allowed to use the device.
   * Commands with no direction are administration
   * commands, and you need write permission
   * for this */

  if ( _IOC_DIR(opcode) == _IOC_NONE ) {
    if (! ( file->f_mode & FMODE_WRITE) ) return -EPERM;
  } else {
    if (! ( file->f_mode & FMODE_READ) ) return -EPERM;
  }

  if (down_interruptible(&devices[minor].sem)) {
    return -ERESTARTSYS;
  }


  switch (opcode) {

    case INIT:
      {
        struct mst_pci_init_st initd;
        struct pci_dev* pci_dev;
        /* Now have to init the new device */

        /* printk("<1> --------------------- ioctl INIT: copy_from_user(%p, %p/%lx/, %d)\n", */
        /*        &initd, udata, udata_l, (int)sizeof(initd)); */
        /* printk("<1> --------------------- copy_from_user:%d\n", */
        /*        (int)copy_from_user(&initd, udata, (int)sizeof(initd))); */
        if (copy_from_user(&initd, udata, sizeof(initd))) {
          ret=-EFAULT;
          goto fin;
        }

        printk("<1> --------------------- ioctl INIT: pci_find_slot(%d, %d)\n",
               initd.bus, initd.devfn);
        pci_dev=pci_find_slot (initd.bus, initd.devfn);
        if (!pci_dev)  {
          ret=-ENXIO;
          goto fin;
        }

#ifdef OLD_PCI_METHOD
        dev->size=mst_pci_resource_len(pci_dev, initd.bar);
        dev->bar=mst_pci_resource_start(pci_dev, initd.bar);
#else
        dev->size=pci_resource_len(pci_dev, initd.bar);
        dev->bar=pci_resource_start(pci_dev, initd.bar);
#endif
        dev->bufused=0;


        goto fin;
      }

    case STOP:
      devices[minor].size=0;
      goto fin;

    case PARAMS:
      {
        struct mst_pci_params_st paramsd;
        paramsd.bar=dev->bar;
        paramsd.size=dev->size;

        if (copy_to_user(udata, &paramsd, sizeof(paramsd))) {
          ret=-EFAULT;
        }
        goto fin;
      }
    case PARAMS_OLD:
      {
        struct mst_pci_params_old_st paramsd;
        paramsd.bar=dev->bar;
        paramsd.size=dev->size;

        if (copy_to_user(udata, &paramsd, sizeof(paramsd))) {
          ret=-EFAULT;
        }
        goto fin;
      }

    default:
      ret= -ENOTTY;
      goto fin;
  }

  fin:
  up(&devices[minor].sem);
  return ret;
}

/* from drivers/char/mem.c */
#ifndef pgprot_noncached

/*
 * This should probably be per-architecture in <asm/pgtable.h>
 */
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
	unsigned long prot = pgprot_val(_prot);

#if defined(__i386__) || defined(__x86_64__)
	/* On PPro and successors, PCD alone doesn't always mean
	    uncached because of interactions with the MTRRs. PCD | PWT
	    means definitely uncached. */
	if (boot_cpu_data.x86 > 3)
		prot |= _PAGE_PCD | _PAGE_PWT;
#elif defined(__powerpc__)
	prot |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#elif defined(__mc68000__)
#ifdef SUN3_PAGE_NOCACHE
	if (MMU_IS_SUN3)
		prot |= SUN3_PAGE_NOCACHE;
	else
#endif
	if (MMU_IS_851 || MMU_IS_030)
		prot |= _PAGE_NOCACHE030;
	/* Use no-cache mode, serialized */
	else if (MMU_IS_040 || MMU_IS_060)
		prot = (prot & _CACHEMASK040) | _PAGE_NOCACHE_S;
#endif

	return __pgprot(prot);
}
#endif /* !pgprot_noncached */
/*
 * Architectures vary in how they handle caching for addresses
 * outside of main memory.
 */
static inline int noncached_address(unsigned long addr)
{
#if defined(__i386__)
	/*
	 * On the PPro and successors, the MTRRs are used to set
	 * memory types for physical addresses outside main memory,
	 * so blindly setting PCD or PWT on those pages is wrong.
	 * For Pentiums and earlier, the surround logic should disable
	 * caching for the high addresses through the KEN pin, but
	 * we maintain the tradition of paranoia in this code.
	 */
 	return !( test_bit(X86_FEATURE_MTRR, (void *)&boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_K6_MTRR, (void *)&boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CYRIX_ARR, (void *)&boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CENTAUR_MCR, (void *)&boot_cpu_data.x86_capability) )
	  && addr >= __pa(high_memory);
#else
	return addr >= __pa(high_memory);
#endif
}

static int mmap(struct file *file, struct vm_area_struct *vma)
{

  int ret=0;
  unsigned long off;
  unsigned long vsize;
  unsigned long offset;
  struct dev_data* dev=(struct dev_data*) file->private_data;

  if (!dev) return -ENOTTY; /*Shouldn't happend, as open assigns it
                              but just in case */

  if (down_interruptible(&dev->sem))
    return -ERESTARTSYS;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  off = vma->vm_pgoff << PAGE_SHIFT;
#else
  off = vma->vm_offset;
#endif


  vsize = vma->vm_end - vma->vm_start;

  if ((dev->size <= off) || (dev->size < off+vsize)) {
    ret=-EINVAL;
    goto fin;
  }

  offset = dev->bar+off;
  /*
   *    * Accessing memory above the top the kernel knows about or
   *    * through a file pointer that was marked O_SYNC will be
   *    * done non-cached.
   *    */
  if (noncached_address(offset) || (file->f_flags & O_SYNC))
    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);


  /* Mark this area as I/O, that is non-cacheable.
   * Don't do it by default for regular memory  */
  if (offset >= __pa(high_memory) || (file->f_flags & O_SYNC)) {
    vma->vm_flags |= VM_IO;
  }

	/* Don't try to swap out physical pages.. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  vma->vm_flags |= VM_RESERVED; /* Why do I insist on reserved? is locked ok? */
#else
  vma->vm_flags |= VM_LOCKED;
#endif
 /* vma->vm_ops->nopage=nopage;*/

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10))
  ret=remap_pfn_range(vma, vma->vm_start, offset >> PAGE_SHIFT,
                       vsize, vma->vm_page_prot);
#else
#if defined(RED_HAT_LINUX_KERNEL) && ( \
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,4,9)) ||\
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,4,20)) ||\
    (LINUX_VERSION_CODE == KERNEL_VERSION(2,4,21)) ||\
    0 )
#define REMAP_PREFIX current->mm->mmap,
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define REMAP_PREFIX current->mm->mmap,
#else
#define REMAP_PREFIX
#endif

  ret=remap_page_range(REMAP_PREFIX vma->vm_start, offset,
                       vsize, vma->vm_page_prot);
#endif
  if (ret) goto fin;

fin:
  up(&dev->sem);
  return ret;
}

#define SUPPORT_MMAP 1

#include "driver_common.c"


