#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/pci.h>

#include "mst_pciconf.h"
#define  INIT   PCICONF_INIT
#define  STOP   PCICONF_STOP
#define  READ4  PCICONF_READ4
#define  WRITE4 PCICONF_WRITE4
#define  MODIFY PCICONF_MODIFY

#define PCICONF_READ4_OLD  _IOR (PCICONF_MAGIC,1,struct mst_read4_old_st)
struct mst_read4_old_st {
        unsigned long long __attribute__((packed)) offset;
        unsigned long long __attribute__((packed)) data; /*OUT*/
};

#define PCICONF_WRITE4_OLD _IOW (PCICONF_MAGIC,2,struct mst_write4_old_st)
struct mst_write4_old_st {
        unsigned long long __attribute__((packed)) offset;
        unsigned long long __attribute__((packed)) data;
};

#define PCICONF_MODIFY_OLD _IOWR(PCICONF_MAGIC,3,struct mst_modify_old_st)
struct mst_modify_old_st {
        unsigned long long __attribute__((packed)) offset;
        unsigned long long __attribute__((packed)) data;
        unsigned long long __attribute__((packed)) mask;
        unsigned long long __attribute__((packed)) old_data; /*OUT*/
};
#define  READ4_OLD  PCICONF_READ4_OLD
#define  WRITE4_OLD PCICONF_WRITE4_OLD
#define  MODIFY_OLD PCICONF_MODIFY_OLD

static char* name="mst_pciconf";

#if CONFIG_COMPAT && !HAVE_COMPAT_IOCTL
static unsigned int
ioctl_cmds[] ={
    READ4,
    WRITE4,
    MODIFY,
    READ4_OLD,
    WRITE4_OLD,
    MODIFY_OLD
};
#endif

/* Allow minor numbers 0-255 */
#define MAXMINOR 256
#define BUFFER_SIZE 256

MODULE_AUTHOR("Michael S. Tsirkin - Mellanox Technologies LTD");
MODULE_DESCRIPTION("Mellanox Software Tools driver for accessing Tavor Configuration");
MODULE_SUPPORTED_DEVICE("MT23109");

struct dev_data {
  struct pci_dev* pci_dev;
  int addr_reg;
  int data_reg;
  volatile int valid;
  struct semaphore sem;
  char buf[BUFFER_SIZE];
  int bufused;
};

static struct dev_data devices[MAXMINOR];


static int ioctl (struct inode *inode, struct file *file, unsigned int opcode, unsigned long udata_l)
{
  void* udata=(void*)udata_l;
  int minor=MINOR(inode->i_rdev);
  struct dev_data* dev=&devices[minor];
  int ret=0;
  unsigned int d;

  /* By convention, any user gets read access
   * and is allowed to use the device.
   * Commands with no direction are administration
   * commands, and you need write permission
   * for this */

  if ( _IOC_DIR(opcode) == _IOC_NONE ) {
    if (! ( file->f_mode & FMODE_WRITE) ) return -EPERM;
  } else {
    if (! ( file->f_mode & FMODE_READ) ) return -EPERM;
  }

  if (down_interruptible(&devices[minor].sem)) {
    return -ERESTARTSYS;
  }


  switch (opcode) {

    case INIT: 
      {
        struct mst_pciconf_init_st initd;
        /* Now have to init the new device */

        if (copy_from_user(&initd, udata, sizeof(initd))) {
          ret=-EFAULT;
          goto fin;
        }
        dev->pci_dev=pci_find_slot (initd.bus, initd.devfn);
        if (!dev->pci_dev)  {
          ret=-ENXIO;
          goto fin;
        }

        dev->bufused=0;

        dev->valid=1;

		dev->addr_reg=initd.addr_reg;
		dev->data_reg=initd.data_reg;

        goto fin;
      }

    case STOP:
      dev->valid=0;
      goto fin;


    case WRITE4:
      {
        struct mst_write4_st write4d;
        if (copy_from_user(&write4d, udata, sizeof(write4d))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }
        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, write4d.offset);
        if (ret) goto fin;

        ret=pci_write_config_dword(dev->pci_dev, dev->data_reg, write4d.data);
        goto fin;
      }
    case READ4:
      {
        struct mst_read4_st read4d;
        struct mst_read4_st* r_udata=(struct mst_read4_st*)udata;
        if (copy_from_user(&read4d, udata, sizeof(read4d))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }

        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, read4d.offset);
        if (ret) goto fin;

        ret=pci_read_config_dword(dev->pci_dev, dev->data_reg, &d);
        if (ret) goto fin;

        ret=put_user(d,&(r_udata->data))?-EFAULT:0;
        goto fin;
      }
    case MODIFY:
      {
        struct mst_modify_st modifyd;
        struct mst_modify_st* m_udata=(struct mst_modify_st*)udata;
        if (copy_from_user(&modifyd, udata, sizeof(modifyd))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }

        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, modifyd.offset);
        if (ret) goto fin;

        ret=pci_read_config_dword(dev->pci_dev, dev->data_reg, &d);
        if (ret) goto fin;

        d= (d & ~modifyd.mask) | (modifyd.data & modifyd.mask);

        ret=pci_write_config_dword(dev->pci_dev, dev->data_reg, d);
        if (ret) goto fin;

        ret=put_user(d,&(m_udata->old_data))?-EFAULT:0;
        goto fin;
      }
    case WRITE4_OLD:
      {
        struct mst_write4_old_st write4d;
        if (copy_from_user(&write4d, udata, sizeof(write4d))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }
        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, write4d.offset);
        if (ret) goto fin;

        ret=pci_write_config_dword(dev->pci_dev, dev->data_reg, write4d.data);
        goto fin;
      }
    case READ4_OLD:
      {
        struct mst_read4_old_st read4d;
        struct mst_read4_old_st* r_udata=(struct mst_read4_old_st*)udata;
        if (copy_from_user(&read4d, udata, sizeof(read4d))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }

        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, read4d.offset);
        if (ret) goto fin;

        ret=pci_read_config_dword(dev->pci_dev, dev->data_reg, &d);
        if (ret) goto fin;

	read4d.data=d;

        ret=copy_to_user(&(r_udata->data),&(read4d.data),sizeof(read4d.data))?
		-EFAULT:0;
        goto fin;
      }
    case MODIFY_OLD:
      {
        struct mst_modify_old_st modifyd;
        struct mst_modify_old_st* m_udata=(struct mst_modify_old_st*)udata;
        if (copy_from_user(&modifyd, udata, sizeof(modifyd))) {
          ret=-EFAULT;
          goto fin;
        }

        if (!dev->valid) {
          ret=-ENOTTY;
          goto fin;
        }

        ret=pci_write_config_dword(dev->pci_dev, dev->addr_reg, modifyd.offset);
        if (ret) goto fin;

        ret=pci_read_config_dword(dev->pci_dev, dev->data_reg, &d);
        if (ret) goto fin;

        d= (d & ~modifyd.mask) | (modifyd.data & modifyd.mask);

        ret=pci_write_config_dword(dev->pci_dev, dev->data_reg, d);
        if (ret) goto fin;

        ret=copy_to_user(&(m_udata->data),&(modifyd.data),sizeof(modifyd.data))?
		-EFAULT:0;
        goto fin;
      }
    default:
      ret= -ENOTTY;
      goto fin;
  }

  fin:
  up(&dev->sem);
  return ret;
}

#include "driver_common.c"


