#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/ioctl.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/mpc8xx.h>

#include "mst_ppc.h"
#define  INIT   PPC_INIT
#define  STOP   PPC_STOP
#define  PARAMS PPC_PARAMS

static char name[]="mst_ppc";

/* #define MST_PPC_DEBUG */

/* Allow minor numbers 0-255 */
#define MAXMINOR 256
#define BUFFER_SIZE 256

MODULE_AUTHOR("Alex Kholodenko - Mellanox Technologies LTD");
MODULE_DESCRIPTION("Mellanox Software Tools driver for accessing Anafa via PPC IF");
MODULE_SUPPORTED_DEVICE("MT43132");

struct dev_data {
      unsigned int cs;
	  unsigned int addr;
      volatile ssize_t size;
      struct semaphore sem;
      char buf[BUFFER_SIZE];
      int bufused;
};

static struct dev_data devices[MAXMINOR];

static unsigned int get_immr(void);
static int _8xx_cs_map( unsigned int cs_num, unsigned int size);
static int _8xx_cs_unmap( unsigned int cs_num );
static unsigned int _8xx_find_free_segment(unsigned int size);
static unsigned int mst_virt_to_phys (unsigned long va);
static void mst_vm_close (struct vm_area_struct *vma);

static struct vm_operations_struct mst_vm_ops = {
        close:&mst_vm_close,
};

/*
 * Translate a virtual address to physical address
 * Assume we always work with mapped user-space memory here.
 * For kernel we'd need passing &init_mm instead of current->mm
 */
static unsigned int
mst_virt_to_phys (va)
        unsigned long va;
{
#if LINUX_VERSION_CODE - KERNEL_VERSION(2,4,9)

        pgd_t *pgd;
        pmd_t *pmd;
        pte_t *pte;
        unsigned int pa = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        spin_lock (&current->mm->page_table_lock);
#else
#ifdef __SMP__
        lock_kernel ();
#endif
#endif

        pgd = pgd_offset (current->mm, va);
        pmd = pmd_offset (pgd, va);
        if (!pmd_present (*pmd))
                goto fin;
        pte = pte_offset (pmd, va);

        if (!pte_present (*pte))
                goto fin;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        pa = __pa ((unsigned int) page_address (pte_page (*pte)));
#else
        pa = __pa ((unsigned int) pte_page (*pte));
#endif

        pa |= ((PAGE_SIZE - 1) & (unsigned int) va);
      fin:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
        spin_unlock (&current->mm->page_table_lock);
#else
#ifdef __SMP__
        unlock_kernel ();
#endif
#endif

        return pa;

#else
        return 0;
#endif
}

static int ioctl (struct inode *inode, struct file *file, unsigned int opcode, unsigned long udata_l)
{
  void* udata=(void*)udata_l;
  int minor=MINOR(inode->i_rdev);
  struct dev_data* dev=&devices[minor];
  int ret=0;

  /* By convention, any user gets read access
   * and is allowed to use the device.
   * Commands with no direction are administration
   * commands, and you need write permission
   * for this */

  if ( _IOC_DIR(opcode) == _IOC_NONE ) {
    if (! ( file->f_mode & FMODE_WRITE) ) return -EPERM;
  } else {
    if (! ( file->f_mode & FMODE_READ) ) return -EPERM;
  }

  if (down_interruptible(&devices[minor].sem)) {
    return -ERESTARTSYS;
  }


  switch (opcode) {

    case INIT:
      {
        struct mst_ppc_init_st initd;
		unsigned int addr;
#ifdef MST_PPC_DEBUG
		printk( "%s: INIT IOCTL\n", __func__ );  
#endif
        if (copy_from_user(&initd, udata, sizeof(initd))) {
          ret=-EFAULT;
          goto fin;
        }
        if( dev->cs != (unsigned int)(-1) ){
			if( dev->cs != initd.cs)
                ret=-EINVAL;
            goto fin;             /*  Device already initialized */
		}
		if( (addr=_8xx_cs_map( initd.cs, initd.size )) == (unsigned int)(-1) ) {
			ret=-EAGAIN;
			goto fin;
		}
		dev->cs = initd.cs;
		dev->size = initd.size;
		dev->addr = addr;
		goto fin;
      }

    case STOP:
#ifdef MST_PPC_DEBUG
		printk( "%s: STOP IOCTL\n", __func__ );  
#endif
      if( dev->cs == (unsigned int)(-1) )
	   	  goto fin;
	   _8xx_cs_unmap( dev->cs );
	   dev->cs = (unsigned int)(-1);
      goto fin;

    case PARAMS:
      {
        struct mst_ppc_params_st paramsd;
#ifdef MST_PPC_DEBUG
		printk( "%s: PARAMS IOCTL\n", __func__ );  
#endif
		paramsd.cs = dev->cs;
		paramsd.size = dev->size;
		paramsd.addr = dev->addr;
        if (copy_to_user(udata, &paramsd, sizeof(paramsd))) {
          ret=-EFAULT;
        }
        goto fin;
      }
    case PPC_VTOP:
      {
        struct mst_ppc_vtop vtop;
        if (copy_from_user(&vtop, udata, sizeof(vtop))) {
            ret = -EFAULT;
            goto fin;
        }

        vtop.pa = mst_virt_to_phys ((unsigned long) vtop.va);

        if (copy_to_user (udata, &vtop, sizeof (vtop))) {
            ret = -EFAULT;
            goto fin;
        }
        break;
      }

    default:
      ret= -ENOTTY;
      goto fin;
  }

  fin:
  up(&devices[minor].sem);
  return ret;
}


/*
*
* Code from drivers/char/mem.c (XOL)
*
*/
#ifndef pgprot_noncached

/*
 * This should probably be per-architecture in <asm/pgtable.h>
 */
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
	unsigned long prot = pgprot_val(_prot);

#if defined(__i386__) || defined(__x86_64__)
	/* On PPro and successors, PCD alone doesn't always mean 
	    uncached because of interactions with the MTRRs. PCD | PWT
	    means definitely uncached. */ 
	if (boot_cpu_data.x86 > 3)
		prot |= _PAGE_PCD | _PAGE_PWT;
#elif defined(__powerpc__)
	prot |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#elif defined(__mc68000__)
#ifdef SUN3_PAGE_NOCACHE
	if (MMU_IS_SUN3)
		prot |= SUN3_PAGE_NOCACHE;
	else
#endif
	if (MMU_IS_851 || MMU_IS_030)
		prot |= _PAGE_NOCACHE030;
	/* Use no-cache mode, serialized */
	else if (MMU_IS_040 || MMU_IS_060)
		prot = (prot & _CACHEMASK040) | _PAGE_NOCACHE_S;
#endif

	return __pgprot(prot);
}

#endif /* !pgprot_noncached */

/*
 * Architectures vary in how they handle caching for addresses 
 * outside of main memory.
 */
static inline int noncached_address(unsigned long addr)
{
#if defined(__i386__)
	/* 
	 * On the PPro and successors, the MTRRs are used to set
	 * memory types for physical addresses outside main memory, 
	 * so blindly setting PCD or PWT on those pages is wrong.
	 * For Pentiums and earlier, the surround logic should disable 
	 * caching for the high addresses through the KEN pin, but
	 * we maintain the tradition of paranoia in this code.
	 */
 	return !( test_bit(X86_FEATURE_MTRR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_K6_MTRR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CYRIX_ARR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CENTAUR_MCR, &boot_cpu_data.x86_capability) )
	  && addr >= __pa(high_memory);
#else
	return addr >= __pa(high_memory);
#endif
}


static int
mmap_gfp (struct file *file, struct vm_area_struct *vma)
{
        char *kva;
        char *pageva;
        int error = 0;
        unsigned int ksize = vma->vm_end - vma->vm_start;

        /* make it non cacheable */
        vma->vm_page_prot = pgprot_noncached (vma->vm_page_prot);

        /* Don't try to swap out physical pages */
        vma->vm_flags |= VM_RESERVED;

        /* Don't dump to a core file  */
        vma->vm_flags |= VM_IO;

        /* Allocate memory */
        /*  kva = kmalloc (ksize, GFP_KERNEL); */
        kva = (void *) __get_free_pages (GFP_KERNEL, get_order (ksize));

        if (!kva)
                return -ENOMEM;

        for (pageva = kva; pageva < (kva + ksize); pageva += PAGE_SIZE)
                SetPageReserved (virt_to_page (pageva));

        if (remap_page_range (vma->vm_start, __pa (kva),
                              vma->vm_end - vma->vm_start, vma->vm_page_prot))
                error = -EAGAIN;

        if (error) {
                for (pageva = kva; pageva < (kva + ksize); pageva += PAGE_SIZE)
                        ClearPageReserved (virt_to_page (pageva));
                free_pages ((unsigned long) kva, get_order (ksize));

        } else {
                vma->vm_pgoff = ((unsigned int) kva) >> PAGE_SHIFT;
                vma->vm_ops = &mst_vm_ops;
        }

        return error;
}

static int
mmap_cr (struct dev_data * dev, struct file * file, struct vm_area_struct * vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

	unsigned long vsize;

	vsize = vma->vm_end - vma->vm_start;
	if ((dev->size <= offset) || (dev->size < offset+vsize)) {
	  return -EINVAL;
	}
	offset += dev->addr;

	/*
	 * Accessing memory above the top the kernel knows about or
	 * through a file pointer that was marked O_SYNC will be
	 * done non-cached.
	 */
	if (noncached_address(offset) || (file->f_flags & O_SYNC))
		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	/* Don't try to swap out physical pages.. */
	vma->vm_flags |= VM_RESERVED;

	/*
	 * Don't dump addresses that are not real memory to a core file.
	 */
	if (offset >= __pa(high_memory) || (file->f_flags & O_SYNC))
		vma->vm_flags |= VM_IO;

	if (remap_page_range(vma->vm_start, offset, vma->vm_end-vma->vm_start,
			     vma->vm_page_prot))
		return -EAGAIN;

        return 0;

}

static int
mmap (struct file *file, struct vm_area_struct *vma)
{
        int rc;

	struct dev_data* dev=(struct dev_data*) file->private_data;
	if (!dev) return -ENOTTY; /*Shouldn't happend, as open assigns it
                                    but just in case */
	if (down_interruptible(&dev->sem))
	  return -ERESTARTSYS;

        if (vma->vm_pgoff)
                rc = mmap_gfp (file, vma);
        else
                rc = mmap_cr (dev, file, vma);

        up(&dev->sem);
        return rc;
}

static void
mst_vm_close (vma)
        struct vm_area_struct *vma;
{
        char *kva = (char *) (vma->vm_pgoff << PAGE_SHIFT);
        unsigned int ksize = vma->vm_end - vma->vm_start;
        char *pageva;

        for (pageva = kva; pageva < (kva + ksize); pageva += PAGE_SIZE)
                ClearPageReserved (virt_to_page (pageva));
        free_pages ((unsigned long) kva, get_order (ksize));
}


/*
* PPC 8xx MMU operations
*
*/
static unsigned int get_immr(void)
{
    unsigned int result;
    asm ( "mfspr 3, 638");
    asm ( "stw 3, %0" : "=m"(result) );
    return result;
}                                               

#define NUM_OF_CS 8
#define ADDR_MASK		0xFFFF8000

struct mmu_regs_st
{
	struct 
	{
		unsigned int BR;	/*  Base Register */
		unsigned int OR;	/*  Options Register */
	} bank[NUM_OF_CS];	/*  Banks */
} *mmu_regs;

#define __BR_VALID 		0x00000001	/*  Bank CS Valid */
#define __BR_PORT16		0x00000800	/*  Port is 16 bit port */

#define __OR_BIH		0x00000100	/*  Burst Mode not supported */
#define __OR_SCY		0x000000F0	/*  Cycle length is 16 wait states */
#define __OR_SETA		0x00000008  /*  Peripheral generates Transport ACK */

#define MIN_MAP_SIZE 0x00008000

static int _8xx_cs_map( unsigned int cs_num, unsigned int size)
{
    unsigned int segment;

	mmu_regs = (struct mmu_regs_st*)(ioremap_nocache((get_immr() & 0xffff0000)+0x100, sizeof(struct mmu_regs_st)));
	if( ! mmu_regs )
	{
		printk( "%s: MMU registers ioremap_nocache failed\n", __func__ );
		return -1;
	}

    /*  Check that requested chip selected if free */
    if( mmu_regs->bank[cs_num].BR & __BR_VALID )
    {
        printk("%s: Chip select (cs = %d) allready active.\n", __func__, cs_num);
        iounmap(mmu_regs);
        return(-1);
    }
    /*  It is possible to map only a minimum of 32K */
    if(size < MIN_MAP_SIZE)
    {
        printk("%s: region size is too small ( 0x%08x < 0x%08x )\n", __func__, size, MIN_MAP_SIZE);
        iounmap(mmu_regs);
        return(-1);
    }
    
	segment = _8xx_find_free_segment(size);
    if(segment == (unsigned int)(-1))
    {
		iounmap(mmu_regs);
        return(-1);
    }
    printk("%s: Allocating segment(0x%08x) for chipselect %x\n", __func__, segment, cs_num);
    /* initialize chip of ppc to work */
	mmu_regs->bank[cs_num].OR = ((~(size-1)) & ADDR_MASK) | __OR_BIH | __OR_SCY | __OR_SETA;
	mmu_regs->bank[cs_num].BR = (segment & ADDR_MASK)  | __BR_PORT16 | __BR_VALID;
#ifdef MST_PPC_DEBUG
	printk("%s: Region CS=%d BR=%08x OR=%08x\n", __func__, cs_num, 
										mmu_regs->bank[cs_num].BR, mmu_regs->bank[cs_num].OR );
#endif
	iounmap(mmu_regs);
    return(segment);
}     

static int _8xx_cs_unmap( unsigned int cs_num )
{
	mmu_regs = (struct mmu_regs_st*)(ioremap_nocache((get_immr() & 0xffff0000)+0x100, sizeof(struct mmu_regs_st)));
	if( ! mmu_regs )
	{
		printk( "%s: MMU registers ioremap_nocache failed\n", __func__ );
		return -1;
	}

	mmu_regs->bank[cs_num].BR &= (~(__BR_VALID));
    printk("%s: Chip select (cs = %d) now not valid\n", __func__, cs_num);
    iounmap(mmu_regs);
	return 0;
}     

static unsigned int _8xx_find_free_segment(unsigned int size)
{
	int i;
	unsigned int addr, naddr, mask, umask;
	struct 
	{
		unsigned int addr;
		unsigned int mask;
	} map[NUM_OF_CS+1];

	/*  Fill CS Info */
	for( i=0; i<NUM_OF_CS; i++ )
	{
		if( mmu_regs->bank[i].BR & __BR_VALID )
		{
			map[i].addr = mmu_regs->bank[i].BR & ADDR_MASK;
			map[i].mask = mmu_regs->bank[i].OR & ADDR_MASK;
#ifdef MST_PPC_DEBUG
			printk( " CS%d BR=%08x OR=%08x\n", i, 
										mmu_regs->bank[i].BR, mmu_regs->bank[i].OR );
#endif
		}
		else
		{	/*  Not Valid segment */
			map[i].addr = (unsigned int)(-1);
			map[i].mask = (unsigned int)(-1);
		}
	}
    map[NUM_OF_CS].addr = (get_immr() & 0xffff0000);	/*  Put IMMR region  */
	map[NUM_OF_CS].mask = 0xFFFF0000;
#ifdef MST_PPC_DEBUG
	printk( " IMMR A=%08x  M=%08x\n", map[NUM_OF_CS].addr, map[NUM_OF_CS].mask );  
#endif

	/*  Look for first suitable segment */
	addr=0;
	mask=~(size-1);
	while( 1 )
	{
		for( i=0; i<NUM_OF_CS+1; i++ )
        {
            umask = mask & map[i].mask;

			if( (addr & umask) == (map[i].addr & umask) )
			{
				naddr = addr + (~umask)+1;
				if( naddr < addr )
					return -1;
				addr = naddr;
				break;
			}
        }
		if( i == NUM_OF_CS+1 )
			return addr;
	}
}

/*
* Common MST Driver code
*
*/
#define SUPPORT_MMAP 1
#define MST_PPC_DRIVER
#include "driver_common.c"

