/*
 *
 *  mtcr_i2c.c - Mellanox Software tools (mst) driver: I2C master routines
 *
 */

#include <stdio.h>

#include <asm/byteorder.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <endian.h>

#include "mtcr.h"
#include "mtcr_i2c.h"
#include "mst_pci.h"
#include "mst_pciconf.h"

#define REPEAT_WRITE_IF_NACK
/* #define LOWR_LEVEL_DEBUG */
/* #define LOWW_LEVEL_DEBUG */
/* #define MID_LEVEL_DEBUG */
/* #define HIGH_LEVEL_DEBUG */

enum CONSTANTS {
    EPROM_PRIMARY_COUNT = 5000,  /*  Amount of "fast" check */
    EPROM_SECONDARY_COUNT = 10,  /*  Amount of "slow" check */
    EPROM_REQUEST_PAUSE = 100,   /*  Microseconds for pause */
    MAX_REPETITION = 100,        /*  Maximal number of transaction repet. */
    ST_IN_TRANSACTION = 7,       /*  Status: cycle in transaction completed OK */
    ST_TIMEOUT = 6,              /*  Status: transaction timeout */
    ST_IDLE = 0                  /*  Status: IDLE */
};

/*  SPM registers offset */
#define SPM_OFFS ((mf->itype == MST_TAVOR) ? 0xf0180 : 0x3180)

#define CLEAR(st) memset(&(st), 0, sizeof(st))

/*  Like mread4, but only for PCI/PCICONF interfaces and doesn't check is_i2cm */
static int mread4_(mfile *mf, unsigned int offset, unsigned int *value)
{
    int                 rc = 4;
    struct mst_read4_st r4;

#ifdef LOWR_LEVEL_DEBUG
    printf("mread4_: tp=%d,itype=%d - try read offs=0x%x\n",
           mf->tp, mf->itype, offset);
#endif

    switch(mf->tp)
    {
    case MST_PCI:
        if (!mf->ptr)
        {
            errno = EFAULT;
            return -1;
        }

        if (mf->itype == MST_TAVOR)
            *value = __be32_to_cpu(*((unsigned int *)((char *)mf->ptr + offset)));
        else
            *value = *((unsigned int *)((char *)mf->ptr + offset));
        break;
    case MST_PCICONF:
        CLEAR(r4);
        r4.offset = offset;
        if ((rc = ioctl(mf->fd, PCICONF_READ4, &r4)) < 0)
            return rc;
        *value = r4.data;
        rc = 4;
        break;
    default:
        errno = EINVAL;
        rc = -1;
    }

#ifdef LOWR_LEVEL_DEBUG
    printf("    mread4_: rc=%d,offs=0x%x val=0x%08x\n", rc, offset, *value);
#endif

    return rc;
}

/*  Like mwrite4, but only for PCI/PCICONF interfaces and doesn't check is_i2cm */
static int mwrite4_(mfile *mf, unsigned int offset, unsigned int value)
{
    int                  rc = 4;
    struct mst_write4_st r4;

#ifdef LOWW_LEVEL_DEBUG
    printf("mwrite4_: tp=%d,itype=%d - try write offs=0x%x val=0x%08x\n",
           mf->tp, mf->itype, offset, value);
#endif

    switch(mf->tp)
    {
    case MST_PCI:
        if (!mf->ptr)
        {
            errno = EFAULT;
            return -1;
        }

        if (mf->itype == MST_TAVOR)
            *((unsigned int *)((char *)mf->ptr + offset)) = __cpu_to_be32(value);
        else
            *((unsigned int *)((char *)mf->ptr + offset)) = value;
        break;
    case MST_PCICONF:
        CLEAR(r4);
        r4.offset = offset;
        r4.data = value;
        if ((rc = ioctl(mf->fd, PCICONF_WRITE4, &r4)) < 0)
            return rc;
        rc = 4;
        break;
    default:
        errno = EINVAL;
        rc = -1;
    }

#ifdef LOWW_LEVEL_DEBUG
    printf("    mwrite4_: rc=%d\n", rc);
#endif

    return rc;
}


/*  Return -1 when I/O error or cmd doesn't become zero */
/*  Return status field otherwise */
static int wait_trans(mfile *mf)
{
    u_int32_t     val;
    u_int32_t     cnt = 0;

    while(1)
    {
        if (mread4_(mf, SPM_OFFS, &val) != 4)
            return -1;
        if (!(val >> 29))
            break;
        if (cnt > EPROM_PRIMARY_COUNT)
            usleep(EPROM_REQUEST_PAUSE);
        if (cnt > EPROM_PRIMARY_COUNT+EPROM_SECONDARY_COUNT)
        {
            errno = EBUSY;
            return -1;
        }
        cnt++;
    }
    return (val>>16) & 0x07;
}

/*  Return status field or -1 on fatal failure */
static int w_trans(mfile *mf, void *data_, int len)
{
    char         *data = (char *)data_;
    int           vbt = -1;
    u_int32_t     val = 0;

#ifdef MID_LEVEL_DEBUG
    printf("w_trans: data=0x%08x, len:%d\n", *(unsigned int *)data, len);
#endif

    switch (len)
    {
    case 1:
        vbt = 0;
        val = (*data << 24) & 0xff000000;
        break;
    case 2:
        vbt = 1;
        val = ((*data << 24) & 0xff000000) | ((data[1] << 16) & 0xff0000);
        break;
    case 3:
        vbt = 2;
        val = ((*data << 24) & 0xff000000) | ((data[1] << 16) & 0xff0000) |
            ((data[2] << 8) & 0xff00);
        break;
    case 4:
        vbt = 3;
        val = ((*data << 24) & 0xff000000) | ((data[1] << 16) & 0xff0000) |
            ((data[2] << 8) & 0xff00) | (data[3] & 0xff);
        break;
    default:
        return -1;
    }
    if (mwrite4_(mf, SPM_OFFS + 4, val) != 4)
        return -1;
    if (mwrite4_(mf, SPM_OFFS, (2<<29) | (vbt << 22) | (mf->i2c_slave & 0x7f)) != 4)
        return -1;

    return wait_trans(mf);
}


/*  Return status field or -1 on fatal failure */
static int r_trans(mfile *mf, void *data_, int len)
{
    char         *data = (char *)data_;
    int           vbt = -1, rc;
    u_int32_t     val = 0;

#ifdef MID_LEVEL_DEBUG
    printf("r_trans: len:%d\n", len);
#endif

    switch (len)
    {
    case 1:
        vbt = 0;
        break;
    case 2:
        vbt = 1;
        break;
    case 3:
        vbt = 2;
        break;
    case 4:
        vbt = 3;
        break;
    default:
        return -1;
    }

    if (mwrite4_(mf, SPM_OFFS, (1<<29) | (vbt << 22) | (mf->i2c_slave & 0x7f)) != 4)
        return -1;
    rc = wait_trans(mf);
    if (rc != 7)
        return rc;

    if (mread4_(mf, SPM_OFFS + 4, &val) != 4)
        return -1;

    switch (len)
    {
    case 1:
        *data = (val >> 24) & 0xff;
        break;
    case 2:
        *data++ = (val >> 24) & 0xff;
        *data   = (val >> 16) & 0xff;
        break;
    case 3:
        *data++ = (val >> 24) & 0xff;
        *data++ = (val >> 16) & 0xff;
        *data   = (val >> 8)  & 0xff;
        break;
    case 4:
        *data++ = (val >> 24) & 0xff;
        *data++ = (val >> 16) & 0xff;
        *data++ = (val >> 8)  & 0xff;
        *data   = val         & 0xff;
        break;
    default:
        return -1;
    }
    return 7;
}

/*  Return status field or -1 on fatal failure */
static int end_trans(mfile *mf)
{
    if (mwrite4_(mf, SPM_OFFS, (3<<29) | (mf->i2c_slave & 0x7f)) != 4)
        return -1;
    return wait_trans(mf);
}

int i2c_master_write_cr(mfile *mf, unsigned int value, unsigned int offset,
                        int len)
{
    unsigned int   ivalue = __cpu_to_be32(value);
    unsigned short svalue = __cpu_to_be16((unsigned short)(value & 0xffff));
    unsigned char  cvalue = value & 0xff;
    int            rc=0, repeat=1;
    unsigned short offs = __cpu_to_be16(offset & 0xffff);
    unsigned char  off1 = offset & 0xff;

#ifdef HIGH_LEVEL_DEBUG
    printf("i2c_master_write_cr i2c:0x%x, offs:0x%x, val:0x%x, len:%d\n",
           mf->i2c_slave, offset, value, len);
#endif

    offset = __cpu_to_be32(offset);

    while (repeat)
    {
        /*  Write address (1byte for DIMM, 2bytes for Gamla and 4bytes for Tavor) */
        switch(mf->dtype)
        {
        case MST_TAVOR:
            rc = w_trans(mf, &offset, sizeof(offset));
            break;
        case MST_GAMLA:
            rc = w_trans(mf, &offs, sizeof(offs));
            break;
        case MST_DIMM:
            rc = w_trans(mf, &off1, sizeof(off1));
            break;
        case MST_NOADDR:
            rc = ST_IN_TRANSACTION;  /*  FAKE! No address transaction */
            break;
        }

        /*  Write data */
        if (rc == ST_IN_TRANSACTION)
            switch(len)
            {
            case 1:
                rc = w_trans(mf, &cvalue, sizeof(cvalue));
                break;
            case 2:
                rc = w_trans(mf, &svalue, sizeof(svalue));
                break;
            case 4:
                rc = w_trans(mf, &ivalue, sizeof(ivalue));
                break;
            default:
                errno = EINVAL;
                return -1;
            }

        /*  End transaction (anyway, even it was an error)s */
        if (end_trans(mf) < 0)
            return -1;

        /*  Now check Gw status and repeat transaction if timeout */
        switch (rc)
        {
        case ST_IN_TRANSACTION:
            repeat = 0;
            break;
#ifdef REPEAT_WRITE_IF_NACK
        default:
            if (++repeat > MAX_REPETITION)
            {
                errno = EIO;
                return -1;
            }
#else
        case ST_TIMEOUT:
            if (++repeat > MAX_REPETITION)
                return -1;
            /* printf("i2c_master_write_cr: repeat:%d\n", repeat); */
            break;
        default:
            errno = EIO;
            return -1;
#endif
        }
    }

#ifdef HIGH_LEVEL_DEBUG
    printf("    i2c_master_write_cr: rc=%d\n", len);
#endif

    return len;
}

int i2c_master_read_cr(mfile *mf, void *value, unsigned int offset, int len)
{
    unsigned int   *ivalue = (unsigned int *)value;
    unsigned short *svalue = (unsigned short *)value;
    unsigned char  *cvalue = (unsigned char *)value;
    int            rc=0, repeat=1;
    unsigned short offs = __cpu_to_be16(offset & 0xffff);
    unsigned char  off1 = offset & 0xff;

#ifdef HIGH_LEVEL_DEBUG
    printf("i2c_master_read_cr i2c:0x%x, offs:0x%x, len:%d\n",
           mf->i2c_slave, offset, len);
#endif

    offset = __cpu_to_be32(offset);

    while (repeat)
    {
        /*  Write address (1byte for DIMM, 2bytes for Gamla and 4bytes for Tavor) */
        switch(mf->dtype)
        {
        case MST_TAVOR:
            rc  = w_trans(mf, &offset, sizeof(offset));
            break;
        case MST_GAMLA:
            rc = w_trans(mf, &offs, sizeof(offs));
            break;
        case MST_DIMM:
            rc = w_trans(mf, &off1, sizeof(off1));
            break;
        case MST_NOADDR:
            rc = ST_IN_TRANSACTION;  /*  FAKE! No address transaction */
            break;
        }

        /*  Read data */
        if (rc == ST_IN_TRANSACTION)
            switch(len)
            {
            case 1:
                rc = r_trans(mf, cvalue, 1);
                break;
            case 2:
                rc = r_trans(mf, svalue, 2);
                break;
            case 4:
                rc = r_trans(mf, ivalue, 4);
                break;
            default:
                errno = EINVAL;
                return -1;
            }

        /*  End transaction (anyway, even it was an error)s */
        if (end_trans(mf) < 0)
            return -1;

        /*  Now check Gw status and repeat transaction if timeout */
        switch (rc)
        {
        case ST_IN_TRANSACTION:
            repeat = 0;
            break;
        case ST_TIMEOUT:
            if (++repeat > MAX_REPETITION)
                return -1;
            /* printf("i2c_master_read_cr: repeat:%d\n", repeat); */
            break;
        default:
            errno = EIO;
            return -1;
        }
    }

    /*  Convert output value */
    switch(len)
    {
    case 2:
        *svalue = __be16_to_cpu(*svalue);
        break;
    case 4:
        *ivalue = __be32_to_cpu(*ivalue);
        break;
    }

#ifdef HIGH_LEVEL_DEBUG
    printf("    i2c_master_read_cr: val:0x%x rc=%d\n", *ivalue, len);
#endif

    return len;
}
int i2c_master_read(mfile *mf, void *value, unsigned int offset, int len)
{
    unsigned int   *ivalue = (unsigned int *)value;
    unsigned short *svalue = (unsigned short *)value;
    unsigned char  *cvalue = (unsigned char *)value;
    int            rc=0, repeat=1;
    unsigned short offs = __cpu_to_be16(offset & 0xffff);
    unsigned char  off1 = offset & 0xff;

#ifdef HIGH_LEVEL_DEBUG
    printf("i2c_master_read_cr i2c:0x%x, offs:0x%x, len:%d\n",
           mf->i2c_slave, offset, len);
#endif

    offset = __cpu_to_be32(offset);

    while (repeat)
    {
        /*  Write address (1byte for DIMM, 2bytes for Gamla and 4bytes for Tavor) */
        switch(mf->dtype)
        {
        case MST_TAVOR:
            rc  = w_trans(mf, &offset, sizeof(offset));
            break;
        case MST_GAMLA:
            rc = w_trans(mf, &offs, sizeof(offs));
            break;
        case MST_DIMM:
            rc = w_trans(mf, &off1, sizeof(off1));
            break;
        case MST_NOADDR:
            rc = ST_IN_TRANSACTION;  /*  FAKE! No address transaction */
            break;
        }

        /*  Read data */
        if (rc == ST_IN_TRANSACTION)
            switch(len)
            {
            case 1:
                rc = r_trans(mf, cvalue, 1);
                break;
            case 2:
                rc = r_trans(mf, svalue, 2);
                break;
            case 3:
                rc = r_trans(mf, svalue, 3);
                break;
            case 4:
                rc = r_trans(mf, ivalue, 4);
                break;
            default:
                errno = EINVAL;
                return -1;
            }

        /*  End transaction (anyway, even it was an error)s */
        if (end_trans(mf) < 0)
            return -1;

        /*  Now check Gw status and repeat transaction if timeout */
        switch (rc)
        {
        case ST_IN_TRANSACTION:
            repeat = 0;
            break;
        case ST_TIMEOUT:
            if (++repeat > MAX_REPETITION)
                return -1;
            /* printf("i2c_master_read_cr: repeat:%d\n", repeat); */
            break;
        default:
            errno = EIO;
            return -1;
        }
    }

#ifdef HIGH_LEVEL_DEBUG
    printf("    i2c_master_read_cr: val:0x%x rc=%d\n", *ivalue, len);
#endif

    return len;
}

void i2c_master_set(mfile *mf)
{
    unsigned int val;
    int          i;

    if (mf->itype == MST_GAMLA)
    {
        /*  Screw up SPM clock. Works w/ Gamla only */
        mread4_(mf, SPM_OFFS + 12, &val);
        mf->i2c_RESERVED = val;
        val &= 0xffff0000;
        val |= 0x100;
        mwrite4_(mf, SPM_OFFS + 12, val);
    }

    /*  Tavor I2C bug workaround */
    if (mf->dtype == MST_TAVOR)
    {
        for (i=0; i<9; i++)
            end_trans(mf);

        i2c_master_read_cr(mf, &val, 0xf0014, 4);

        for (i=0; i<9; i++)
            end_trans(mf);
    }
    else
        i2c_master_read_cr(mf, &val, 0x2800, 4);
}

void i2c_master_restore(mfile *mf)
{
    if (mf->itype == MST_GAMLA)
        /*  Restore SPM clock. */
        mwrite4_(mf, SPM_OFFS + 12, mf->i2c_RESERVED);
}
