/*
 *
 *  mtcr_ib.c - Mellanox Software tools (mst) IB related driver implementation
 *
 */
#include <errno.h>

#include "mtcr.h"
#include "mtcr_ib.h"

#ifndef USE_IB_MGT
/*
 * Stabs for IB interface
 * ----------------------
 */

int mib_open(const char *name, mfile *mf)
{
    errno = ENOSYS;
    return -1;
}

int mib_close(mfile *mf)
{
    errno = ENOSYS;
    return -1;
}

int mib_read4(mfile *mf, unsigned int offset, u_int32_t *value)
{
    errno = ENOSYS;
    return -1;
}

int mib_write4(mfile *mf, unsigned int offset, u_int32_t value)
{
    errno = ENOSYS;
    return -1;
}

int mib_read64(mfile *mf, unsigned int offset, void *data, int length)
{
    errno = ENOSYS;
    return -1;
}

int mib_write64(mfile *mf, unsigned int offset, void *data, int length)
{
    errno = ENOSYS;
    return -1;
}

#else

/**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <asm/byteorder.h>

#define CPU_TO_BE32(b, c) do {      \
    u_int32_t *p = (u_int32_t*)(b); \
    int        n = (c);             \
    int i;                          \
    for (i=0; i<n; i++,p++)         \
        *p = __cpu_to_be32(*p);     \
    } while(0)
#define BE32_TO_CPU(b, c) do {      \
    u_int32_t *p = (u_int32_t*)(b); \
    int        n = (c);             \
    int i;                          \
    for (i=0; i<n; i++,p++)         \
        *p = __be32_to_cpu(*p);     \
    } while(0)

#ifdef MIB_DEBUG
#define RC_CHECK(s, r) do {                         \
    IB_MGT_ret_t  rc = (r);                         \
    if (rc != IB_MGT_OK) {                          \
        printf(s " error: %s\n", mtl_strerror(rc)); \
        errno = EIO;                                \
        return -1;                                  \
    }                                               \
    } while(0)
/***********
#define RC_CHECK(s, r) do {                         \
    IB_MGT_ret_t  rc;                               \
    printf("before " s " .. \n");                   \
    rc = r;                                         \
    printf("after " s " rc=%d\n\n", rc);            \
    if (rc != IB_MGT_OK) {                          \
        printf(s " error: %s\n", mtl_strerror(rc)); \
        errno = EIO;                                \
        return -1;                                  \
    }                                               \
    } while(0)
***********/
#else
#define RC_CHECK(s, r)
#endif

/*
 * Real IB interface
 * ----------------------
 */
static u_int64_t    wait_trid;
static volatile int wait_flag;
static u_int32_t    wait_value;

void mib_recv(IB_MGT_mad_hndl_t        mad_hndl,
              void*                    private_ctx_p,
              void*                    payload_p,
              IB_MGT_mad_rcv_desc_t    *rcv_remote_info_p)
{
    MAD_payload  p;
    memcpy(&p, payload_p, sizeof(MAD_payload));
    BE32_TO_CPU(&p, sizeof(MAD_payload)/4);
    if (p.r  &&  !p.Status  &&  p.TransactionID == wait_trid)
    {
        wait_value = p.Data[0];
        wait_flag = 1;
    }
}

int mib_open(const char *name, mfile *mf)
{
    char *endp;

    /* This is IB connection */
    mf->tp = MST_IB;
    mf->sock = -1;

    /* Initialize miscellaneous MIB_Private fields */
    mf->mib.trid = 1;

    /* determine target lid */
    mf->mib.udav.dlid = strtol(name, &endp, 10);
    if (*endp)
    {
        errno = ENXIO;
        return -1;
    }

    /* Prepare other udav fields */
    mf->mib.udav.sl            = 0;
    mf->mib.udav.src_path_bits = 0;
    mf->mib.udav.static_rate   = 0;
    mf->mib.udav.grh_flag      = 0;

    RC_CHECK("IB_MGT_get_handle",
             IB_MGT_get_handle(DEVICE_NAME, DEVICE_PORT, IB_MGT_GSI, &mf->mib.hndl));
    RC_CHECK("IB_MGT_bind_gsi_class",
             IB_MGT_bind_gsi_class(mf->mib.hndl, VENDOR_CLASS));
    RC_CHECK("IB_MGT_reg_cb",
             IB_MGT_reg_cb(mf->mib.hndl, mib_recv, 0, 0, 0, IB_MGT_RCV_CB_MASK));

    return 0;
}

int mib_close(mfile *mf)
{
    RC_CHECK("IB_MGT_reg_cb",
             IB_MGT_reg_cb(mf->mib.hndl, 0, 0, 0, 0, IB_MGT_RCV_CB_MASK));
    RC_CHECK("IB_MGT_unbind_gsi_class",
             IB_MGT_unbind_gsi_class(mf->mib.hndl, VENDOR_CLASS));
    RC_CHECK("IB_MGT_release_handle",
             IB_MGT_release_handle(mf->mib.hndl));
    return 0;
}

int mib_read4(mfile *mf, unsigned int offset, u_int32_t *value)
{
    int          cnt_repl, cnt_resend;
    MAD_payload  p;

    /* Prepare read request */
    memset(&p, 0, sizeof(MAD_payload));
    p.BaseVersion = 1;
    p.MgmtClass = VENDOR_CLASS;
    p.ClassVersion = 1;
    p.method = VENDOR_GET;
    p.TransactionID = mf->mib.trid;
    p.AttributeID = 0x50;
    p.Mode = 0;
    p.Count = 1;
    p.BurstAddress = offset;

    CPU_TO_BE32(&p, sizeof(MAD_payload)/4);
    for (cnt_resend=0; cnt_resend < RESEND_TRIES; cnt_resend++)
    {
        /* Send read request */
        wait_trid = mf->mib.trid++;
        wait_flag = 0;
        RC_CHECK("IB_MGT_send_mad",
                 IB_MGT_send_mad(mf->mib.hndl, &p, &mf->mib.udav, 0, SEND_TIMEOUT));

        /* Wait for response */
        for (cnt_repl=0; cnt_repl <= FAST_CHECKS + SLOW_CHECKS; cnt_repl++)
        {
            if (wait_flag)
            {
                *value = wait_value;
                return 4;
            }
            if (cnt_repl > FAST_CHECKS)
                usleep(SLOW_PAUSE);
        }
    }
    errno = ETIMEDOUT;
    return -1;
}

int mib_write4(mfile *mf, unsigned int offset, u_int32_t value)
{
    int          cnt_repl, cnt_resend;
    MAD_payload  p;

    /* Prepare read request */
    memset(&p, 0, sizeof(MAD_payload));
    p.BaseVersion = 1;
    p.MgmtClass = VENDOR_CLASS;
    p.ClassVersion = 1;
    p.method = VENDOR_SET;
    p.TransactionID = mf->mib.trid;
    p.AttributeID = 0x50;
    p.Mode = 0;
    p.Count = 1;
    p.BurstAddress = offset;
    p.Data[0] = value;

    CPU_TO_BE32(&p, sizeof(MAD_payload)/4);
    for (cnt_resend=0; cnt_resend < RESEND_TRIES; cnt_resend++)
    {
        /* Send read request */
        wait_trid = mf->mib.trid++;
        wait_flag = 0;
        RC_CHECK("IB_MGT_send_mad",
                 IB_MGT_send_mad(mf->mib.hndl, &p, &mf->mib.udav, 0, SEND_TIMEOUT));

        /* Wait for response */
        for (cnt_repl=0; cnt_repl <= FAST_CHECKS + SLOW_CHECKS; cnt_repl++)
        {
            if (wait_flag)
            {
                if (value == wait_value)
                    return 4;
                else
                {
                    errno = EIO;
                    return -1;
                }
            }
            if (cnt_repl > FAST_CHECKS)
                usleep(SLOW_PAUSE);
        }
    }
    errno = ETIMEDOUT;
    return -1;
}

int mib_read64(mfile *mf, unsigned int offset, void *data, int length)
{
    errno = ENOSYS;
    return -1;
}

int mib_write64(mfile *mf, unsigned int offset, void *data, int length)
{
    errno = ENOSYS;
    return -1;
}

#endif
