#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>

#include "mtcr.h"

/*  Linux USB Support */
#include <linux/types.h>
#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

#define DIMAX_MST_DEV   "/dev/mst/mtusb-1"
#define DIMAX_IN_EP     0x84
#define DIMAX_OUT_EP    0x02
#define DIMAX_FW_MAJOR  2
#define DIMAX_FW_MINOR  8
#define DIMAX_IF        0x00

static int bulk_transaction( int fd, char cmd, int out_len, int in_len, char* buf );
static void print_line_status( unsigned char status, char* line );


int main( int argc, char*argv[] )
{
    mfile* mf;
    char dev_name[40] = DIMAX_MST_DEV;
    char buf[300];
    int rc, i, tmp;

       
    // Open device via mtcr
    mf = mopen(  dev_name );
    if( ! mf )
    {
        if( ! errno ) errno =-1;
        printf( "Can't open \"%s\" (%s)\n", dev_name, strerror(errno) );
        return errno;
    }

    /*  Claim Interface */
    tmp = DIMAX_IF;
    rc = ioctl( mf->fd, USBDEVFS_CLAIMINTERFACE, &tmp );
    if( rc )
    {
        rc = errno;
        printf( "CLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", rc, strerror(rc) );
        return -1;
    }

    // Test1 Check FW version
    printf( "Check FW Version             : " );
    rc = bulk_transaction( mf->fd, 0x0a, 1, 3, buf );
    if( rc ) return rc;
    if( buf[0] < DIMAX_FW_MAJOR )
    {
        printf( "Version is %d.%d but must be at least %d.0\n",
                    (int)buf[0], (int)buf[1], (int)DIMAX_FW_MAJOR );
        return -1;
    }
    printf( " %d.%d ", (int)buf[0], (int)buf[1] );
    printf( "PASSED\n" );

    // Seconf claim Interface
    tmp = DIMAX_IF;
    rc = ioctl( mf->fd, USBDEVFS_RELEASEINTERFACE, &i ); 
    if( rc )
    {
        rc = errno;
        printf( "RELEASEINTERFACE ioctl error. rc=%d, \"%s\"\n", rc, strerror(rc) );
        return -1;
    }
    tmp = DIMAX_IF;
    rc = ioctl( mf->fd, USBDEVFS_CLAIMINTERFACE, &tmp );
    if( rc )
    {
        rc = errno;
        printf( "Second CLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", rc, strerror(rc) );
        return -1;
    }

    // Test SDA, SCL Idle status
    printf( "Please disconnect I2C cable from target device. Press Enter... " );
    getchar();
    printf( "\rSDA Lines Idle status test   : " );
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "FAILED\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    printf( "PASSED\n" );

    printf( "\rSCL Lines Idle status test   : " );
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "FAILED\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    printf( "PASSED\n" );
   
 
    // Test SDA, SCL High/Low
    printf( "\rSDA HI/LOW transition test   : " );
    rc = bulk_transaction( mf->fd, 0x06, 1, 1, buf );   // Drop SDA
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0x00 )
    {
        printf( "SDA LOW problem\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    rc = bulk_transaction( mf->fd, 0x04, 1, 1, buf );   // Release SDA
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SDA HIGH problem\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    printf( "PASSED\n" );
    
    printf( "\rSCL HI/LOW transition test   : " );
    rc = bulk_transaction( mf->fd, 0x07, 1, 1, buf );   // Drop SCL
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0x00 )
    {
        printf( "SCL LOW problem\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    rc = bulk_transaction( mf->fd, 0x05, 1, 1, buf );   // Release SCL
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SCL HIGH problem\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    printf( "PASSED\n" );
    
    
    // Test SDA, SCL Target status
    printf( "Please Connect I2C cable to target device. Press Enter... " );
    getchar();
    printf( "\rSDA Line Target test         : " );
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SDA Target Line Idle problem\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    printf( "PASSED\n" );

    printf( "\rSCL Line Target test         : " );
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SCL Target Line Idle problem\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    printf( "PASSED\n" );
   
 
    // Test Target SDA, SCL High/Low
    printf( "\rSDA HI/LOW Target test       : " );
    rc = bulk_transaction( mf->fd, 0x06, 1, 1, buf );   // Drop SDA
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0x00 )
    {
        printf( "SDA Target LOW problem\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    rc = bulk_transaction( mf->fd, 0x04, 1, 1, buf );   // Release SDA
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x08, 1, 2, buf );   // Read SDA
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SDA Target HIGH problem\n" );
        print_line_status( buf[0], "SDA" );
        return -1;
    }
    printf( "PASSED\n" );

    printf( "\rSCL HI/LOW Target test       : " );
    rc = bulk_transaction( mf->fd, 0x07, 1, 1, buf );   // Drop SCL
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0x00 )
    {
        printf( "SCL Target LOW problem\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    rc = bulk_transaction( mf->fd, 0x05, 1, 1, buf );   // Release SCL
    if( rc ) return rc;
    rc = bulk_transaction( mf->fd, 0x09, 1, 2, buf );   // Read SCL
    if( rc ) return rc;
    if( buf[0] != (char)0xC0 )
    {
        printf( "SCL Target HIGH problem\n" );
        print_line_status( buf[0], "SCL" );
        return -1;
    }
    printf( "PASSED\n" );

    // Send stop to target
    printf( "\rSend I2C STOP to Target      : " );
    rc = bulk_transaction( mf->fd, 0x0d, 1, 1, buf );   // I2C Stop
    if( rc ) return rc;
    printf( "PASSED\n" );

    // Detect all present slave devices
    printf( "\rI2C Slave devices Scan. Found devices: " );
    rc = bulk_transaction( mf->fd, 0x03, 1, sizeof(buf), buf );   // I2C Scan
    for( i=0; i<rc-1; i++ )
        printf( "0x%02x ", (int)((buf[i]>>1)&0x7F) );
    printf( "\n" );
     
    printf( "Test PASSED\n" );
    return 0;
}
     
static int bulk_transaction( int fd, char cmd, int out_len, int in_len, char* buf )
{
    int rc;
    struct usbdevfs_bulktransfer bulk_st;

    errno=0;

    buf[0]=cmd;
    bulk_st.data = buf;
    bulk_st.ep = DIMAX_OUT_EP;
    bulk_st.len = out_len;
    bulk_st.timeout = 100;  /*  ms */
    rc = ioctl( fd, USBDEVFS_BULK, &bulk_st );
    if( rc != out_len )
    {
        if( ! errno ) errno=-1;
        printf( "BULK OUT error (\"%s\") rc=%d\n", strerror(errno), rc );
        return errno;
    }
    bulk_st.data = buf;
    bulk_st.ep = DIMAX_IN_EP;
    bulk_st.len = in_len;
    bulk_st.timeout = 100;  /*  ms */
    rc = ioctl( fd, USBDEVFS_BULK, &bulk_st );
    if( rc != in_len || errno )
    {
        if( cmd == 3 && !errno && rc>0 )
            return rc;  // Special case. Reply size unknown before
        if( ! errno ) errno=-1;
        printf( "BULK IN error (\"%s\") rc=%d\n", strerror(errno), rc );
        return errno;
    }
    return 0;
}

static void print_line_status( unsigned char status, char* line )
{
    switch( status )
    {
    case 0x00:
        printf( "%s Line is dropped by MTUSB-1 signal is LOW\n", line );
        break;
    case 0x80:
        printf( "%s Line is dropped by MTUSB-1 but signal is HIGH\n", line );
        break;
    case 0x40:
        printf( "%s Line is released by MTUSB-1 but signal is LOW\n", line );
        break;
    case 0xCC:
        printf( "%s Line is released by MTUSB-1 signal is HIGH\n", line );
        break;
    default:
        printf( "%s Line has unknown status=0x%x\n", line, (unsigned int)(status) );
    }
}

