#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h> /*This header likes to sit before linux/ headers*/
#include <linux/types.h>
#include <linux/usb.h>
#include <linux/usbdevice_fs.h>
#include <asm/byteorder.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>

#include "usbioctl.h"

#if 0
struct usbdevfs_ctrltransfer {
	__u8 requesttype;
	__u8 request;
	__u16 value;
	__u16 index;
	__u16 length;
	__u32 timeout;  /* in milliseconds */
 	void *data;
};
#endif

/* USB 2.0: bmRequestType - Characteristics of request: 
 * D7: Data transfer direction 0 = Host-to-device 1 = Device-to-host 
 * D6...5: Type 0 = Standard 1 = Class 2 = Vendor 3 = Reserved 
 * D4...0: Recipient 0 = Device 1 = Interface 2 = Endpoint 3 = Other 4...31 = Reserved
 * */

#define WRITE_I2C (USB_DIR_OUT|USB_TYPE_VENDOR)
#define READ_I2C (USB_DIR_IN|USB_TYPE_VENDOR)
#define PORTS_REQUEST 0xF1
#define I2C_REQUEST 0xF2
#define DEBUG_REQUEST 0xFF

int mst_WriteI2c(int fd, PI2C_TRANS tr) {
  struct usbdevfs_ctrltransfer ctrl={
     WRITE_I2C,     /* requesttype */
     I2C_REQUEST,   /* request */
     tr->byTransType << 0 | tr->bySlvDevAddr << 8, /* value: bytes 2 and 3 -LE? */
     tr->wMemoryAddr,/* index */
     tr->wCount,    /* length */
     5000,          /* timeout */
     tr->Data,      /* data */
  };
  return ioctl(fd,USBDEVFS_CONTROL, &ctrl);
}
int mst_ReadIoPorts(int fd, PI2C_TRANS tr) {
  struct usbdevfs_ctrltransfer ctrl={
     READ_I2C,      /* requesttype */
     PORTS_REQUEST, /* request */
     0,             /* value */
     0,             /* index */
     4,             /* length */
     5000,          /* timeout */
     tr->Data,      /* data */
  };
  return ioctl(fd,USBDEVFS_CONTROL, &ctrl);
}
int mst_WriteIoPorts(int fd, PI2C_TRANS tr) {
  struct usbdevfs_ctrltransfer ctrl={
     WRITE_I2C,     /* requesttype */
     PORTS_REQUEST, /* request */
     0,             /* value */
     0,             /* index */
     8,             /* length */
     5000,          /* timeout */
     tr->Data,      /* data */
  };
  return ioctl(fd,USBDEVFS_CONTROL, &ctrl);
}

int mst_ReadI2c(int fd, PI2C_TRANS tr) {
  struct usbdevfs_ctrltransfer ctrl={
     READ_I2C,      /* requesttype */
     I2C_REQUEST,   /* request */
     tr->byTransType << 0 | tr->bySlvDevAddr << 8,/* value:bytes 2 and 3 -LE? */
     tr->wMemoryAddr,/* index */
     tr->wCount,    /* length */
     5000,          /* timeout */
     tr->Data,      /* data */
  };
  return ioctl(fd,USBDEVFS_CONTROL, &ctrl);
}

int mst_Debug(int fd, PI2C_TRANS tr) {
  struct usbdevfs_ctrltransfer ctrl={
     READ_I2C,      /* requesttype */
     DEBUG_REQUEST, /* request */
     0,             /* value */
     0,             /* index */
     tr->wCount,    /* length */
     5000,          /* timeout */
     tr->Data,      /* data */
  };
  return ioctl(fd,USBDEVFS_CONTROL, &ctrl);
}


/*  DiMax support */

#define dimax_debug	0
#define DIMAX_IN_EP		0x84
#define DIMAX_OUT_EP	0x02
#define DIMAX_IF   		0x00


/*
Input:
   size  - read data size
   tr->wCount - address size
   tr->Data - memory address
Output:
   0-ok, else error code
*/   
int dimax_ReadI2c( int fd, PI2C_TRANS tr, int size )
{
/*
Read transaction Format ( non zero address )
1 : Command = 1
1 : M - Addr Size
1 : wsa - Write Slave Address
M : memory address M bytes
1 : rsa - Read Slave Address
1 : N - read data size

Read transaction Format ( zero address )
1 : Command = 1
1 : Addr Size = 0
1 : rsa - Read Slave Address
1 : N - read data size

Read Reply Format
N : Read Data
1 : Status, 0-ok, else error
*/

	unsigned char cbuf[12];
	unsigned int i,sz=0;
	struct usbdevfs_bulktransfer bulk_st;
	int rc=0,sent,recv=0;
    sigset_t set, oldset; 

	/*  Fill command */
	if( tr->wCount )
	{
		cbuf[sz++] = 1;								/*  Read Command Code */
		cbuf[sz++] = tr->wCount;					/*  Address Size */
		cbuf[sz++] = tr->bySlvDevAddr & 0xFE;		/*  Device Slave Address for Write */
		memcpy( cbuf+sz, tr->Data, tr->wCount );    /*  Memory address */
		sz+=tr->wCount;
		cbuf[sz++] = tr->bySlvDevAddr|0x01; 		/*  Device Slave Address for Read */
		cbuf[sz++] = size; 							/*  Read Size */
	}
	else
	{
		cbuf[sz++] = 1;
		cbuf[sz++] = 0;
		cbuf[sz++] = tr->bySlvDevAddr|0x01;
		cbuf[sz++] = size;
	}

	/*  Claim Interface */
	i = DIMAX_IF;
	rc = ioctl( fd, USBDEVFS_CLAIMINTERFACE, &i );
	if( rc )
	{
		rc = errno;
		if( dimax_debug )
            printf( "CLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", rc, strerror(rc) );
		return rc;
	}

    /* Block signals to make transaction atomic */
    memset( &set, 0xFF, sizeof(sigset_t) );
    sigprocmask(SIG_BLOCK, &set, &oldset);

	/*  OUT Command */
    bulk_st.data = cbuf;
	bulk_st.ep = DIMAX_OUT_EP;
	bulk_st.len = sz;
	bulk_st.timeout = 100;			/*  ms */

	sent = ioctl( fd, USBDEVFS_BULK, &bulk_st );
	if( sent != sz )
	{
		rc = errno;
        if( !rc ) rc=-1; 
		if( dimax_debug )
			printf( "%s : BULK ioctl OUT error. rc=%d, \"%s\"\n", __func__, rc, strerror(rc) );
		goto cleanup;
	}

	/*  Bulk IN. Read Result */
	bulk_st.ep = DIMAX_IN_EP;
	bulk_st.len = size+1;
	bulk_st.timeout = 100;			/*  ms */
	bulk_st.data = tr->Data;

	recv = ioctl( fd, USBDEVFS_BULK, &bulk_st );
	if( recv != size+1 )
	{
		rc = errno;
        if( !rc ) rc=-1; 
		if( dimax_debug )
			printf( "%s : BULK ioctl IN error. rc=%d, \"%s\"\n", __func__, rc, strerror(rc) );
		goto cleanup;
	}
	/*  Check Status */
	if( tr->Data[size] != 0 )
	{
		rc = 200+tr->Data[size];
		if( dimax_debug )
			printf( "%s : Status error %d\n", __func__, tr->Data[size] );
		goto cleanup;
	}
	rc=0; 

cleanup:
    /* Unblock signals which was not blocked */
    for( i=0; i<sizeof(sigset_t); i++ )
        ((char*)(&set))[i]=~((char*)(&oldset))[i];
    sigprocmask(SIG_UNBLOCK, &set, &oldset);

	/*  Unclaim Interface */
	i = DIMAX_IF;
	if( ioctl( fd, USBDEVFS_RELEASEINTERFACE, &i ) )
	{
		if( dimax_debug )
			printf( "UNCLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", errno, strerror(errno) );
	}

	if( dimax_debug )
	{
		printf( "%s : BULK ioctl OUT %d bytes sent\n", __func__, sent );
		for( i=0; i<sz; i++ )
			printf( "0x%02x ", cbuf[i] );
		printf( "\n" );
		if( rc==0 )
		{
			printf( "%s: BULK ioctl IN %d bytes received\n", __func__, recv ); 
			if( recv > 0 )
                for( i=0; i<recv; i++ )
                    printf( "0x%02x ", tr->Data[i] );
			printf( "\n" );
		}
	}
	return rc;
}

/*
Input:
   size  - write data size
   tr->wCount - address size
   tr->Data - memory address followed by write data
Output
   0-ok, else error code
*/   
int dimax_WriteI2c( int fd, PI2C_TRANS tr, int size )
{
/*
Write transaction Format ( non zero address )
1 : Command = 2
1 : M - Addr Size
1 : wsa - Write Slave Address
M : memory address M bytes
1 : N - write data size (up to 63 bytes)
N : Data

Read transaction Format ( zero address )
1 : Command = 2
1 : Addr Size = 0
1 : wsa - Write Slave Address
1 : N - write data size (up to 63 bytes)
N : Data

Write Reply Format
1 : Status, 0-ok, else error
*/
	unsigned char cbuf[80];
	unsigned int i,sz=0;
	struct usbdevfs_bulktransfer bulk_st;
	int rc=0,sent,recv=0;
    sigset_t set, oldset; 

	/*  Fill command */
	if( tr->wCount )
	{
		cbuf[sz++] = 2;								/*  Write Command Code */
		cbuf[sz++] = tr->wCount;					/*  Address Size */
		cbuf[sz++] = tr->bySlvDevAddr & 0xFE;		/*  Device Slave Address for Write */
		memcpy( cbuf+sz, tr->Data, tr->wCount );    /*  memory address */
		sz+=tr->wCount;
		cbuf[sz++] = size; 								/*  Write Size */
		memcpy( cbuf+sz, tr->Data+tr->wCount, size );	/*  Data */
		sz+=size;
	}
	else
	{
		cbuf[sz++] = 2;		
		cbuf[sz++] = 0;		
		cbuf[sz++] = tr->bySlvDevAddr & 0xFE;
		cbuf[sz++] = size;
		memcpy( cbuf+sz, tr->Data, size );
		sz+=size;
	}

	/*  Claim Interface */
	i = DIMAX_IF;
	rc = ioctl( fd, USBDEVFS_CLAIMINTERFACE, &i );
	if( rc )
	{
		rc = errno;
		if( dimax_debug )
            printf( "CLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", rc, strerror(rc) );
		return rc;
	}

    /* Block signals to make transaction atomic */
    memset( &set, 0xFF, sizeof(sigset_t) );
    sigprocmask(SIG_BLOCK, &set, &oldset);

	/*  OUT Command */
    bulk_st.data = cbuf;
	bulk_st.ep = DIMAX_OUT_EP;
	bulk_st.len = sz;
	bulk_st.timeout = 100;			/*  ms */

	sent = ioctl( fd, USBDEVFS_BULK, &bulk_st );
	if( sent != sz )
	{
		rc = errno;
        if( !rc ) rc=-1; 
		if( dimax_debug )
			printf( "%s: BULK ioctl OUT error. rc=%d, \"%s\"\n", __func__, rc, strerror(rc) );
		goto cleanup;
	}

	/*  Bulk IN. Read Result */
	bulk_st.ep = DIMAX_IN_EP;
	bulk_st.len = 1;
	bulk_st.timeout = 100;			/*  ms */
	bulk_st.data = tr->Data;

	recv = ioctl( fd, USBDEVFS_BULK, &bulk_st );
	if( recv != 1 )
	{
		rc = errno;
        if( !rc ) rc=-1; 
		if( dimax_debug )
			printf( "%s: BULK ioctl IN error. rc=%d, \"%s\"\n", __func__, rc, strerror(rc) );
		goto cleanup;
	}
	/*  Check Status */
	if( tr->Data[0] != 0 )
	{
		rc = 200+tr->Data[size];
		if( dimax_debug )
			printf( "%s: Status error %d\n", __func__, tr->Data[0] );
		goto cleanup;
	}
	rc=0; 

cleanup:
    /* Unblock signals which was not blocked */
    for( i=0; i<sizeof(sigset_t); i++ )
        ((char*)(&set))[i]=~((char*)(&oldset))[i];
    sigprocmask(SIG_UNBLOCK, &set, &oldset);

	/*  Unclaim Interface */
	i = DIMAX_IF;
	if( ioctl( fd, USBDEVFS_RELEASEINTERFACE, &i ) )
	{
		if( dimax_debug )
			printf( "UNCLAIMINTERFACE ioctl error. rc=%d, \"%s\"\n", errno, strerror(errno) );
	}

	if( dimax_debug )
	{
		printf( "%s: BULK ioctl OUT %d bytes sent\n", __func__, sent );
		for( i=0; i<sz; i++ )
			printf( "0x%02x ", cbuf[i] );
		printf( "\n" );
		if( rc==0 )
		{
			printf( "%s: BULK ioctl IN %d bytes received\n", __func__, recv ); 
			if( recv > 0 )
                for( i=0; i<recv; i++ )
                    printf( "0x%02x ", tr->Data[i] );
			printf( "\n" );
		}
	}
	return rc;
}

#ifdef MAIN
int main(int argc, char** argv)
{
  int fd;
  int ret;
  unsigned int buf=0;
  unsigned int slave;
  /* unsigned int mask=-1; */
  I2C_TRANS rtrans={
    byTransType:I2C_TRANS_16ADR,
    bySlvDevAddr:0x5a<<1, /* Gamla, fix for Tavor */
    wMemoryAddr:0,
    wCount:4,
    Data:{},
  };
#ifdef DOWRITE
  I2C_TRANS wtrans={
    byTransType:I2C_TRANS_16ADR,
    bySlvDevAddr:0x5a<<1, /* Gamla, fix for Tavor */
    wMemoryAddr:_be16(0x2800),
    wCount:4,
    Data:{0,28,0,0},
  };
#endif
  I2C_TRANS dtrans={
    wCount:64,
    Data:{},
  };

  fd=open(argv[1],O_WRONLY);
  perror("OPEN");

  sscanf(argv[2]+2,"%x",&slave);
  {
    unsigned addr;
  sscanf(argv[3]+2,"%x",&addr);
  rtrans.wMemoryAddr=addr;

  }

#ifdef DOWRITE
  ret=mst_WriteI2c(fd,&wtrans);
  printf("fd=%d, ret=%d\n",fd, ret);
  perror("WRITE");
#endif

  rtrans.bySlvDevAddr=slave;
  printf("SLAVE:0x%x\n", slave);
  printf("ADDR:0x%x\n", rtrans.wMemoryAddr);

#if 0
  ret=mst_ReadIoPorts(fd,&rtrans);
  memcpy(&buf,rtrans.Data,4);
  printf("fd=%d, ret=%d, result=%x\n",fd, ret, buf);
  perror("READ PORTS");

  buf=0x000bff0f;
  memcpy(rtrans.Data,&buf,4);
  memcpy(rtrans.Data+4,&mask,4);
  ret=mst_WriteIoPorts(fd,&rtrans);
  printf("fd=%d, ret=%d, result=%x\n",fd, ret, buf);
  perror("WRITE PORTS");

  ret=mst_ReadIoPorts(fd,&rtrans);
  memcpy(&buf,rtrans.Data,4);
  printf("fd=%d, ret=%d, result=%x\n",fd, ret, buf);
  perror("READ PORTS2");
#endif

  ret=mst_ReadI2c(fd,&rtrans);
  memcpy(&buf,rtrans.Data,4);
  printf("fd=%d, ret=%d, result=%x\n",fd, ret, buf);
  perror("READ");

  printf("DEBUG>\n");
  do {
  ret=mst_Debug(fd,&dtrans);
  dtrans.Data[ret]=0;
  if (ret >0) printf("%64s",dtrans.Data);
  } while (ret == 64);
  printf("<DEBUG\n");


  return ret;
}
#endif
