/*
 * Copyright (C) Voltaire Ltd. 2001-2003.  ALL RIGHTS RESERVED.
 *
 * This source code (the "Code") is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest in and to the Code,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * The Code is provided "as is". To the maximum extent permitted by law,
 * the Company disclaims any warranties of any kind, either expressed or
 * implied, including, without limitation, implied warranties of
 * merchantability, fitness for a particular purpose and non-infringement.
 * The Company does not warrant that the Code will operate error free, or
 * in an uninterrupted fashion or that it is compatible with any particular
 * platform. In no event shall the Company be liable for any damages
 * whatsoever (including, without limitation, incidental, direct, indirect,
 * special and consequential damages, damages for loss of business profits,
 * business interruption, loss of business information, or other pecuniary
 * loss) arising out of the use or inability to use the Code, even if
 * advised of the possibility of such damages.
 */

#ifndef __GSI_H__
#define __GSI_H__

//#include "ib-types.h"
#include "mad.h"
#include "class-port-info.h"

#define GSI_POOL_TRACE

#define GSI_DTGRM_POOL_MAX_SIZE   5000

#define GSI_QP1_WELL_KNOWN_Q_KEY  0x80010000

#ifndef  ARCH_64_BIT  // if not 32 BIT 
#define GSI_DTGRM_PRIV_HDR_SIZE    104
#else
#define GSI_DTGRM_PRIV_HDR_SIZE    104
#endif

typedef void *gsi_class_handle_t;
typedef void *gsi_dtgrm_pool_handle_t;

#define       IB_GRH_LEN  40

//
// A datagram element structure
// including the MAD itself, address information and other.
//
typedef struct gsi_dtgrm_st{
        // The private header used in gsi.c
        // ( scatter-gather, pool handler etc.)
        char                    priv[GSI_DTGRM_PRIV_HDR_SIZE];

        // Management datagram
        char                    mad[MAD_BLOCK_SIZE]; 
        //
        // For SW cache flushing/invaliation align the MAD buffer's end
        // on 32 bytes, to prevent following data corruption
        char                    cache_pad[(32 - ((sizeof(mad_t)+IB_GRH_LEN)%32))%32];

        void                    *rmpp_payload;
        int                     rmpp_payload_size;
        void                    *rmpp_context;
        int                     rmpp_dir_switch_needed;

        vv_qp_t                 rqp;
        vv_q_key_t              rqk;
        vv_lid_t                rlid;
        u_int8_t                sl;
        int                     pkey_idx;
        
        vv_guid_t               guid;
        u_int8_t                path_bits;
        u_int8_t                static_rate;

        u_int8_t                is_marked_for_release;
        u_int8_t                is_in_use;			   

} __attribute__((packed)) gsi_dtgrm_t;




typedef struct gsi_redirect_mad_st
{
        mad_hdr_t         hdr;
        u_int64_t         m_key;
        char              reserved[32];
        ib_class_port_info_t class_port_info;
} __attribute__((packed))  gsi_redirect_mad_t;

//
// Function return codes:
//
// vv_return_ok (zero)
// vv_return_err - negative value


typedef void (*gsi_receive_cb_t) ( IN gsi_class_handle_t      handle,
                                   IN void                    *context,
                                   IN gsi_dtgrm_t             *dtgrm);
typedef void (*gsi_send_compl_cb_t)    ( IN gsi_class_handle_t      handle,
                                         IN void                    *context,
                                         IN gsi_dtgrm_t             *dtgrm);

//
// Register GSI class service
//
// Parameters:
//    mgt_class    -  class ID,
//    version  -  class version,
//    hca_name -  HCA name like "InfiniHost0"
//    port     -  HCA port number,
//    server   -  boolean value indicating that the registered service is
//                an Agent (TRUE) or a Manager (FALSE)
//    RMPP     -  boolean value indicating that the registered service
//                supports RMPP (Relable Multi-Packet Protocol)
//    send_compl_cb - user's send completion callback,
//    receive_cb    - user's receive callback,
//    context       - user's context,
//    handle   -  the class service handle returned to the user

int gsi_register_class(IN u_int8_t                  mgt_class,
                       IN u_int8_t                  version,
                       IN char                      *hca_name,
                       IN u_int                     port,
                       IN u_int8_t                  server,
                       IN u_int8_t                  rmpp,
                       IN gsi_send_compl_cb_t       send_compl_cb,
                       IN gsi_receive_cb_t          receive_cb,
                       IN void                      *context,
                       OUT gsi_class_handle_t       *handle );

//
// Deregister GSI class service
//
int gsi_deregister_class(IN gsi_class_handle_t handle);


//
// Create a datagram pool (see gsi_dtgrm_t )
//
#ifdef GSI_POOL_TRACE
int gsi_dtgrm_pool_create_named( IN     u_int32_t    cnt,
                           OUT    gsi_dtgrm_pool_handle_t     *handle, 
                           IN     char                        *modname
                         );
#define gsi_dtgrm_pool_create(cnt, handle)  gsi_dtgrm_pool_create_named((cnt), (handle), (char *)((THIS_MODULE)->name)) 
#else
int gsi_dtgrm_pool_create( IN     u_int32_t    cnt,
                           OUT    gsi_dtgrm_pool_handle_t     *handle )
#endif
//
// Destroy datagram pool
//
void gsi_dtgrm_pool_destroy( IN   gsi_dtgrm_pool_handle_t     pool_handle );


//
// Get a datagram element from a pool (see gsi_dtgrm_t ) 
// 
int  gsi_dtgrm_pool_get( IN      gsi_dtgrm_pool_handle_t    pool_handle,
                         OUT     gsi_dtgrm_t                 **dtgrm);
 
//
// Return datagram element to pool
//
int  gsi_dtgrm_pool_put( IN     gsi_dtgrm_t          *dtgrm);

//
// Send a datagram
//
// Before calling gsi_post_send_dtgrm() for sending a non-RMPP datagram:
//   1. The user must fill some fields in the general MAD header --
//      the method and response bit,
//   2. The user must fill the address information in the datagram element,
//   3. The user must fill the MAD payload,
//
// For RMPP -- before calling gsi_post_send_dtgrm():
//   1. The user must fill some fields in the  general MAD header like
//      in a regular case --
//      the method and response bit,
//   2. The user must fill the address information in the datagram element,
//   3. The user must allocate and fill the RMPP buffer with the payload
//      (see gsi_dtgrm_alloc_rmpp_buf() description)
//   4. The user shall not fill the MAD payload,
//   5. The user does not need to fill any RMPP related fields in the MAD -
//      the class service will do it automatically,
//   6. If an RMPP response is expected, the user must set
//      'rmpp_dir_switch_needed' field in the datagram element structure to be TRUE.
//
// 
int gsi_post_send_dtgrm( IN gsi_class_handle_t handle, IN gsi_dtgrm_t *dtgrm);



// Allocate a buffer for RMMP (Relable Multi-Packet Protocol) payload.
// The buffer does not include general MAD header or RMMP header.
// Allocating the buffer automatically indicates that the datagram must be sent
// using RMPP protocol.
// The function shall not be used for non-RMPP management classes.
void *gsi_dtgrm_alloc_rmpp_buf( IN gsi_dtgrm_t *dtgrm,IN  int size );
 
// Free the RMPP buffer.
// Note - calling gsi_dtgrm_pool_put() releases the buffer automatiacally.
void gsi_dtgrm_free_rmpp_buf( IN gsi_dtgrm_t *dtgrm );
 
// Get the buffer pointer and size.
void *gsi_dtgrm_get_rmpp_buf( IN gsi_dtgrm_t *dtgrm, OUT int *size );

#endif
