/*
 * Copyright (C) Voltaire Ltd. 2001-2003.  ALL RIGHTS RESERVED.
 *
 * This source code (the "Code") is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest in and to the Code,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * The Code is provided "as is". To the maximum extent permitted by law,
 * the Company disclaims any warranties of any kind, either expressed or
 * implied, including, without limitation, implied warranties of
 * merchantability, fitness for a particular purpose and non-infringement.
 * The Company does not warrant that the Code will operate error free, or
 * in an uninterrupted fashion or that it is compatible with any particular
 * platform. In no event shall the Company be liable for any damages
 * whatsoever (including, without limitation, incidental, direct, indirect,
 * special and consequential damages, damages for loss of business profits,
 * business interruption, loss of business information, or other pecuniary
 * loss) arising out of the use or inability to use the Code, even if
 * advised of the possibility of such damages.
 */

/* user level GSI headers. To be merged with kernel level headers */
#ifndef _UGSI_H
#define _UGSI_H

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

#define IN
#define OUT

#define GSI_DTGRM_POOL_MAX_SIZE   5000
#define GSI_QP1_WELL_KNOWN_Q_KEY  0x80010000

typedef void *ugsi_dtgrm_pool_handle_t;

typedef struct ugsi_dtgrm_st{
        // Management datagram
        char                    mad[MAD_BLOCK_SIZE]; 

        void                    *rmpp_payload;
        int                     rmpp_payload_size;
        void                    *rmpp_context;
        int                     rmpp_dir_switch_needed;

        vv_qp_t                 rqp;
        vv_q_key_t              rqk;
        vv_lid_t                rlid;
        u_int8_t                sl;
        
        vv_guid_t               guid;
        u_int8_t                path_bits;
        u_int8_t                static_rate;

        u_int8_t                is_marked_for_release;
        u_int8_t                is_in_use;			   

	/*
	 * handle of the kernel buffer. If not 0 then user uses
	 * manual allocation of the datagram buffers
	 */
	unsigned long khandle;     // set to preallocated kernel buffer handle

        ugsi_dtgrm_pool_handle_t pool_handle;
	/* 
	 * returned by the kernel via read() if mux mode is on and is set to 
	 * the address of the user mode buffer. It signifies a completion
	 * of the write() event. 
	 */
	unsigned long           uhandle;     

} __attribute__((packed)) ugsi_dtgrm_t;


typedef struct ugsi_class_handle_st *ugsi_class_handle_t;

typedef void (*ugsi_receive_cb_t) (ugsi_class_handle_t      handle,
		void                    *context,
		ugsi_dtgrm_t             *dtgrm);

typedef void (*ugsi_send_compl_cb_t)    (ugsi_class_handle_t      handle,
		void                    *context,
		ugsi_dtgrm_t             *dtgrm);


struct ugsi_class_handle_st {
	int fd;  /* file descriptor */
	ugsi_receive_cb_t rx_cb;
	ugsi_send_compl_cb_t tx_cb;
};



// Register GSI class service
//
// Parameters:
//    mgt_class-  class ID,
//    version  -  class version,
//    hca_name -  HCA name like "InfiniHost0"
//    port     -  HCA port number,
//    server   -  boolean value indicating that the registered service is
//                an Agent (TRUE) or a Manager (FALSE)
//    RMPP     -  boolean value indicating that the registered service
//                supports RMPP (Relable Multi-Packet Protocol)
//    send_compl_cb - user's send completion callback,
//    receive_cb    - user's receive callback,
//    context       - user's context,
//    handle   -  the class service handle returned to the user

int ugsi_register_class(IN u_int8_t                  mgt_class,
                       IN u_int8_t                  version,
                       IN char                      *hca_name,
                       IN u_int                     port,
                       IN u_int8_t                  server,
                       IN u_int8_t                  rmpp,
                       IN ugsi_send_compl_cb_t       send_compl_cb,
                       IN ugsi_receive_cb_t          receive_cb,
                       IN void                      *context,
                       OUT ugsi_class_handle_t       *handle );

//
// Deregister GSI class service
//
int ugsi_deregister_class(IN ugsi_class_handle_t handle);


//
// Create a datagram pool (see ugsi_dtgrm_t )
//
int ugsi_dtgrm_pool_create( IN     u_int32_t                     cnt,
                            OUT    ugsi_dtgrm_pool_handle_t     *handle);
 
//
// Destroy datagram pool
//
void ugsi_dtgrm_pool_destroy(IN   ugsi_dtgrm_pool_handle_t     handle );
 
//
// Get a datagram element from a pool (see gsi_dtgrm_t ) 
// 
int  ugsi_dtgrm_pool_get(IN      ugsi_dtgrm_pool_handle_t   handle,
                         OUT     ugsi_dtgrm_t                 **dtgrm);
 
//
// Return datagram element to pool
//
int  ugsi_dtgrm_pool_put( IN     ugsi_dtgrm_t          *dtgrm);

//
// Send a datagram
//
// 
int ugsi_post_send_dtgrm( IN ugsi_class_handle_t handle, IN ugsi_dtgrm_t *dtgrm);

//
// Allocate a buffer for RMMP (Relable Multi-Packet Protocol) payload.
// The buffer does not include general MAD header or RMMP header.
// Allocating the buffer automatically indicates that the datagram must be sent
// using RMPP protocol.
// The function shall not be used for non-RMPP management classes.
//
void *ugsi_dtgrm_alloc_rmpp_buf( IN ugsi_dtgrm_t *dtgrm,IN  int size );

//
// Free the RMPP buffer.
// Note - calling ugsi_dtgrm_pool_put() releases the buffer automatiacally.
//
void ugsi_dtgrm_free_rmpp_buf( IN ugsi_dtgrm_t *dtgrm );

//
// Get the buffer pointer and size.
//
void *ugsi_dtgrm_get_rmpp_buf( IN ugsi_dtgrm_t *dtgrm, OUT int *size );



#define UGSI_HCA_NAME_LEN 32
struct ugsi_hca_name {
		char name[UGSI_HCA_NAME_LEN];
};

struct ugsi_hca_properties {
	int  num_hca_names;
	struct ugsi_hca_name names[0];
};

int ugsi_list_hcas(struct ugsi_hca_properties *hcas);



#ifdef __cplusplus
};
#endif // __cplusplus
#endif
