/*
 * Copyright (C) Voltaire Ltd. 2001-2003.  ALL RIGHTS RESERVED.
 *
 * This source code (the "Code") is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest in and to the Code,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * The Code is provided "as is". To the maximum extent permitted by law,
 * the Company disclaims any warranties of any kind, either expressed or
 * implied, including, without limitation, implied warranties of
 * merchantability, fitness for a particular purpose and non-infringement.
 * The Company does not warrant that the Code will operate error free, or
 * in an uninterrupted fashion or that it is compatible with any particular
 * platform. In no event shall the Company be liable for any damages
 * whatsoever (including, without limitation, incidental, direct, indirect,
 * special and consequential damages, damages for loss of business profits,
 * business interruption, loss of business information, or other pecuniary
 * loss) arising out of the use or inability to use the Code, even if
 * advised of the possibility of such damages.
 */

#ifndef __REPOSITORY_H__
#define __REPOSITORY_H__

#ifdef __KERNEL__
#include <linux/time.h>
#else
#include <sys/time.h>
#endif

typedef long rep_key_t;

typedef struct 
{
    long  sub_keys;
    long  max_sub_key_len;
    long  value_count;
    long  max_value_name_len;
    long  max_value_len;
    long  security_descriptor;
    struct timeval  last_write_time;
} rep_key_attributes_t;

typedef long rep_security_t;
typedef long rep_notifier_t;

#define REP_MAX_KEY_NAME 32
#define REP_MAX_VAL_NAME 32

#define MAX_BIN_DATA 8000

#define KEY_QUERY_VALUE         (0x0001)
#define KEY_SET_VALUE           (0x0002)
#define KEY_CREATE_SUB_KEY      (0x0004)
#define KEY_ENUMERATE_SUB_KEYS  (0x0008)
#define KEY_NOTIFY              (0x0010)
#define KEY_CREATE_LINK         (0x0020)
 
#define KEY_READ                ((KEY_QUERY_VALUE            |\
                                  KEY_ENUMERATE_SUB_KEYS     |\
                                  KEY_NOTIFY)                 \
                                 )
 
 
#define KEY_WRITE               ((KEY_SET_VALUE              |\
                                  KEY_CREATE_SUB_KEY)         \
                                 )
 
#define KEY_EXECUTE             ((KEY_READ)                   \
                                 )
 
#define KEY_ALL_ACCESS          ((KEY_QUERY_VALUE            |\
                                  KEY_SET_VALUE              |\
                                  KEY_CREATE_SUB_KEY         |\
                                  KEY_ENUMERATE_SUB_KEYS     |\
                                  KEY_NOTIFY                 |\
                                  KEY_CREATE_LINK)            \
                                 )
                                 


//
// Open/Create Options
//

#define REP_OPTION_NON_VOLATILE     (0x00000000L)   // Key is preserved when system is rebooted
#define REP_OPTION_VOLATILE         (0x00000001L)   // Key is not preserved when system is rebooted

//
// Key creation/open disposition
//
#define REP_CREATED_NEW_KEY         (0x00000001L)   // New Registry Key created
#define REP_OPENED_EXISTING_KEY     (0x00000002L)   // Existing Key opened

//
// Notify filter values
//
#define REP_NOTIFY_CHANGE_KEY           (0x00000001L) // Create or delete (child)
#define REP_NOTIFY_CHANGE_VALUE         (0x00000002L)

//
//
// Predefined Value Types.
//
 
#define REP_NONE                    ( 0 )   // No value type
#define REP_STRING                  ( 1 )   // nul terminated string
#define REP_BINARY                  ( 3 )   // Free form binary
#define REP_LONG                    ( 4 )   // 32-bit number
#define REP_64BIT                   ( 5 )   // 64-bit number

void tree_printk(void);

rep_key_t
rep_get_root(void);

int
rep_close_key (
            rep_key_t key
            );


int
rep_create_key (
                rep_key_t key,
                char * subkey,
                long options,
                rep_security_t security_attr,
                rep_key_t * new_key,
                long *disposition
                );


int
rep_delete_key (
              rep_key_t key,
              char * subkey
              );



int
rep_delete_value (
                rep_key_t key,
                char * name
                );



int
rep_enum_key (
            rep_key_t key,
            long index,
            char * name,
            long name_len,
            struct timeval *last_write_time
            );




int
rep_enum_value (
              rep_key_t key,
              long index,
              char * name,
              long name_len,
              long * type,
              char * data,
              long * data_len
              );



int
rep_flush(void);


int
rep_load_file(
                char *  filename
            );



typedef void (*key_change_callback)(rep_key_t key, long change, void *context);

int
rep_notify_key_change(
                        rep_key_t key,
                        int watch_subtree,
                        long filter,
                        key_change_callback fn,
                        void *context,
                        rep_notifier_t *notifier
                    );

int
rep_notify_del(rep_notifier_t notifier);



int
rep_open_key (
              rep_key_t key,
              char * subkey,
              long options,
              rep_key_t * new_key
              );



int
rep_query_key(
                 rep_key_t key,
                 rep_key_attributes_t *key_attr
             );


int
rep_get_value (
                 rep_key_t key,
                 char * name,
                 long * type,
                 char * data,
                 long * data_len
                 );

int
rep_set_value (
               rep_key_t key,
               char * name,
               long type,
               char * data,
               long data_len
               );


int
rep_save_key(
            rep_key_t key,
            char * filename,
            rep_security_t * security_attributes
            );

/*
int
rep_set_key_security (
                  rep_key_t key,
                  SECURITY_INFORMATION SecurityInformation,
                  PSECURITY_DESCRIPTOR pSecurityDescriptor
                  );
                  
int
rep_get_key_security (
                  rep_key_t key,
                  SECURITY_INFORMATION SecurityInformation,
                  PSECURITY_DESCRIPTOR pSecurityDescriptor,
                  long * lpcbSecurityDescriptor
                  );

*/


///////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef _DEBUG

static char *kerror_table[] __attribute__ ((unused)) = {
    "Success" ,
    "Operation not permitted" ,
    "No such file or directory" ,
    "No such process" ,
    "Interrupted system call" ,
    "Input/output error" ,
    "No such device or address" ,
    "Argument list too long" ,
    "Exec format error" ,
    "Bad file descriptor" ,
    "No child processes" ,
    "Resource temporarily unavailable" ,
    "Cannot allocate memory" ,
    "Permission denied" ,
    "Bad address" ,
    "Block device required" ,
    "Device or resource busy" ,
    "File exists" ,
    "Invalid cross-device link" ,
    "No such device" ,
    "Not a directory" ,
    "Is a directory" ,
    "Invalid argument" ,
    "Too many open files in system" ,
    "Too many open files" ,
    "Inappropriate ioctl for device" ,
    "Text file busy" ,
    "File too large" ,
    "No space left on device" ,
    "Illegal seek" ,
    "Read-only file system" ,
    "Too many links" ,
    "Broken pipe" ,
    "Numerical argument out of domain" ,
    "Numerical result out of range" ,
    "Resource deadlock avoided" ,
    "File name too long" ,
    "No locks available" ,
    "Function not implemented" ,
    "Directory not empty" ,
    "Too many levels of symbolic links" ,
    "Unknown error 41" ,
    "No message of desired type" ,
    "Identifier removed" ,
    "Channel number out of range" ,
    "Level 2 not synchronized" ,
    "Level 3 halted" ,
    "Level 3 reset" ,
    "Link number out of range" ,
    "Protocol driver not attached" ,
    "No CSI structure available" ,
    "Level 2 halted" ,
    "Invalid exchange" ,
    "Invalid request descriptor" ,
    "Exchange full" ,
    "No anode" ,
    "Invalid request code" ,
    "Invalid slot" ,
    "Unknown error 58" ,
    "Bad font file format" ,
    "Device not a stream" ,
    "No data available" ,
    "Timer expired" ,
    "Out of streams resources" ,
    "Machine is not on the network" ,
    "Package not installed" ,
    "Object is remote" ,
    "Link has been severed" ,
    "Advertise error" ,
    "Srmount error" ,
    "Communication error on send" ,
    "Protocol error" ,
    "Multihop attempted" ,
    "RFS specific error" ,
    "Bad message" ,
    "Value too large for defined data type" ,
    "Name not unique on network" ,
    "File descriptor in bad state" ,
    "Remote address changed" ,
    "Can not access a needed shared library" ,
    "Accessing a corrupted shared library" ,
    ".lib section in a.out corrupted" ,
    "Attempting to link in too many shared libraries" ,
    "Cannot exec a shared library directly" ,
    "Invalid or incomplete multibyte or wide character" ,
    "Interrupted system call should be restarted" ,
    "Streams pipe error" ,
    "Too many users" ,
    "Socket operation on non-socket" ,
    "Destination address required" ,
    "Message too long" ,
    "Protocol wrong type for socket" ,
    "Protocol not available" ,
    "Protocol not supported" ,
    "Socket type not supported" ,
    "Operation not supported" ,
    "Protocol family not supported" ,
    "Address family not supported by protocol" ,
    "Address already in use" ,
    "Cannot assign requested address" ,
    "Network is down" ,
    "Network is unreachable" ,
    "Network dropped connection on reset" ,
    "Software caused connection abort" ,
    "Connection reset by peer" ,
    "No buffer space available" ,
    "Transport endpoint is already connected" ,
    "Transport endpoint is not connected" ,
    "Cannot send after transport endpoint shutdown" ,
    "Too many references: cannot splice" ,
    "Connection timed out" ,
    "Connection refused" ,
    "Host is down" ,
    "No route to host" ,
    "Operation already in progress" ,
    "Operation now in progress" ,
    "Stale NFS file handle" ,
    "Structure needs cleaning" ,
    "Not a XENIX named type file" ,
    "No XENIX semaphores available" ,
    "Is a named type file" ,
    "Remote I/O error" ,
    "Disk quota exceeded" ,
    "No medium found" ,
    "Wrong medium type"
};

    #define strerror(error) ((error >= 0 && error < 125) ? kerror_table[error] : "")

//#define strerror(error) (kerror_table[error])

    #define DBG_TRACE(fmt , args...) printk(KERN_INFO fmt, ##args)
 

#else // _DEBUG

    #define DBG_TRACE(fmt , args...)
    
    #define strerror(error) ("")

#endif // _DEBUG
#endif // __REPOSITORY_H__
