 /* Copyright (C) June 2000, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 * Except as specifically permitted herein, no portion of the information,
 * including but not limited to object code and source code, may be reproduced,
 * modified, distributed, republished or otherwise exploited in any form or by
 * any means for any purpose without the prior written permission of Mellanox
 * Technologies Ltd. Use of software subject to the terms and conditions
 * detailed in the file "LICENSE.txt".
 *
 * End of legal section ......................................................
 *
 */ 



#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioctl.h>
#include <linux/fs.h>

#include <asm/uaccess.h>
#include <asm/mpc8xx.h>
#include <asm/commproc.h>

/*  Hack to avoid sys/type from loading from mtcr.h */
#define _SYS_TYPES_H 1

#include "mtcr.h"
#include "mtcr_mif.h"

#define CPU_FREQ_MHZ 80

/*
 * Maximum number of times to poll for ACK after EEPROM write.
 */
#define MAX_ACKNOWLEDGE_POLLS  10
#define _8XX_MAX_I2C_LEN       64
#define MAX_I2C_RETRY         100

/*
 * Bit rate in bits per second
 */
#define XI2C_BITRATE 200000

/* we don't want to keep recalculating this each time we wiggle SCL or SDA */
static volatile cpm8xx_t* cp;

/* number of cycles per bit time (calculated by xi2c_init()) */
static int machine_cycles_per_bit;

int i2c_8xx_debug = 0;



#define SCL_UP() (cp->cp_pbdat |=  (0x80000000 >> 26))
#define SCL_DOWN() (cp->cp_pbdat &= ~(0x80000000 >> 26))
#define SDA_UP() (cp->cp_pbdat |=  (0x80000000 >> 27))
#define SDA_DOWN() (cp->cp_pbdat &= ~(0x80000000 >> 27))
#define SDA_INPUT() (cp->cp_pbdir &= ~(0x80000000 >> 27))
#define SDA_OUTPUT() (cp->cp_pbdir |=  (0x80000000 >> 27))
#define SDA_VAL() (cp->cp_pbdat & (0x80000000 >> 27))


#define WAIT_90()                             do {  \
    volatile int i = machine_cycles_per_bit >> 3;   \
    while (i--) ;                                   \
/*    schedule(); */                                \
} while(0)
#define WAIT_180() WAIT_90(); WAIT_90()


#define WIGGLE()                                \
SDA_OUTPUT();                                   \
SDA_DOWN();                                     \
SCL_DOWN();                                     \
WAIT_90();                                      \
SDA_UP();                                       \
WAIT_90();                                      \
SDA_DOWN();                                     \
WAIT_90();


static void init_sda(void)
{
    /* SDA is port B pin 27 (see 34-8 of MPC860UM) */
    SDA_OUTPUT();
    cp->cp_pbpar &= ~(0x80000000 >> 27);
    cp->cp_pbodr |=  (0x8000 >> (27-16));
    SDA_UP();
}

static void init_scl(void)
{

    /* SCL is port B pin 26 (see 34-8 of MPC860UM) */
    cp->cp_pbdir |=  (0x80000000 >> 26);
    cp->cp_pbpar &= ~(0x80000000 >> 26);
    cp->cp_pbodr |=  (0x8000 >> (26-16));
    SCL_UP();
}

static void send_start_cond(void)
{
    SDA_OUTPUT();
    SDA_UP();
    SCL_UP();

    WAIT_180();

    SDA_DOWN();

    WAIT_90();

    SCL_DOWN();

    WAIT_90();

    /* SCL must be low at this point */
}

static void send_stop_cond(void)
{
    /* SCL must be low at this point */

    SCL_DOWN();
    SDA_OUTPUT();
    SDA_DOWN();

    WAIT_90();

    SCL_UP();

    WAIT_90();

    SDA_UP();

    WAIT_180();
}

/* returns after all bytes sent, or after a byte has not been acknowledged */
/* returns 0 if all n acknowledged, or -1 if one byte not acknowledged */
static int send_bytes(char *b, int n)
{
    int nack = 0;

    /* SCL must be low at this point */
    SCL_DOWN();
    SDA_OUTPUT();

    while (n--)
    {
        /* send *b */
        unsigned char ch = (unsigned char) *b++;
        unsigned char bitmsk = 0x80;


        while (bitmsk)
        {
            (ch & bitmsk)? SDA_UP() : SDA_DOWN();
            WAIT_90();
            SCL_UP();
            WAIT_180();
            SCL_DOWN();
            WAIT_90();
            bitmsk >>=1;
        }



        /* recv ack */
        SDA_INPUT();
        WAIT_90();
        SCL_UP();
	
        WAIT_180();
        nack = SDA_VAL();
	if( !nack)
		goto got_ack;
        WAIT_180();
        nack = SDA_VAL();
	if( !nack)
		goto got_ack;
        WAIT_180();
        nack = SDA_VAL();
	if( !nack)
		goto got_ack;
        WAIT_180();
        nack = SDA_VAL();
got_ack:	
        WAIT_90();
        SCL_DOWN();
        WAIT_90();
        SDA_OUTPUT();

        if (nack)
        {
            if(i2c_8xx_debug) printk("%s: Nack was received n=%x\n",__func__, n );
            break;
        }
    }

    return nack?(-1):0;

    /* SCL must be low at this point */
}

static void recv_bytes(char *b, int n)
{
    /* SCL must be low at this point */


    while (n--)
    {
        int i;
        unsigned char ch = 0;

        SDA_INPUT();

        /* recv byte */
        for (i = 0; i < 8; i++)
        {
            unsigned char bit;
            WAIT_90();
            SCL_UP();
            WAIT_90();
            bit = (unsigned char)(SDA_VAL()?1:0);
            ch = bit | (ch<<1);
            WAIT_90();
            SCL_DOWN();
            WAIT_90();
        }
        *b++ = ch;



        /* acknowledge if we want more data still */
        SDA_OUTPUT();

        if (n)
            SDA_DOWN(); /* send ack */
        else
            SDA_UP(); /* don't send ack */

        WAIT_90();
        SCL_UP();
        WAIT_180();
        SCL_DOWN();
        WAIT_90();
    }

    /* SCL must be low at this point */
}

int send_iic_addr(char abyte)
{
    int rc;
    int retry = MAX_I2C_RETRY;


    while(1)
    {
            send_start_cond();
            rc = send_bytes(&abyte,1);
            if (rc == 0)
                    break;

            send_stop_cond();

            WAIT_180();

            retry--;

            if(retry == 0)
            {
               if(i2c_8xx_debug) printk("%s: error :  addr = %x.\n", __func__, abyte & 0xff);
               return(-1);
            }
    }

    return(0);
}




/* Read from IIC...
 * abyte = address byte, with r/w flag already set
 */
static int
i2c_8xx_recv(u_char abyte, char *buf, int count)
{
    if (i2c_8xx_debug) printk("read(abyte=0x%x)\n", abyte);


    if(send_iic_addr((abyte<<1)|1))
    {
            if(i2c_8xx_debug) printk("%s: read operation failed\n", __func__);
            return(-1);
    }

    recv_bytes(buf, count);  /*  last byte not acknowledged */
    send_stop_cond();

    return count;
}



/* Write to IIC...
 * addr = address byte, with r/w flag already set
 */
static int
i2c_8xx_send(u_char abyte, char *buf,int count)
{
    int rc = 0;

    if (count > _8XX_MAX_I2C_LEN+4)
        return -EINVAL;

    send_start_cond();

    if(send_iic_addr((abyte<<1)&0xfe))
    {
            if(i2c_8xx_debug) printk("%s: write operation (%x)\n", __func__, 0xff & abyte);
            return(-1);
    }

    rc = send_bytes(buf, count);
    send_stop_cond();

    return rc?(-EFAULT):count;
}

static int 
i2c_8xx_read(u_char addr, char * outbuf, unsigned int offset, unsigned int count, DType dtype)
{
        int rc;
        char ee[4];

        ee[0] = (offset >> 24) & 0xff;
        ee[1] = (offset >> 16) & 0xff;
        ee[2] = (offset >>  8) & 0xff;
        ee[3] = offset & 0xff;

        /*  Send address */
	switch(dtype)
	{
		
        case MST_TAVOR:
                rc = i2c_8xx_send(addr, ee, 4);

                if(rc != 4)
                {
			if(i2c_8xx_debug) printk("%s: failed sending 4-byte address.\n", __func__);
                        return(-1);
                }
                break;
		
	case MST_GAMLA:
		rc = i2c_8xx_send(addr, ee+2, 2);

		if(rc != 2)
		{
			if(i2c_8xx_debug) printk("%s: failed sending 2-byte address.\n", __func__);
			return(-1);
		}
		break;

        case MST_DIMM:
                rc = i2c_8xx_send(addr, ee+3, 1);

                if(rc != 1)
                {
			if(i2c_8xx_debug) printk("%s: failed sending 1-byte address.\n", __func__);
			return(-1);
                }
                break;		
		
	case MST_NOADDR:
		break;

	default:
		return(-1);
	}

        /*  Receive bytes */
        rc = i2c_8xx_recv(addr, outbuf, count);
        if(rc != count)
        {
                if(i2c_8xx_debug) printk("%s: failed reading from I2C.\n", __func__);
                return(-1);
        }

        return(count);
}

static int 
i2c_8xx_write(u_char addr, char * outbuf, unsigned int offset, unsigned int count, DType dtype)
{
        int rc;

        outbuf[0] = (offset >> 24) & 0xff;
        outbuf[1] = (offset >> 16) & 0xff;
        outbuf[2] = (offset >>  8) & 0xff;
        outbuf[3] = offset & 0xff;

        /*  Send  */
	switch(dtype)
	{
        case MST_TAVOR:
                rc = i2c_8xx_send(addr, outbuf, count+4);

                if(rc != (count+4))
                {
			if(i2c_8xx_debug) printk("%s: failed 4-byte addr write\n", __func__);
			return(-1);
		}
                break;		
		
	case MST_GAMLA:
            	rc = i2c_8xx_send(addr, outbuf+2, count+2);

		if(rc != (count+2))
		{
			if(i2c_8xx_debug) printk("%s: failed 2-byte addr write\n", __func__);
			return(-1);
		}
		break;

        case MST_DIMM:
                rc = i2c_8xx_send(addr, outbuf+3, count+1);
                if(rc != (count+1))
		{
			if(i2c_8xx_debug) printk("%s: failed 1-byte addr write\n", __func__);
			return(-1);
		}
                break;
		
	case MST_NOADDR:
                rc = i2c_8xx_send(addr, outbuf+4, count);

		if(rc != count)
		{
			if(i2c_8xx_debug) printk("%s: failed --byte addr write\n", __func__);
			return(-1);
		}
		break;

	default:
		return(-1);
	}

        return(count);
}



static int __init i2c_8xx_init(void)
{
    /***bd_t* bd = (bd_t*)__res;***/
    /***unsigned int system_clock = bd->bi_busfreq * 1000000;***/
    unsigned int system_clock = CPU_FREQ_MHZ * 1000000;
    machine_cycles_per_bit = system_clock/XI2C_BITRATE;
    if( i2c_8xx_debug )
	    printk( "machine_cycles_per_bit=%d, I2C_BitRate=%d\n", machine_cycles_per_bit, XI2C_BITRATE ); 

    cp = cpmp;


    init_sda();
    init_scl();

    return 0;
}


static int major;

static int gen_dev_open(struct inode *inode, struct file *pfile);
static int gen_dev_ioctl(struct inode* pinode, struct file * pfile,
                         unsigned int cmd, unsigned long arg);
static int gen_dev_close(struct inode *inode, struct file *pfile);

struct file_operations fops = {
    ioctl   :       gen_dev_ioctl,
    open    :       gen_dev_open,
    release :       gen_dev_close,
};


static int gen_dev_close(struct inode *inode, struct file *pfile)
{
    MOD_DEC_USE_COUNT;

    return(-1);
}



static int gen_dev_open(struct inode *inode, struct file *pfile)
{
    MOD_INC_USE_COUNT;
    return(0);
}

int gen_dev_ioctl(struct inode* pinode, struct file * pfile, 
                  unsigned int cmd, unsigned long arg)
{
    int rc=0;
    mif_param_t * p = (mif_param_t *)arg;
    char buf[0x80];
    
    switch ( cmd )
    {
    case MACCELERATE:
        printk("ACCELERATE: mode acceleration no supported.\n");
        return(ENOTTY);
        

    case MRESTORE:
        printk("RESTORE: mode resotre no supported.\n");
        return(ENOTTY);

        
    case MREAD64:
    case MREAD4:
        rc = i2c_8xx_read(p->addr, buf, p->offset, p->len, p->dtype);
        if(rc != p->len)
        {
                if(i2c_8xx_debug) printk("%s: failed in MREAD.\n", __func__);
                return(-1);
        }
        copy_to_user(p->buf, buf, p->len);
        return(p->len);
        
    case MWRITE64:
    case MWRITE4:
        copy_from_user(buf + 4, p->buf, p->len);
        rc = i2c_8xx_write(p->addr, buf, p->offset, p->len, p->dtype);
        if(rc != p->len)
        {
                if(i2c_8xx_debug) printk("%s: failed in MWRITE.\n", __func__);
                return(-1);
        }
        return(p->len);

    default:
        rc = ENOTTY;
    }

    return(rc);
}



int init_module(void)
{
    int rc;


    rc = i2c_8xx_init();
    if(rc)
    {
        printk("%s: failed in init.\n", __func__);
        return(rc);
    }

    major = register_chrdev(0, "mif-i2c-8xx", &fops);

    if ( major < 0)
    {
        return major;
    }


    printk( "Loaded ppc-8xx driver\n" );
    return(0);
}


void cleanup_module(void)
{
    int rc;

    rc=unregister_chrdev(major, "mif-i2c-8xx");

}



